////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \defgroup GrpEventReceiverStdOut EventReceiver StdOut
/// @{
/// \ingroup GrpEventReceiver 
/// \brief
/// Class EventReceiverStdOut analyzes the received test events and writes
/// appropriate messages to stdout.
/// Error events are counted and can be used to decide whether a test was successful.
/// 
/// \sa EventReceiverStdOut.h (File documentation)
/// \sa TestToolBox::IEventReceiver (interface documentation)
/// \sa TestToolBox::EventReceiverStdOut (simple receiver class writing to stdout)
///
////////////////////////////////////////////////////////////////////////////////

#pragma once

#include "IEventReceiver.h"
#include <iostream>

namespace TestToolBox
{

/// \brief EventReceiverStdOutanalyzes the received test events and writes
/// appropriate messages to stdout.
/// Error events are counted and can be used to decide whether a test was successful.
class EventReceiverStdOut : public IEventReceiver
{
public:

/// \cond do not document
    EventReceiverStdOut()
        : m_filterMask (0xffff)
        , m_numErrors  (0)
    {}
/// \endcond

    /// Define the event types which shall be written to stdout
    void SetOutputFilter (long in_filterMask)
    {m_filterMask = in_filterMask;}

    /// Reset the number of errors
    void ResetErrors (void)
    {m_numErrors = 0;}

    /// Get the number of errors which have occured during the test cases
    long GetNumErrors (void)
    {return m_numErrors;}

private:
/// \cond do not document

    long m_filterMask;
    long m_numErrors;

private:

	/// Signal actual event as occured during test execution
	virtual void EventMsg(
		EventContext in_context,
		const char*  in_eventDescription)
	{
        if (!(in_context & m_filterMask))
            return;

        // Generate message containing two lines to enable both jumping to
        // this line and the line where the TestEvent-Macro was called
        // (assume that in_eventDescription contains location info)
        std::ostringstream fullInfoStr;

        if (in_context & CTX_PROT)
        {
            fullInfoStr << in_eventDescription;
        }
        else
        {
            if (in_context & CTX_INFO)
            {
                fullInfoStr << "INFO: " << in_eventDescription;
            }
            else if (in_context & CTX_TEST_EVENT)
            {
                fullInfoStr << "EVENT: " << in_eventDescription;
            }
            else
            {
                // increment error count
                ++m_numErrors;
                if (in_context & CTX_UNEXPECTED_EVENTS)
                {
                    fullInfoStr << in_eventDescription << " (UNEXPECTED_EVENTS)";
                }
                else if (in_context & CTX_FATAL_ERROR)
                {
                    fullInfoStr << in_eventDescription << " (FATAL_ERROR)";
                }
                else
                {
                    fullInfoStr << in_eventDescription;
                }
            }
        }
        std::cout << fullInfoStr.str().c_str() << std::endl;
    }
/// \endcond do not document

}; // EventReceiverStdOut

}; // namespace TestToolBox


/// @}
