////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \defgroup GrpEnvironmentForBoostTest Using TestEnvironmentForBoost
/// @{
/// \ingroup GrpTestUtils 
/// \brief
/// Provides tools to support several test problems when running tests
///  based on Boost.Test library.
///
/// \n The supported features are:
/// 
/// - synchronisation of test execution running within
///   several threads
/// 
/// - automatic verification of asynchronously occured events
///   which may be in fixed or non deterministic order
/// 
/// - catching and verifying error objects emitted from the
///   system under test, support for filtering and error trees
///
/// \sa BoostDefaultTestEnvironment.h (File documentation)
/// \sa TestToolBox::AutoSetupDefaultTestEnv (class documentation)
/// \sa TestToolBox::AutoSetupTestEnvWithErrors (class documentation)
/// \sa GrpTestExtensionsForBoostTest (unit test using boost.Test and ErrorEvents)
/// 
////////////////////////////////////////////////////////////////////////////////

#pragma once
#pragma warning(push, 3)
#include "../Synchronizer.h"
#include "../Environment.h"
#include "../TestEvents.h"
#include "../ErrorEvents.h"
#include "../Protocol.h"
#include "../TestTime.h"
#include "BoostEventReceiverForSynchronizer.h"
#include "BoostEventReceiverForTestEvents.h"

#include <iostream>

#pragma warning(disable: 4311)
#pragma warning(disable: 4312)
#include <boost/thread.hpp>
#pragma warning(default: 4311)
#pragma warning(default: 4312)

#pragma warning(pop) 

namespace TestToolBox
{


/// Test environment with support for synchronisation and asynchronous test events
class AutoSetupDefaultTestEnv
{
///\cond

public:

    // Default constructor, always using (possibly already existing) protocol file
    AutoSetupDefaultTestEnv();

    /// Constructor allowing use of externally controlled protocol file
    /// with possibly longer lifetime than the DefaultTestEnvironment
    AutoSetupDefaultTestEnv(
        IProtocol* in_pProtocol);

    virtual ~AutoSetupDefaultTestEnv();

protected:

    bool m_cleanupProtocol;
    bool m_cleanupTestEvents;

    BoostEventReceiverForSynchronizer m_eventReceiverForSynchronizer;
    BoostEventReceiverForTestEvents   m_eventReceiverForTestEvents;
///\endcond

};



///\cond

///\endcond


class AutoTestCase
{
public:
    explicit AutoTestCase(
        std::string const & in_testCaseDescription,
        char const *        in_file = 0,
        int const           in_line = 0);
    ~AutoTestCase(void);

    private:
        char const * m_file;
        int const    m_line;
};

class FinalCleanup
{
public:
    virtual ~FinalCleanup (void);
};

class DefaultSettingsForBoostAndTTB : public FinalCleanup
{
public:
    DefaultSettingsForBoostAndTTB (
        unsigned char in_protocolMode = OutputMode::M_STDOUT | OutputMode::M_OUTFILE  | OutputMode::M_RPTFILE,
        bool          in_detailedLog  = true);

    ~DefaultSettingsForBoostAndTTB();

protected:
    BoostEventReceiverForSynchronizer m_eventReceiverForSynchronizer;
    BoostEventReceiverForTestEvents   m_eventReceiverForTestEvents;
};


#define TTB_AUTO_TEST_CASE(DESC) TestToolBox::AutoTestCase autoTestCase(DESC,__FILE__,__LINE__)

#ifdef TTB_BOOST_NO_AUTOREGISTER
#undef TTB_BOOST_AUTOREGISTER
#else
#ifndef TTB_BOOST_AUTOREGISTER
#define TTB_BOOST_AUTOREGISTER
#endif
#endif

#ifdef TTB_BOOST_AUTOREGISTER
#define _TEST_CASE(name) BOOST_AUTO_TEST_CASE (name)
#else
#define _TEST_CASE(name) void name()
#endif

#define TTB_BOOST_TEST_CASE(name) _TEST_CASE(name) { TTB_AUTO_TEST_CASE(#name);
#define TTB_BOOST_FIXTURE_TEST_CASE(name, fixture) BOOST_FIXTURE_TEST_CASE(name,fixture) { TTB_AUTO_TEST_CASE(#name);


#define TTB_TEST_CASE(DESC)                                                                          \
    TestToolBox::TestEvents::Get()->CheckForUnexpectedEvents  ("next test case", __FILE__,__LINE__); \
    TestToolBox::TestEvents::Get()->StartNewTestCase(DESC);

#define TTB_BOOST_CHECK_FOR_TEST_EVENTS()                       \
    BOOST_CHECK(TestToolBox::TestEvents::Get()->AllTestsHaveSucceeded());



inline void DummyFunction(){}

//class WorkAroundBoostThreadMemLeak
//{
//public:
//    WorkAroundBoostThreadMemLeak ();
//    ~WorkAroundBoostThreadMemLeak ();
//
//private:
//    _CrtMemState memStateStartup;
//    _CrtMemState memStateShutdown;
//
//};

}; // namespace TestToolBox

/// @} // GrpTestEnvironmentForBoost

