////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \sa BoostDefaultTestEnvironmentImpl.h (File documentation)
/// \sa GrpTestExtensionsForBoostTest (unit test using boost.Test and ErrorEvents)
/// 
////////////////////////////////////////////////////////////////////////////////

#pragma once


#include "../CommonSources/CommonDefinitions.h"
#include "../ConditionalFixtureImpl.h"
#include "../SynchronizerImpl.h"
#include "../EnvironmentImpl.h"
#include "../TestEventsImpl.h"
#include "../ErrorEventsImpl.h"
#include "../ProtocolImpl.h"
#include "../TestTimeImpl.h"
#include "../TestStreamImpl.h"
#include "../SimulatedObjectBaseImpl.h"
#include "BoostEventReceiverForSynchronizer.h"
#include "BoostEventReceiverForTestEvents.h"
#include "BoostDefaultTestEnvironment.h"

#include <iostream>

#pragma warning(disable: 4311)
#pragma warning(disable: 4312)
#include <boost/thread.hpp>
#pragma warning(default: 4311)
#pragma warning(default: 4312)
#include <boost/test/results_collector.hpp>
//#include <boost/test/unit_test_suite_impl.hpp>
#include <boost/test/unit_test_suite.hpp>

namespace TestToolBox
{


///\cond

    // Default constructor, always using (possibly already existing) protocol file
    TTB_INLINE AutoSetupDefaultTestEnv::AutoSetupDefaultTestEnv()
    : m_cleanupTestEvents (true)
    , m_cleanupProtocol   (true)
    {
	    Synchronizer::Get()->
            ConnectWithEventReceiver(&m_eventReceiverForSynchronizer);
        TestEvents::Get()->
            ConnectWithEventReceiver(&m_eventReceiverForTestEvents);
        TestEvents::Get()->
            ConnectWithProtocol (Protocol::Get());
    }

    /// Constructor allowing use of externally controlled protocol file
    /// with possibly longer lifetime than the DefaultTestEnvironment
    TTB_INLINE AutoSetupDefaultTestEnv::AutoSetupDefaultTestEnv(
        IProtocol* in_pProtocol)
    : m_cleanupTestEvents (true)
    , m_cleanupProtocol   (false)
    {
	    Synchronizer::Get()->
            ConnectWithEventReceiver(&m_eventReceiverForSynchronizer);
        TestEvents::Get()->
            ConnectWithEventReceiver(&m_eventReceiverForTestEvents);
        if (in_pProtocol)
        {
            TestEvents::Get()->
                ConnectWithProtocol (in_pProtocol);
        }
    }

    TTB_INLINE AutoSetupDefaultTestEnv::~AutoSetupDefaultTestEnv()
    {
        if (m_cleanupTestEvents)
        {
            TestEvents::Cleanup();
        }
        Synchronizer::Cleanup();
        Environment::Cleanup();
        GlobalOptions::Cleanup();
        Formatter::Cleanup();
        RegisteredSimulatedObjects::Cleanup();

        if (m_cleanupProtocol)
        {
            Protocol::Cleanup();
        }
    }
///\endcond





///\cond

    //-------------------------------------------------------------------------

    TTB_INLINE AutoTestCase::AutoTestCase(
        std::string const & in_testCaseDescription,
        char const *        in_file,
        int const           in_line)
    : m_file(in_file)
    , m_line (in_line)
    {
        TestToolBox::TestEvents::Get()->StartNewTestCase(in_testCaseDescription);
        MyErrorEvents::Get()->DeleteStoredErrors(); // ignore all preceding errors

    }

    //-------------------------------------------------------------------------

    TTB_INLINE AutoTestCase::~AutoTestCase(void)
    {
        TestToolBox::Protocol::Get()->Writeln(
            TestToolBox::OutputLevel::eREGULAR, TestToolBox::OutputMode::M_OUTFILE,
            "\n--- checks at end of test case ---");

        // Verify that no unchecked errors are left
        MyErrorEvents::Get()->ReportStoredErrors();
        TTB_EXP("No errors stored");



        // Check if boost checks have failed
        bool boostTestCaseHasSucceeded = boost::unit_test::results_collector.results( 
            boost::unit_test::framework::current_test_case().p_id ).passed();
        if (!boostTestCaseHasSucceeded)
        {
            TTB_ACT_S("TTB: BOOST-Checks have failed!");
        }

        // Does only work when working without test suites
        //unsigned int boostTestCaseAssertCounter = 0;
        //{
        //    using boost::unit_test::test_results;
        //    using boost::unit_test::results_collector;
        //    test_results tr = results_collector.results(boost::unit_test::framework::master_test_suite().get(
        //        TestToolBox::TestEvents::Get()->GetDescCurTest().c_str()));
        //    boostTestCaseAssertCounter = tr.p_assertions_failed;
        //    if (boostTestCaseAssertCounter > 0)
        //    {
        //        TTB_ACT_S("BOOST-Checks failed: numFailedAssertions=" << boostTestCaseAssertCounter);
        //    }
        //}
        // bool boostTestCaseHasSucceeded = (boostTestCaseAssertCounter == 0);

        // Check for any unexpected events
        TestToolBox::TestEvents::Get()->CheckForUnexpectedEvents  ("~AutoTestCase", m_file,m_line);

        // inform boost about TTB failures only if BOOST still assumes success
        if (boostTestCaseHasSucceeded)
        {
            bool TTB_EndTestCaseIsOk = TestToolBox::TestEvents::Get()->CurTestIsStillOk();
            BOOST_CHECK(TTB_EndTestCaseIsOk);
        }

        TestToolBox::Protocol::Get()->Writeln(
            TestToolBox::OutputLevel::eREGULAR, TestToolBox::OutputMode::M_OUTFILE,
            "\nTTB: test case " + std::string((TestToolBox::TestEvents::Get()->CurTestIsStillOk() ? "succeeded" : "failed")));
    }

    //-------------------------------------------------------------------------

    TTB_INLINE FinalCleanup::~FinalCleanup (void)
    {
        ConditionalFixture::Cleanup();
        Synchronizer::Cleanup();
        TestEvents::Cleanup();
        MyErrorEvents::Cleanup();
        Protocol::Cleanup();
        TestTime::Cleanup();
        Environment::Cleanup();
        GlobalOptions::Cleanup();
        Formatter::Cleanup();
        RegisteredSimulatedObjects::Cleanup();
#if defined(TTB_USE_PUGI_XML) || defined (TTB_USE_TINY_XML)
        TestToolBox::XmlCheck::Cleanup();
#endif

    }

    //-------------------------------------------------------------------------

    TTB_INLINE DefaultSettingsForBoostAndTTB::DefaultSettingsForBoostAndTTB (
        unsigned char in_protocolMode,
        bool          in_detailedLog)
    {
        //-----  Create components of test environment  -----

        TheEnvironment()->CutSourceDirAfter("MySwArchive\\");

        Protocol::Get(in_protocolMode);

        Synchronizer::Get();
        TestEvents::Get();
        MyErrorEvents::Get();

        // Automatically remove errors which have been reported as TestEvents (recommended!)
        MyErrorEvents::Get()->DeleteStoredErrorsAfterReport(true);

        if (in_protocolMode == OutputMode::M_NULL) // i.e. no protocol file
        {
            // Connect with Boost macros (only recommended if not connected to Protocol/TestEvents!)
	        Synchronizer::Get()->
                ConnectWithEventReceiver(&m_eventReceiverForSynchronizer);
            TestEvents::Get()->
                ConnectWithEventReceiver(&m_eventReceiverForTestEvents);
        }
        else // prefer usage of Protocol and TestEvents
        { 
	        Synchronizer::Get()->
                ConnectWithEventReceiver(TestEvents::Get());

            // Connect with test protocol (recommended!)
            // (additional connection to BOOST macros is not needed)
            TestEvents::Get()->
                ConnectWithProtocol (Protocol::Get());
        }

        TestEvents::Get()->SetDetailedLog(in_detailedLog);
#if defined(TTB_USE_PUGI_XML) || defined (TTB_USE_TINY_XML)
		TestToolBox::TheXmlCheck()->SetEventReceiver(TestToolBox::TestEvents::Get());
#endif
    }

    //-------------------------------------------------------------------------

    TTB_INLINE DefaultSettingsForBoostAndTTB::~DefaultSettingsForBoostAndTTB ()
    {
    }

    //-------------------------------------------------------------------------

    //TTB_INLINE WorkAroundBoostThreadMemLeak::WorkAroundBoostThreadMemLeak ()
    //{
    //    _CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF /*| _CRTDBG_LEAK_CHECK_DF */);
    //    _CrtSetReportMode( _CRT_WARN, _CRTDBG_MODE_DEBUG );
    //    _CrtSetReportMode( _CRT_ERROR, _CRTDBG_MODE_DEBUG );
    //    _CrtSetReportMode( _CRT_ASSERT, _CRTDBG_MODE_DEBUG );
    //    _CrtSetReportFile(_CRT_WARN, _CRTDBG_FILE_STDOUT);

    //    // Use leaky boost methods before storing memory state
    //    {
    //         // onetime initialization causes
    //         // memory leak because there is no cleanup on exit program
    //         boost::thread t(DummyFunction);
    //         t.join();
    //    }
    //    // From now on we are sensible to memleaks
    //    // Reporting is done to stdout
    //    _CrtMemCheckpoint(&memStateStartup);
    //}
    //TTB_INLINE WorkAroundBoostThreadMemLeak::~WorkAroundBoostThreadMemLeak ()
    //{
    //    cout << "MEMLEAK REPORT (Workaround Boost thread memleak / run with --detect_memory_leak=0)" << endl;
    //    _CrtMemCheckpoint(&memStateShutdown);
    //    _CrtMemState diffState;
    //    IGNORE_PARAM(diffState); // avoid warning in Release builds
    //    if (_CrtMemDifference(&diffState,&memStateStartup,&memStateShutdown))
    //    {
    //        _CrtMemDumpStatistics(&diffState);
    //        _CrtMemDumpAllObjectsSince(&memStateStartup);
    //        cout << "MEMLEAK REPORT: End of leaks report!" << endl;
    //    } 
    //    else
    //    {
    //        cout << "MEMLEAK REPORT: No leaks detected!" << endl;
    //    }
    //    //_CrtMemDumpAllObjectsSince(&memStateStartup);
    //}

}; // namespace TestToolBox

/// @} // GrpTestEnvironmentForBoost

