////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \defgroup GrpFilter Using Filter
/// @{
/// \ingroup GrpTestUtils 
/// \brief
/// Classes Filter and Match allow to define complex filter conditions which can be used
/// for arbitrary purposes. One possible usage is within simulated objetcs to
/// trigger a specific reaction only when special (filtered) conditions apply.
///
/// <b>\n Example</b>\n
///
/// <b>Define some filters</b>
/// \code
/// FilterSp spFilterHugoAndNotOtto = Match("Hugo") & !Match("Otto");
/// FilterSp spFilterHugoAndNotOtto_Or_HugoAndPeter = (Match("Hugo") & !Match("Otto")) | (Match("Hugo") & Match("Peter"));
/// FilterSp spFilterHugoXorFelix = Match("Hugo") ^ Match("Felix");
/// FilterSp spFilterHugoAndNotOtto_Or_HugoXorFelix = Match(spFilterHugoAndNotOtto) | Match(spFilterHugoXorFelix);
/// \endcode
///
////////////////////////////////////////////////////////////////////////////////

#pragma once

#include <memory>
#include <string>

namespace TestToolBox
{
	//-----------------------------------------------------------------------------

	class Filter;
	typedef std::shared_ptr<Filter> FilterSp;

	/// Class Filter and its derived classes allows the definition of complex filter
	/// conditions for text strings
	class Filter : std::enable_shared_from_this<Filter>
	{
	public:
		enum class LogicOp
		{
			AND,
			OR,
			XOR
		};

		virtual ~Filter(){}

		virtual bool Matches(std::string const & in_someText) const = 0;
	};

	//-----------------------------------------------------------------------------

	/// Supports a simple text match
	class TextFilter : public Filter
	{
	public:
		explicit TextFilter(std::string const & in_pattern)
			: m_pattern(in_pattern)
		{}

		bool Matches(std::string const & in_someText) const override
		{
			bool patternFound = (in_someText.find(m_pattern) != std::string::npos);
			return patternFound;
		}
	private:
		std::string m_pattern;
	};

	//-----------------------------------------------------------------------------

	/// Supports negation of an other filter
	class NegateFilter : public Filter
	{
	public:
		explicit NegateFilter(FilterSp const & in_spFilter)
			: m_spFilter(in_spFilter)
		{}

		bool Matches(std::string const & in_someText) const override
		{return !(m_spFilter->Matches(in_someText));}

	private:
		FilterSp m_spFilter;
	};

	//-----------------------------------------------------------------------------

	/// Supports logical combination /AND, OR, XOR) of two filters
	class LogicFilter : public Filter
	{
	public:
		explicit LogicFilter(LogicOp const & in_logicOp, FilterSp const & in_lhs, FilterSp const & in_rhs)
			: m_logicOp(in_logicOp)
			, m_lhs(in_lhs)
			, m_rhs(in_rhs)
		{}

		bool Matches(std::string const & in_someText) const override;

	private:
		LogicOp m_logicOp;
		FilterSp m_lhs;
		FilterSp m_rhs;
	};

	//-----------------------------------------------------------------------------

	/// Class Match is a helper class which allows definition and combination of filter conditions
	class Match
	{
	public:
		explicit Match(std::string const & in_someText)
		{m_spFilter = std::make_shared<TextFilter>(in_someText);}

		explicit Match(FilterSp const & in_spFilter)
		{m_spFilter = in_spFilter;}

		Match And(Match const & in_rhs) const
		{return Match(std::make_shared<LogicFilter>(Filter::LogicOp::AND, m_spFilter, in_rhs.m_spFilter));}

		Match Or(Match const & in_rhs) const
		{return Match(std::make_shared<LogicFilter>(Filter::LogicOp::OR, m_spFilter, in_rhs.m_spFilter));}

		Match Xor(Match const & in_rhs) const
		{return Match(std::make_shared<LogicFilter>(Filter::LogicOp::XOR, m_spFilter, in_rhs.m_spFilter));}

		friend Match operator& (Match const & lhs, Match const & rhs)
		{return lhs.And(rhs);}

		friend Match operator| (Match const & lhs, Match const & rhs)
		{return lhs.Or(rhs);}

		friend Match operator^ (Match const & lhs, Match const & rhs)
		{return lhs.Xor(rhs);}

		friend Match operator! (Match const & rhs)
		{return Match(std::make_shared<NegateFilter>(rhs.m_spFilter));}

		// allow assignment to FilterSp
		operator FilterSp() const {return m_spFilter;}
	private:
		FilterSp m_spFilter;
	};

	//-----------------------------------------------------------------------------

	bool LogicFilter::Matches(std::string const & in_someText) const
	{
		bool retVal = false;
		switch (m_logicOp)
		{
		case LogicOp::AND:
			retVal = m_lhs->Matches(in_someText) && m_rhs->Matches(in_someText);
			break;
		case LogicOp::OR:
			retVal = m_lhs->Matches(in_someText) || m_rhs->Matches(in_someText);
			break;
		case LogicOp::XOR:
		{
			bool lhsMatch = m_lhs->Matches(in_someText);
			bool rhsMatch = m_rhs->Matches(in_someText);
			retVal = (lhsMatch && !rhsMatch) || (!lhsMatch && rhsMatch);
		}
		break;
		}
		return retVal;
	}

} // namespace TestToolBox

////////////////////////////////////////////////////////////////////////////////
/// 
/// <b>\n Test environment</b>\n
/// \code
/// std::vector<std::string> testData = {
/// "Hugo, Anna",
/// "Hugo, Anna, Otto, Felix",
/// "Peter, Otto, Hugo, Anna",
/// "Peter, Felix, Anna, Julia",
/// "Hugo, Felix, Anna, Julia"
/// };
/// 
/// #define TEST_MATCH(filter) TestMatch(filter, #filter);
/// 
/// void TestMatch(FilterSp spFilter, std::string const & in_nameFilter)
/// {
/// 	std::cout << std::endl;
/// 
/// 	for (auto const & text : testData)
/// 	{
/// 		std::cout << in_nameFilter.c_str()
/// 			<< (spFilter->Matches(text) ? "     Matched " : " Not Matched ")
/// 			<< text.c_str() << std::endl;
/// 	}
/// }
///
/// // Run test cases within main:
/// TEST_MATCH(spFilterHugoAndNotOtto);
/// TEST_MATCH(spFilterHugoAndNotOtto_Or_HugoAndPeter);
/// TEST_MATCH(spFilterHugoXorFelix);
/// TEST_MATCH(spFilterHugoAndNotOtto_Or_HugoXorFelix);
/// \endcode
///
/// <b>\n Output</b>\n
/// \code
/// spFilterHugoAndNotOtto     Matched Hugo, Anna
/// spFilterHugoAndNotOtto Not Matched Hugo, Anna, Otto, Felix
/// spFilterHugoAndNotOtto Not Matched Peter, Otto, Hugo, Anna
/// spFilterHugoAndNotOtto Not Matched Peter, Felix, Anna, Julia
/// spFilterHugoAndNotOtto     Matched Hugo, Felix, Anna, Julia
/// 
/// spFilterHugoAndNotOtto_Or_HugoAndPeter     Matched Hugo, Anna
/// spFilterHugoAndNotOtto_Or_HugoAndPeter Not Matched Hugo, Anna, Otto, Felix
/// spFilterHugoAndNotOtto_Or_HugoAndPeter     Matched Peter, Otto, Hugo, Anna
/// spFilterHugoAndNotOtto_Or_HugoAndPeter Not Matched Peter, Felix, Anna, Julia
/// spFilterHugoAndNotOtto_Or_HugoAndPeter     Matched Hugo, Felix, Anna, Julia
/// 
/// spFilterHugoXorFelix     Matched Hugo, Anna
/// spFilterHugoXorFelix Not Matched Hugo, Anna, Otto, Felix
/// spFilterHugoXorFelix     Matched Peter, Otto, Hugo, Anna
/// spFilterHugoXorFelix     Matched Peter, Felix, Anna, Julia
/// spFilterHugoXorFelix Not Matched Hugo, Felix, Anna, Julia
/// 
/// spFilterHugoAndNotOtto_Or_HugoXorFelix     Matched Hugo, Anna
/// spFilterHugoAndNotOtto_Or_HugoXorFelix Not Matched Hugo, Anna, Otto, Felix
/// spFilterHugoAndNotOtto_Or_HugoXorFelix     Matched Peter, Otto, Hugo, Anna
/// spFilterHugoAndNotOtto_Or_HugoXorFelix     Matched Peter, Felix, Anna, Julia
/// spFilterHugoAndNotOtto_Or_HugoXorFelix     Matched Hugo, Felix, Anna, Julia
/// \endcode
/// 
/// @} // end group GrpFilter
