////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// @{
/// \ingroup GrpSynchronizer 
/// \brief interface to interact with a Synchronizer object
///
/// \sa ISynchronizer.h (File documentation)
/// \sa TestToolBox::ISynchronizer (interface documentation)
///
////////////////////////////////////////////////////////////////////////////////

#pragma once

namespace TestToolBox
{

    struct IEventReceiver;

    // ----------  Interface ISynchronizer  ----------

    /// \brief This interface is used for synchronizing a linear test sequence
    //( with parallel activities of the test objects.
    struct ISynchronizer
    {
        /// \brief Connect with callback interface. This method is intended to be
        /// used by the specific report facility of the test environment.
        /// The synchronizer will inform the tets environment
        /// through the callback interface when error conditions are detected.
        virtual void ConnectWithEventReceiver(
            IEventReceiver* in_pIEventReceiver) = 0;

        /// Define maximum timeout to wait for sync events.
        virtual void SetSyncTimeout(long in_timeoutMs) = 0;

        /// Init wait conditions. Define the number of events to wait for.
        virtual void InitSync(long in_numSyncEventsToWaitFor) = 0;

        /// \brief Wait until the configured number of sync events has occurred.
        /// The number of sync events to wait for is given by the last call of
        /// InitSync().
        /// If not all sync events arrive within a specific period of time
        /// a timeout error is generated and the function returns.
        /// The timeout is defined by the last call of SetSyncTimeout().
        virtual void WaitSync(
            const char* in_fileName = 0,
            const long  in_lineNum  = 0) = 0;

        /// \brief Signals that a single sync event has occured.
        /// The event represents an arbitrary action that has occured
        /// during the test (e.g. an asynchronous callback was executed,
        /// a notification arrived, an asynchronous telegram was received).
        /// In most cases before generating the sync event the information
        /// about the kind of the event (e.g. the contents of the received
        /// telegram) has to be stored within the specific test environment
        /// to make possible an automatic verification of test events even
        /// in case of asynchronous occurance.
        virtual void Sync(
            const char* in_fileName = 0,
            const long  in_lineNum  = 0) = 0;
    };


}; // namespace TestToolBox

