////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \defgroup GrpXmlCheck Using XmlCheck
/// @{
/// \ingroup GrpTestUtils 
/// \brief
/// Class XmlCheck allows to check the contents of an XML tree/subtree.
///
/// <b>\n Supported features</b>
///
///  - check of XML trees and subtrees
///
///  - support for XPath navigation
///
///  - skipping specific attributes/nodes
///
///  - treating float attributes/nodes with configurable precision
///
/// <b>Depends on</b>
///
///  - TinyXpath (includes TinyXml)
///
/// <b>HowTo access XML nodes with TinyXml</b>
/// \code
///
/// // Read XML string into XML document structure
/// TiXmlDocument doc;
/// doc.Parse(DEMO_XML_STRING);
///
/// // Save XML document to file
/// std::string filePath = MY_TEST_DIR + "/GeneratedXmlFile.xml";
/// doc.SaveFile(filePath.c_str());
///
/// // Read XML file into XML document structure
/// TiXmlDocument docFromFile;
/// docFromFile.LoadFile(filePath.c_str());
/// \endcode
///
///
/// <b>HowTo address subnodes with TinyXpath</b>
/// <table>
///   <tr>
///     <td>/MyContainer/Group_B</td>
///     <td>subnode Group_B</td>
///   </tr>
///   <tr>
///     <td>/MyContainer/Group_B/x</td>
///     <td>first x element within Group_B</td>
///   </tr>
///   <tr>
///     <td>/MyContainer/Group_B/x[3]</td>
///     <td>third x element within Group_B</td>
///   </tr>
///   <tr>
///     <td>/MyContainer/Group_B/x[last()]</td>
///     <td>last x element of Group_B</td>
///   </tr>
///   <tr>
///     <td>/MyContainer/Country[\@name='Germany']</td>
///     <td>first country with specific attribute 'name'</td>
///   </tr>
///   <tr>
///     <td>/MyContainer/Country[attribute::name='USA']</td>
///     <td>first country with specific attribute 'name' (explicit syntax)</td>
///   </tr>
/// </table>
///
///
/// <b>Typical code snippets when used with Boost.Test</b><br><br>
///    see also \ref GrpTestXmlCheckWithBoost
///
///  - comparing complete XML trees
///    <pre>
///
///    BOOST_MESSAGE("\nComparing two XML trees");
///    BOOST_CHECK(TTB::TheXmlCheck()-><b>Compare</b>(&docA,&docB));
///    BOOST_MESSAGE(TTB::TheXmlCheck()->GetCompareResult());
///    </pre>
///
///  - comparing XML subtrees and ignoring attributes/nodes
///    <pre>
///
///    BOOST_MESSAGE("\nComparing two XML subtrees using XPath, ignoring time");
///    TTB::TheXmlCheck()-><b>IgnoreAttribute</b>("time");
///    BOOST_CHECK(TTB::TheXmlCheck()->Compare(
///        TTB::XmlCheck::<b>SubNodeFromXPath</b>(docA.RootElement(), "/MyContainer/Group_B"),
///        TTB::XmlCheck::<b>SubNodeFromXPath</b>(docB.RootElement(), "/MyContainer/Group_B")));
///    BOOST_MESSAGE(TTB::TheXmlCheck()->GetCompareResult());
///    </pre>
///
///
/// <b>Typical code snippets when used with Boost.Test and TestToolBox::TestEvents</b><p>
///    see also \ref GrpTestXmlCheckWithBoostExtensions
///
///  - checking XML contents, ignoring some attributes/nodes
///    <pre>
///
///    TTB::TheXmlCheck()->IgnoreAttribute("date");
///    TTB::TheXmlCheck()->IgnoreAttribute("time");
///    TTB::TheXmlCheck()->IgnoreNode("date");
///    TTB::TheXmlCheck()->UseAsDoubleAttribute("tolerance", true, 3);
///    TTB::TheXmlCheck()->UseAsDoubleNode("x", true, 4);
///    TTB::TheXmlCheck()-><strong>CheckNode(&doc)</strong>;
///
///    TTB_EXP("MyContainer");
///    TTB_EXP("  Group_A");
///    TTB_EXP("    Vector");
///    TTB_EXP("      id: 1");
///    TTB_EXP("      tolerance: 0.026");
///    TTB_EXP("      This is the text of the first vector element");
///    TTB_EXP("      x");
///    TTB_EXP("        3.1424");
///    TTB_EXP("    Vector");
///    TTB_EXP("      id: 2");
///    TTB_EXP("      tolerance: 0.026");
///    TTB_EXP("      This is the text of the second vector element");
///    TTB_EXP("      x");
///    TTB_EXP("        1.3264");
///    </pre>
///
///  - checking XML contents, accessing subtree using XPath
///    <pre>
///
///    TTB::TheXmlCheck()-><b>CheckSubNode(doc.RootElement(),""/MyContainer/Group_A/Vector[last()]")</b>;
///
///    TTB_EXP("Vector");
///    TTB_EXP("  id: 3");
///    TTB_EXP("  tolerance: 1.357");
///    TTB_EXP("  x");
///    TTB_EXP("    6.28532");
///
///     or as alternative syntax:
///
///     TTB::TheXmlCheck()-><b>CheckSubNode</b>(
///         TTB::XmlCheck::<b>SubNodeFromXPath</b>(doc.RootElement(),<b>"/MyContainer/Group_A"</b>),
///         <b>"Vector[last()]"</b>);
///
///    or identifying XML contents via <b>multiple attributes / multiple XPath expressions</b>:
///
///    XPath: Direct acess to third node hierarchy (street within town within country)
///    TTB::TheXmlCheck()->CheckSubNode(doc.RootElement(),
///        "Country[@name='Germany']",
///        "Town[@name='Regensburg']",
///        "Street[@name='LappersdorferStrasse']");
///
///    TTB_EXP("Street");
///    TTB_EXP("  name: LappersdorferStrasse");
///    TTB_EXP("  Contents for LappersdorferStrasse in Regensburg, Germany");
///    </pre>
///
///
///  - comparing two XML trees
///    <pre>
///
///    TTB::TheXmlCheck()->IgnoreAttribute("time");
///    TTB::TheXmlCheck()-><b>Compare(&docA,&docB)</b>;
///
///    TTB_EXP("Difference detected");
///    TTB_EXP("A: tolerance: 0.0258 (line: 7 col: 54, path: //MyContainer/Group_A/Vector)");
///    TTB_EXP("B: tolerance: 0.8258 (line: 9 col: 54, path: //MyContainer/Group_A/Vector)");
///
///
///    TTB::TheXmlCheck()->Compare(
///        TTB::XmlCheck::<b>SubNodeFromXPath</b>(docA.RootElement(), "/MyContainer/Group_B"),
///        TTB::XmlCheck::<b>SubNodeFromXPath</b>(docB.RootElement(), "/MyContainer/Group_B"));
///
///    TTB_EXP("Difference detected");
///    TTB_EXP("A: 3.132437 (line: 18 col: 8, path: //MyContainer/Group_B/x/3.132437/)");
///    TTB_EXP("B: 12.132437 (line: 20 col: 8, path: //MyContainer/Group_B/x/12.132437/)");
///    </pre>
///
///
/// <b>Code usage</b>
///
/// There are two main possibilities to use XmlCheck within your test applications:
///
/// - use as "inline code"
///   - from all source files use <code>#include "TestToolBox\XmlCheckImpl.h"</code><br>
///   - example see \ref TestXmlCheckWithBoost.cpp<br>
///   - Disadvantage: each cpp file contains the same implementation code
///
/// - use as "library" code:<br>
///   - from all source files use <code>#include "TestToolBox\XmlCheck.h"</code><br>
///   - add implementation code to a single source file e.g. stdafx.cpp
/// \code
/// // Compile as real functions which are available to other
/// // source files which need only include TestToolBox\XmlCheck.h
/// #define TTB_NOT_INLINE
/// #include "TestToolBox\XmlCheckImpl.h"
/// \endcode
///   - example see \ref TestXmlCheckWithBoostExtensions.cpp<br>
///
/// \sa XmlCheck.h (File documentation)
/// \sa TestToolBox::XmlCheck (class documentation)
/// \sa GrpTestXmlCheckBasic (basic tests)
/// \sa GrpTestXmlCheckWithBoost (tests with Boost)
/// \sa GrpTestXmlCheckWithBoostExtensions (tests with Boost and TestToolBox::TestEvents)
/// 
////////////////////////////////////////////////////////////////////////////////

#pragma once

#include <set>
#include <map>
#include <vector>

class TiXmlNode;
class TiXmlElement;


//************************************************************************

namespace TestToolBox
{

struct IEventReceiver;
enum EventContext;

//************************************************************************
//************************************************************************


/// \brief Check an XML tree/subtree or compare two trees/subtrees.
/// For more info about usage see \ref GrpXmlCheck.
///
/// \nosubgrouping 
/// \sa XmlCheck.h (File documentation)
/// \sa GrpTestXmlCheckBasic (basic tests)
/// \sa GrpTestXmlCheckWithBoost (tests with Boost)
/// \sa GrpTestXmlCheckWithBoostExtensions (tests with Boost and TestToolBox::TestEvents)
class XmlCheck
{
public:

///\cond do not document

    // Constructor
	XmlCheck();

    // Destructor
	virtual ~XmlCheck();

    // Get the pointer to the only instance of this class
    static XmlCheck* Get (void);

    static XmlCheck* s_pXmlCheck;

    static void Cleanup (void);

///\endcond

    //---------------------------------------------------------------
    //! \name Configuration
    //! \{

    /// \brief Define to which destination test events (e.g. failures) are reported
    /// \return previous EventReceiver
    IEventReceiver* SetEventReceiver (
        IEventReceiver* in_pEventReceiver);

    /// Set default values for the following settings
    void SetDefaults (void);

    /// \brief Set output mode for writing XML contents
    ///
    /// Allowed values: CTX_PROT, CTX_INFO, CTX_TEST_EVENT
    void SetOutputModeForWritingXmlContents (
        EventContext in_outputMode);

    /// Set amount of data written to test output
    void SetDetailedOutput (bool in_detailedOutput);

    /// Define whether the given attribute shall be ignored when checking XML data
    void IgnoreAttribute (
        std::string const & in_attribute,
        bool                in_ignore = true);

    /// Return whether the given attribute shall be ignored when checking XML data
    bool IsAttributeToIgnore (
        std::string const & in_attribute) const;

    /// Define whether the given node shall be ignored when checking XML data
    void IgnoreNode (
        std::string const & in_nodeName,
        bool                in_ignore = true);

    /// Return whether the given node shall be ignored when checking XML data
    bool IsNodeToIgnore (
        std::string const & in_nodeName) const;

    /// \brief Define whether the given attribute is of type double and
    /// in case of type double specify the precision to be used for
    /// checking the value.
    void UseAsDoubleAttribute (
        std::string const & in_attribute,
        bool                in_useAsDouble = true,
        int                 in_precision   = 4);

    /// Return whether the given attribute is of type double
    bool IsDoubleAttribute (
        std::string const & in_attribute) const;

    /// Return the precision used for checking of the given double attribute
    int GetPrecisionOfDoubleAttribute (
        std::string const & in_attribute) const;

    /// \brief Define whether the given node is of type double and
    /// in case of type double specify the precision to be used for
    /// checking the value.
    void UseAsDoubleNode (
        std::string const & in_nodeName,
        bool                in_useAsDouble = true,
        int                 in_precision   = 4);

    /// Return whether the given node is of type double
    bool IsDoubleNode (
        std::string const & in_nodeName) const;

    /// Return the precision used for checking of the given double node
    int GetPrecisionOfDoubleNode (
        std::string const & in_nodeName) const;

    /// \brief Define whether the given attribute is of pointer type
    void UseAsPointerAttribute (
        std::string const & in_attribute,
        bool                in_useAsPointer = true);

    /// Return whether the given attribute is of type pointer
    bool IsPointerAttribute (
        std::string const & in_attribute) const;

    /// \brief Define whether the given node is of pointer type
    void UseAsPointerNode (
        std::string const & in_nodeName,
        bool                in_useAsPointer = true);

    /// Return whether the given node is of type pointer
    bool IsPointerNode (
        std::string const & in_nodeName) const;

    /// \}

    //---------------------------------------------------------------
    /// \name Check XML contents
    /// \{


    /// \brief Writes contents of given XML node to EventReceiver.
    /// Each attribute and text is reported with a separate call
    /// to EventReceiver.
    void CheckNode (
        TiXmlNode*   in_pNode,             ///< XML (sub)tree to check
        char const * in_file         = 0,  ///< optional file name used in case of error
        long         in_lineNum      = 0   ///< optional line number used in case of error
        );

    /// \brief Writes contents of given XML node to EventReceiver.
    /// Each attribute and text is reported with a separate call
    /// to EventReceiver.
    void CheckSubNode (
        TiXmlNode*          in_pNode,                    ///< XML (sub)tree to check
        std::string const & in_xPathExpression_1,        ///< XPath expression defining a subnode
        std::string const & in_xPathExpression_2 = "",   ///< XPath expression defining a subnode of level 2
        std::string const & in_xPathExpression_3 = "",   ///< XPath expression defining a subnode of level 3
        std::string const & in_xPathExpression_4 = "",   ///< XPath expression defining a subnode of level 4
        std::string const & in_xPathExpression_5 = "",   ///< XPath expression defining a subnode of level 5
        char const *        in_file         = 0,         ///< optional file name used in case of error
        long                in_lineNum      = 0          ///< optional line number used in case of error
        );

    /// \brief Same as CheckSubNode() but writes contents only for the first in_maxChilds
    /// direct child elements of the node identified by the XPath expression(s)
    void CheckSubNodeWithMaxChilds (
        int                 in_maxChilds,                ///< max number of childs to check
        TiXmlNode*          in_pNode,                    ///< XML (sub)tree to check
        std::string const & in_xPathExpression_1,        ///< XPath expression defining a subnode
        std::string const & in_xPathExpression_2 = "",   ///< XPath expression defining a subnode of level 2
        std::string const & in_xPathExpression_3 = "",   ///< XPath expression defining a subnode of level 3
        std::string const & in_xPathExpression_4 = "",   ///< XPath expression defining a subnode of level 4
        std::string const & in_xPathExpression_5 = "",   ///< XPath expression defining a subnode of level 5
        char const *        in_file         = 0,         ///< optional file name used in case of error
        long                in_lineNum      = 0          ///< optional line number used in case of error
        );

    /// \brief Compares two XML (sub) trees. The result depends on the settings
    /// for attributes/nodes to ignore and for the used precision of double values.
    /// \return true if the XML trees are identical
    bool Compare (
        TiXmlNode*   in_pNodeA,             ///< XML (sub)tree A to check
        TiXmlNode*   in_pNodeB,             ///< XML (sub)tree B to check
        char const * in_file         = 0,  ///< optional file name used in case of error
        long         in_lineNum      = 0   ///< optional line number used in case of error
        );

    /// Get textual description of the last call to Compare
    std::string GetCompareResult (void) const;

    /// \}

    //---------------------------------------------------------------
    //! \name Helper functions
    //! \{

    /// Get the path for the given node
    static std::string GetFullPath (
        TiXmlNode*          in_pNode,
        std::string const & in_rPathToConcat = "");

    /// Get the node from the given parent node and the given path description 
    static TiXmlNode* SubNodeFromXPath (
        TiXmlNode*          in_pParentNode,
        std::string const & in_xPathExpression_1,
        std::string const & in_xPathExpression_2 = "",
        std::string const & in_xPathExpression_3 = "",
        std::string const & in_xPathExpression_4 = "",
        std::string const & in_xPathExpression_5 = "");

    ///\cond do not document

    struct StringPosition
    {
        explicit StringPosition (
            std::string::size_type in_pos = 0)
        : m_curPos  (in_pos)
        , m_lastPos (in_pos)
        , m_lineStr ()
        {}

        std::string::size_type m_curPos;
        std::string::size_type m_lastPos;
        std::string            m_lineStr;
    };

    // Extract the next substring delimited by new line
    static bool GetNextLineStr (
        std::string const & in_stringWithNewLines,
        StringPosition &    io_current);

    ///\endcond

    /// \}

protected:

    /// Interface for receiving all kinds of test events
    IEventReceiver* m_pEventReceiver;

private:

    /// Indenting to be used when writing an XML tree to test output
    static const unsigned int NUM_INDENTS_PER_SPACE=2;

    /// Output mode for writing XML contents
    EventContext m_outputMode;

    /// Flag indicating whether the output contains more details
    bool m_detailedOutput;

    /// Set of attribute names to ignore when checking an XML tree
    std::set<std::string> m_attributesToIgnore;

    /// Set of node names to ignore when checking an XML tree
    std::set<std::string> m_nodesToIgnore;

    /// Set of attribute names to be treated as doubles with specific precision
    std::map<std::string,int> m_doubleAttributes;

    /// Set of node names to be treated as doubles with specific precision
    std::map<std::string,int> m_doubleNodes;

    /// Set of attribute names to be treated as doubles with specific precision
    std::set<std::string> m_pointerAttributes;

    /// Set of node names to be treated as doubles with specific precision
    std::set<std::string> m_pointerNodes;

    /// Result of last call to Compare()
    std::string m_compareResult;

///\cond do not document

    struct TextWithPosInfo
    {
        TextWithPosInfo ()
        : m_row      (0)
        , m_col      (0)
        , m_fullPath ()
        , m_text     ()
        {}

        TextWithPosInfo (
            int                 in_row,
            int                 in_col,
            std::string const & in_fullPath,
            std::string const & in_text)
        : m_row      (in_row)
        , m_col      (in_col)
        , m_fullPath (in_fullPath)
        , m_text     (in_text)
        {}

        int         m_row;
        int         m_col;
        std::string m_fullPath;
        std::string m_text;
    };

    // Each line contains an XML entry plus additional infos for
    // better error reporting
    typedef std::vector<TextWithPosInfo> FormattedContents;

/// \endcond


private:
    /// \brief Write given XML node to EventReceiver or given optional container
    void WriteNode(
        TiXmlNode*   in_pNode,              ///< XML (sub)tree to check
        unsigned int in_indent,             ///< indentation for pretty format 
        int          in_maxChildNodes = 0,  ///<maximum number of direct childs to check, 0 : all childs
        char const * in_file         = 0,   ///< optional file name used in case of error
        long         in_lineNum      = 0,   ///< optional line number used in case of error
        FormattedContents* in_pContents = 0 ///< optional container to write output to instead
                                            ///< of using configured EventReceiver
        );

    /// \brief Writes attributes of given XML node to EventReceiver.
    ///
    /// Each attribute and text is reported with a separate call
    /// to EventReceiver.
    void WriteAttributes(
        TiXmlElement* in_pElement,        ///< XML (sub)tree to check
        unsigned int  in_indent,          ///< indentation for pretty format 
        char const *  in_file       = 0,  ///< optional file name used in case of error
        long          in_lineNum    = 0,  ///< optional line number used in case of error
        FormattedContents* in_pContents = 0    ///< optional container to write output to
        );

    /// Helper function to produce more detailed output
    std::string IfDetailed (std::string const & in_str);

    ///\cond do not document
    const char * IndentStr( unsigned int in_numIndents )
    {
    //    static const char* INDENT="                                      + ";
        static const char* INDENT="                                        ";
        static const unsigned int LENGTH=static_cast<int>(strlen(INDENT));
        unsigned int n=in_numIndents*NUM_INDENTS_PER_SPACE;
        if ( n > LENGTH ) n = LENGTH;

        return &INDENT[ LENGTH-n ];
    }

    // same as getIndent but no "+" at the end
    const char * IndentEmptyStr( unsigned int in_numIndents )
    {
        static const char* INDENT="                                        ";
        static const unsigned int LENGTH=static_cast<int>(strlen(INDENT));
        unsigned int n=in_numIndents*NUM_INDENTS_PER_SPACE;
        if ( n > LENGTH ) n = LENGTH;

        return &INDENT[ LENGTH-n ];
    }
    ///\endcond

}; // class XmlCheck

///\cond do not document
__declspec(selectany) XmlCheck* XmlCheck::s_pXmlCheck = 0;
__declspec(selectany) const unsigned int XmlCheck::NUM_INDENTS_PER_SPACE;
///\endcond

/// Access/create the singleton instance of PollingCheck
XmlCheck* TheXmlCheck (void);


//************************************************************************
//************************************************************************

} // namespace TestToolBox

///@} // end group GrpXmlCheck

