////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \defgroup GrpPerformanceDataSet Using PerformanceDataSet
/// @{
/// \ingroup GrpTestUtils 
/// \brief
/// Class PerformanceDataSet allows easy access to performance counter values.
///
/// \n This file was originally retrieved from MSDN (q194655)
/// as rkLeak.h/cpp and adjusted for test purposes.
///
/// The code assumes the existence of PDH.DLL on the system.
/// If not available define NO_PDH_DLL to suppress this part of code.
///
/// <b>Usage:</b>\n
/// \code
///
///.// Define counters to check
///.// (assuming Windows XP with German language pack, and a computer named SIRIUS)
/// std::vector<std::string> counterNames;
/// counterNames.push_back("\\\\SIRIUS\\Prozessor(0)\\Prozessorzeit (%)");
/// counterNames.push_back("\\\\SIRIUS\\Prozess(notepad)\\Threadanzahl");
/// counterNames.push_back("\\\\SIRIUS\\Prozess(TestPerformanceDataSet)\\Private Bytes");
/// 
///.// Prepare access to counter names (may fail in case of invalid names)
/// TTB::PerformanceDataSet perfDataSet;
/// perfDataSet.InitPerfCounters (m_counterNames);
///
///.// Access current counter values
/// std::vector<double> counterValues =perfDataSet.GetCurrentCounterValues();
/// \endcode
///
/// \sa PerformanceDataSet.h (File documentation)
/// \sa TestToolBox::PerformanceDataSet (class documentation)
/// \sa GrpTestPerformanceDataSet (unit test for class PerformanceDataSet)
/// 
////////////////////////////////////////////////////////////////////////////////

#pragma once

//************************************************************************
//      P r e c o m p i l e r   I n c l u d e   D i r e c t i v e s
//************************************************************************

#include <vector>
#include <pdh.h> 

// PDH.DLL may not be available for the target system
#ifndef NO_PDH_DLL

// link to library for accessing performance counters
#pragma comment(lib, "pdh.lib")

#endif

#pragma warning(disable: 4511) // copy constructor could not be generated
#pragma warning(disable: 4512) // assignment operator could not be generated

namespace TestToolBox
{

/// vector of counter names to check
typedef std::vector<std::string> PerfCounterNames;

/// vector of returned counter values
typedef std::vector<double> PerfCounterValues;

/// \brief Class wrapping access to performance counters via PDH.DLL.
/// For more info about usage see \ref GrpPerformanceDataSet.
class PerformanceDataSet
{
public:
///\cond do not document

    /// Constructor
    explicit PerformanceDataSet(void);

    /// Destructor
    virtual ~PerformanceDataSet();
///\endcond

    /// \brief Initialize all counter names and try to register them
    /// If something went wrong return false.
    bool InitPerfCounters (PerfCounterNames in_counterNames);

    /// Get the current values for all counters defined by InitPerfCounters
    PerfCounterValues GetCurrentCounterValues (void);


protected:


    /// Handle to the query for performance data
    HQUERY m_hQuery;

    /// Pointer to array of counter handles
    HCOUNTER* m_phCounter;

    /// Vector of performance counter names
    PerfCounterNames m_perfCounterFullNames;

    /// Flag indicating if the initialization has failed
    bool m_initFailed;

}; // end class PerformanceDataSet

}; // namespace TestUtil

#pragma warning(default: 4511) // copy constructor could not be generated
#pragma warning(default: 4512) // assignment operator could not be generated

///@} // end group GrpPerformanceDataSet
