////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// \brief Implementation of class PerformanceDataSet
///
////////////////////////////////////////////////////////////////////////////////


#include "PerformanceDataSet.h"
#include "CommonSources/CommonDefinitions.h"
#include "PDHMsg.h" // error codes
#include <assert.h>

///\cond do not document

namespace TestToolBox
{


//************************************************************************
//************************************************************************

// PDH.DLL may not be available for the target system
#ifdef NO_PDH_DLL

  // dummy implementation of query perfomance data methods
  // replace with real implementation if available

PDH_FUNCTION
PdhCollectQueryData (
    IN      HQUERY      hQuery)
{return NO_ERROR;}

PDH_FUNCTION
PdhCloseQuery (
    IN      HQUERY      hQuery
)
{return NO_ERROR;}


PDH_FUNCTION
PdhAddCounter (
    IN      HQUERY      hMyQuery,
    IN      LPCSTR     szFullCounterPath,
    IN      DWORD       dwUserData,
    IN      HCOUNTER    *phCounter)
{return NO_ERROR;}

PDH_FUNCTION
PdhAddCounterA (
               IN      HQUERY      hMyQuery,
               IN      LPCSTR     szFullCounterPath,
               IN      DWORD       dwUserData,
               IN      HCOUNTER    *phCounter)
{return NO_ERROR;}

PDH_FUNCTION
PdhOpenQuery (
    IN      LPCSTR     szDataSource,
    IN      DWORD       dwUserData,
    IN      HQUERY      *phQuery
)
{return NO_ERROR;}

PDH_FUNCTION
PdhGetFormattedCounterValue (
    IN      HCOUNTER                hCounter,
    IN      DWORD                   dwFormat,
    IN      LPDWORD                 lpdwType,
    IN      PPDH_FMT_COUNTERVALUE   pValue
)
{return NO_ERROR;}


#endif

//************************************************************************
//************************************************************************


inline PerformanceDataSet::PerformanceDataSet(void)
    :
    m_initFailed (false)
{
} // PerformanceDataSet::PerformanceDataSet


//************************************************************************
//************************************************************************


inline PerformanceDataSet::~PerformanceDataSet()
{
    TRCF("PerformanceDataSet::~PerformanceDataSet");
    TRC(TL_DEBUG, "Destructor ");

#ifndef NO_PDH_DLL
    PdhCloseQuery (m_hQuery); GlobalFree(m_phCounter);
#endif

} // PerformanceDataSet::~PerformanceDataSet


//************************************************************************
//************************************************************************


inline bool PerformanceDataSet::InitPerfCounters (PerfCounterNames in_counterNames)
{
    TRCF("PerfCounterBase::InitPerfCounters");
    TRC(TL_DEBUG, "Begin");

    m_initFailed = false;
    m_perfCounterFullNames = in_counterNames;

    // -----  Init query of performance data  -----

    PDH_STATUS pdhStatus = PdhOpenQuery (0, 0, &m_hQuery); 
    assert(pdhStatus == ERROR_SUCCESS);

    // Set default counters
    size_t numCounters = m_perfCounterFullNames.size();
    m_phCounter = (HCOUNTER *)GlobalAlloc(GPTR, (sizeof(HCOUNTER) * numCounters)); 
    assert (m_phCounter ); 
    for (unsigned int n = 0; n < numCounters; n++)
    { 
        TRC(TL_DEBUG, "Counter %d: >>%s<<", n, m_perfCounterFullNames[n].c_str());

        pdhStatus = PdhAddCounterA (m_hQuery,  m_perfCounterFullNames[n].c_str(), 0, &m_phCounter[n]);
        if (ERROR_SUCCESS != pdhStatus)
        {
            m_initFailed = true;
            TRC(TL_ERROR, "Add performance counter failed for %s", m_perfCounterFullNames[n].c_str());
        }
    } 
    if (m_initFailed)
    {
        PdhCloseQuery (m_hQuery); GlobalFree(m_phCounter);
    }
    else
    {
        pdhStatus = PdhCollectQueryData (m_hQuery);
        if (ERROR_SUCCESS != pdhStatus)
        {
            m_initFailed = true;
            TRC(TL_ERROR, "Initialization of performance counters failed! Check your specific PC with utility ScanPerfCounters");
            PdhCloseQuery (m_hQuery); GlobalFree(m_phCounter);
        }
    }
    return !m_initFailed;

} // PerformanceDataSet::InitPerfCounters


//************************************************************************
//************************************************************************


inline PerfCounterValues PerformanceDataSet::GetCurrentCounterValues (void)
{
    TRCF("PerformanceDataSet::GetCurrentCounterValues");

    PerfCounterValues values;

    PDH_STATUS pdhStatus = PdhCollectQueryData (m_hQuery);
    assert (pdhStatus == ERROR_SUCCESS);

    unsigned int n = 0; 
    HCOUNTER*            pCurHand; 
    DWORD                ctrType; 
    PDH_FMT_COUNTERVALUE fmtValue; 

    pCurHand = &m_phCounter[0];  

    while (n < m_perfCounterFullNames.size())
    { 
        assert(*pCurHand) ; // get the current value for this counter 
        pdhStatus = PdhGetFormattedCounterValue ( 
            *pCurHand,
            PDH_FMT_DOUBLE,
            &ctrType, 
            &fmtValue);
        if (pdhStatus == ERROR_SUCCESS)
        {
            values.push_back(fmtValue.doubleValue);
        }
        else
        {
            values.push_back(-1); // error retrieving value
            TRC(TL_ERROR, "Error retrieving value for %s", m_perfCounterFullNames[n].c_str());
        }
        pCurHand++; 
        n++; 
    } 

    return values;

} // PerformanceDataSet::GetCurrentCounterValues


//************************************************************************
//************************************************************************


#ifndef NO_PDH_DLL

// VC6 hack
#undef PdhOpenQuery      //          PdhOpenQueryA
extern "C" long __stdcall 
PdhOpenQuery (
    IN      LPCSTR      szDataSource,
    IN      DWORD       dwUserData,
    IN      HQUERY      *phQuery
);

#endif


//************************************************************************
//************************************************************************


}; // namespace TestToolBox

///\endcond

// end of file PerformanceDataSetImpl.h
