////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \defgroup GrpPollingCheck Using PollingCheck
/// @{
/// \ingroup GrpTestUtils 
/// \brief
/// Class PollingCheck allows to repeatedly execute a funtional check
/// for some condition until it becomes true or a timeout has elapsed.
///
/// In the context of testing this function should only be used when the
/// following conditions apply:
///   - the condition represents the reaction of an test object
///   - the reaction occures asynchronously after some trigger action
///   - there is no other way to detect the moment of reaction (e.g.
///     by receiving a notification about some state change)
///
/// The check of the condition is implemented within a boolean user function.
/// The user function can optionally make use of any other test features
/// (e.g. writing/ checking TestEvents).
///
/// <b>\n Concept for PollingCheck::Run</b>\n
/// A user definable function/functor is repeatedly called until
/// the return value is true. If the user function returns false
/// the next call is executed after a definable sleep time.
/// The polling stops as soon as the user function returns true or a definable
/// timeout value has been reached.
/// The return value of Run() indicates whether the condition was fulfilled.
/// The return value can be checked with one of the following methods:
///   - if
///   - assert
///   - BOOST_CHECK
///   - convenience macro POLL_VERIFY (requires usage of some EventReceiver class
///     e.g. TTB::EventReceiverStdOut)
///
/// <b>\n Elementar usage with if, assert, bind, BOOST_CHECK</b>\n
/// \code
///
/// bool VerifyCounterIsZero ()
/// {return s_pSomeTestObject->GetCounter() == 0;}
///
/// bool VerifyCounterHasValue (
///     int             in_value)
/// {return s_pSomeTestObject->GetCounter() == in_value;}
///
/// ... 
///
/// // Configure timeouts used for polling
/// TTB::ThePollingCheck()->SetSleepTimeMs(100);
/// TTB::ThePollingCheck()->SetTimeoutMs(500);
///
/// // check with if
/// if (!TTB::ThePollingCheck()->Run(&VerifyCounterIsZero))
/// {
///     cout << "VerifyCounterIsZero failed!" << endl;
/// }
///
/// // check with assert
/// assert(TTB::ThePollingCheck()->Run(&VerifyCounterIsZero));
///
/// // Check with Boost verification macro
/// BOOST_CHECK(TTB::ThePollingCheck()->Run(&VerifyCounterIsZero));
///
/// // using bind to make use of any number of function arguments
/// assert(TTB::ThePollingCheck()->Run(boost::bind(&VerifyCounterHasValue,98)));
///
/// // directly using the member function to check
/// assert(TTB::ThePollingCheck()->Run(boost::bind(
///     &SomeTestObject::GetCounter,
///     s_pSomeTestObject)==94));
/// \endcode
///
/// <b>\n Convenience macro POLL_VERIFY</b>\n
/// \code
/// // define an EventReceiver to process events detected from the execution and
/// // compare mechanism (e.g. use TTB::EventReceiverStdOut)
/// TTB::EventReceiverStdOut eventReceiver;
/// TTB::ThePollingCheck()->SetEventReceiver(&eventReceiver);
/// // optional: configure amount of infos written to stdout
/// // eventReceiver.SetOutputFilter(TTB::CTX_ONLY_ERROR);
///
/// // Macro for use with free functions (other variants support also member functions)
/// POLL_VERIFY_F1(VerifyCounterHasValue, 96);
/// \endcode
/// In case of error the following output is produced:
/// \code
/// <VerifyCounterHasValue 96 [polling]
/// INFO: Poll-Timeout after 515 ms
/// c:\userdata\gerald\sw\c\testtoolbox\test\testpollingcheck\testpollingcheck.cpp (184): POLL-ERROR: verification failed
/// \endcode
///
/// \sa PollingCheck.h (File documentation)
/// \sa TestToolBox::PollingCheck (class documentation)
/// \sa GrpTestPollingCheck (unit test for class Environment)
/// 
////////////////////////////////////////////////////////////////////////////////

#pragma once

#include <boost/function.hpp>

//************************************************************************

namespace TestToolBox
{
    struct IEventReceiver;

    //! \brief Defines a boolean functor without arguments
    typedef boost::function<bool()> BoolFunctor;


//************************************************************************
//************************************************************************


/// Convenience macros (output of called function and possible error results via EventReceiver)

/// Using memberfunction
#define POLL_VERIFY(Func) \
{TTB::ThePollingCheck()->Run(boost::bind(&Func,this),#Func,"", __FILE__, __LINE__);}

/// Using member function with single argument
#define POLL_VERIFY1(Func,arg) \
{TTB::ThePollingCheck()->Run(boost::bind(&Func,this,arg),#Func,#arg, __FILE__, __LINE__);}

/// Using free function
#define POLL_VERIFY_F(Func) \
{TTB::ThePollingCheck()->Run(boost::bind(&Func),#Func,"", __FILE__, __LINE__);}

/// Using free function  with single argument
#define POLL_VERIFY_F1(Func,arg) \
{TTB::ThePollingCheck()->Run(boost::bind(&Func,arg),#Func,#arg, __FILE__, __LINE__);}



//************************************************************************
//************************************************************************


/// \brief Repeated polling until some condition comes true.
/// For more info about usage see \ref GrpPollingCheck.
///
/// \nosubgrouping 
/// \sa PollingCheck.h (File documentation)
/// \sa GrpTestPollingCheck (unit test for class Environment)
/// \sa GrpPollingCheck (Usage documentation)
class PollingCheck
{
public:

///\cond do not document

    // Constructor
	PollingCheck();

    // Destructor
	virtual ~PollingCheck();

    // Get the pointer to the only instance of this class
    static PollingCheck* Get (void)
    {if (!s_pPollingCheck) s_pPollingCheck = new PollingCheck;
     return s_pPollingCheck;}
    static PollingCheck* s_pPollingCheck;
    static void Cleanup (void)
    {delete s_pPollingCheck; s_pPollingCheck = 0;}
///\endcond

    //---------------------------------------------------------------
    //! \name Configuration
    //! \{

    /// Set initial sleep time before making the first call
    void SetInitialSleepTimeMs (long in_timeMs)
    {m_initialSleepTimeMs = in_timeMs;}

    /// Set delay time between the repeated polling calls
    void SetSleepTimeMs (long in_timeMs)
    {m_sleepTimeMs = in_timeMs;}

    /// When this amount of time has elapsed the polling will be stopped with failure
    void SetTimeoutMs (long in_timeMs)
    {m_timeoutMs = in_timeMs;}

    /// Define to which destination test events (e.g. failures) are reported
    void SetEventReceiver (IEventReceiver* in_pEventReceiver)
    {m_pEventReceiver = in_pEventReceiver;}

    /// \}

    //---------------------------------------------------------------
    /// \name Execute polling
    /// \{


    /// \brief Execute the check for some condition by calling repeatedly the
    /// functor until true is returned or the configured timeout has elapsed.
    ///
    /// \return false if condition was not fulfilled and timeout has elapsed
    bool Run(
        BoolFunctor  in_functor,           ///< function checking condition
        char const * in_functionName = 0,  ///< optional function name for protocol
        char const * in_argValue     = 0,  ///< optional argument value for protocol
        char const * in_file         = 0,  ///< optional file name used in case of error
        long         in_lineNum      = 0 ///< optional line number used in case of error
        );

    /// \}


protected:

    /// Initial sleep time before making the first call
    long m_initialSleepTimeMs;

    /// Delay time between the repeated calls
    long m_sleepTimeMs;

    /// \brief When this amount of time has elapsed the polling will
    /// be stopped with failure
    long m_timeoutMs;

    /// Interface for receiving all kind of test events
    IEventReceiver* m_pEventReceiver;


}; // class PollingCheck

///\cond do not document
__declspec(selectany) PollingCheck* PollingCheck::s_pPollingCheck = 0;
///\endcond

/// Access/create the singleton instance of PollingCheck
inline PollingCheck* ThePollingCheck (void)
{return PollingCheck::Get();}


//************************************************************************
//************************************************************************

} // namespace TestToolBox

///@} // end group GrpPollingCheck

