////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \brief
/// Implementation of class PollingCheck
/// 
////////////////////////////////////////////////////////////////////////////////

#pragma once

#include "IEventReceiver.h"
#include "PollingCheck.h"
#include "CommonSources\CommonDefinitions.h"

#include <iostream>

///\cond do not document


//*****************************************************************************
//*****************************************************************************


namespace TestToolBox
{

inline PollingCheck::PollingCheck()
    : m_initialSleepTimeMs (10)
    , m_sleepTimeMs        (10)
    , m_timeoutMs          (1000)
    , m_pEventReceiver     (0)
{
    TRCF("PollingCheck::PollingCheck");
    TRC(TL_DEBUG, "Constructor");

} // PollingCheck::PollingCheck


//************************************************************************
//************************************************************************


inline PollingCheck::~PollingCheck()
{
    TRCF("PollingCheck::~PollingCheck");
    TRC(TL_DEBUG, "Destructor");

} // PollingCheck::~PollingCheck


//************************************************************************
//************************************************************************


inline bool PollingCheck::Run (
    BoolFunctor  in_functor,
    char const * in_functionName,
    char const * in_argValue,
    char const * in_file,
    long         in_lineNum)
{
    bool retVal = false;

    TRCF("PollingCheck::Run");
    TRC(TL_DEBUG, "Begin");

    // Write info about start of polling to protocol
    if (m_pEventReceiver && in_functionName)
    {
        // display only method name, suppress class name if given
        char const * pFunctionNameWithoutClass = strrchr(in_functionName,':');
        pFunctionNameWithoutClass = pFunctionNameWithoutClass 
            ? pFunctionNameWithoutClass + 1 : in_functionName;

        std::ostringstream oss;
        oss << "\n<" << pFunctionNameWithoutClass;
        if (in_argValue) oss << " " << in_argValue;
        oss << " [polling]";
        m_pEventReceiver->EventMsg(CTX_PROT, oss.str().c_str());
    }

    // Store start time
    unsigned long startTimeMs = ::GetTickCount();

    // Optionally wait some time before making the first check
    if (m_initialSleepTimeMs >= 0)
    {
        ::Sleep (m_initialSleepTimeMs);
    }

    bool timeout = false;
    do
    {
        // call the user defined check function
        if (in_functor())
        {
            // condition is fulfilled, stop polling
            retVal = true;
            break;
        }

        // wait some time before repeating the check
        ::Sleep(m_sleepTimeMs);

        timeout = (::GetTickCount() - startTimeMs >= 
            static_cast<unsigned long>(m_timeoutMs));

    } while (!timeout);

    // generating error messages
    if (timeout) // condition was not fulfilled within the timeout period
    {
        if (m_pEventReceiver)
        {
            m_pEventReceiver->Event(CTX_INFO,"Poll-Timeout after %d ms",
                ::GetTickCount() - startTimeMs);
        }
    }
    if (!retVal)
    {
        std::ostringstream oss;
        if (in_file)
            oss << in_file << " (" << in_lineNum << "): ";
        oss << "POLL-ERROR: verification failed";
        if (m_pEventReceiver)
            m_pEventReceiver->EventMsg(CTX_ERROR, oss.str().c_str());
    }

    TRC(TL_DEBUG, "End: retVal=%d", retVal);
    return retVal;

} // PollingCheck::Run


//************************************************************************
//************************************************************************

///\endcond

} // namespace TestToolBox


