////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz (extracted from SDK documentation)
/// 
/// \defgroup GrpScanPerfCounters ScanPerfCounters
/// @{
/// \ingroup GrpHelperApps 
/// \brief
/// Win32 console application "ScanPerfCounters.exe" retrieves a list
/// of performance counters available on your windows system. A single
/// counter instance can be selected via dialog to test the periodic retrieving
/// of performance data.
/// 
/// \n
/// The names of performance counters depend both on the windows
/// version and the language pack installed. The helper application
/// ScanPerfCounters provides a list of valid counter names which may be
/// used within some other program to access specific performance information.
///
/// <b>Syntax:</b>\n
/// \code
/// ScanPerfCounters [ALL_INSTANCES]
/// \endcode
///
/// When started the program executes the following three steps:
///
/// - the counter names are searched within registry and all counter instance
///   names are printed to stdout. When option ALL_INSTANCES is given
///   all instances of a counter object are listed otherwise (default) only
///   the first counter instance is listed.
/// - the standard PDH browse dialog is opened; in this dialog an arbitrary
///   performance counter can be selected
/// - the actual counter value of the selected counter is periodically printed
///   to stdout, program can be stopped with any key click
///
/// 
/// <b>Hint:</b>\n\n
/// Redirect the output to some logfile to get a storable list of performance
/// counter names.
///
/// \sa ScanPerfCounters.cpp
///
////////////////////////////////////////////////////////////////////////////////


///\cond do not document

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <tchar.h>

#include <conio.h>
#include <tchar.h>
#include <pdh.h>

#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <string>

#pragma comment(lib,"pdh.lib")

#define SAMPLE_INTERVAL_MS  1000
#define MAXPATH 80


//************************************************************************
//************************************************************************

#define TOTALBYTES    8192
#define BYTEINCREMENT 1024*1024


LPSTR lpNameStrings;
LPSTR *lpNamesArray;

/*****************************************************************
 *                                                               *
 * Functions used to navigate through the performance data.      *
 *                                                               *
 *****************************************************************/

PPERF_OBJECT_TYPE FirstObject( PPERF_DATA_BLOCK PerfData )
{
    return( (PPERF_OBJECT_TYPE)((PBYTE)PerfData + 
        PerfData->HeaderLength) );
}

PPERF_OBJECT_TYPE NextObject( PPERF_OBJECT_TYPE PerfObj )
{
    return( (PPERF_OBJECT_TYPE)((PBYTE)PerfObj + 
        PerfObj->TotalByteLength) );
}

PPERF_INSTANCE_DEFINITION FirstInstance( PPERF_OBJECT_TYPE PerfObj )
{
    return( (PPERF_INSTANCE_DEFINITION)((PBYTE)PerfObj + 
        PerfObj->DefinitionLength) );
}

PPERF_INSTANCE_DEFINITION NextInstance( 
    PPERF_INSTANCE_DEFINITION PerfInst )
{
    PPERF_COUNTER_BLOCK PerfCntrBlk;

    PerfCntrBlk = (PPERF_COUNTER_BLOCK)((PBYTE)PerfInst + 
        PerfInst->ByteLength);

    return( (PPERF_INSTANCE_DEFINITION)((PBYTE)PerfCntrBlk + 
        PerfCntrBlk->ByteLength) );
}

PPERF_COUNTER_DEFINITION FirstCounter( PPERF_OBJECT_TYPE PerfObj )
{
    return( (PPERF_COUNTER_DEFINITION) ((PBYTE)PerfObj + 
        PerfObj->HeaderLength) );
}

PPERF_COUNTER_DEFINITION NextCounter( 
    PPERF_COUNTER_DEFINITION PerfCntr )
{
    return( (PPERF_COUNTER_DEFINITION)((PBYTE)PerfCntr + 
        PerfCntr->ByteLength) );
}

/*****************************************************************
 *                                                               *
 * Load the counter and object names from the registry to the    *
 * global variable lpNamesArray.                                 *
 *                                                               *
 *****************************************************************/

BOOL GetNameStrings( )
{
    HKEY hKeyPerflib;      // handle to registry key
    HKEY hKeyPerflib009;   // handle to registry key
    DWORD dwMaxValueLen;   // maximum size of key values
    DWORD dwBuffer;        // bytes to allocate for buffers
    DWORD dwBufferSize = sizeof(DWORD);  // size of dwBuffer
    LPSTR lpCurrentString; // pointer for enumerating data strings
    DWORD dwCounter;       // current counter index
    LONG lResult;          // return value


    // Get the number of Counter items.
    if( RegOpenKeyEx( HKEY_LOCAL_MACHINE,
        "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Perflib",
        0,
        KEY_READ,
        &hKeyPerflib) != ERROR_SUCCESS)
        return FALSE;

    lResult = RegQueryValueEx( hKeyPerflib,
        "Last Counter",
        NULL,
        NULL,
        (LPBYTE) &dwBuffer,
        &dwBufferSize );

    RegCloseKey( hKeyPerflib );

    if( lResult != ERROR_SUCCESS )
        return FALSE;

    // Allocate memory for the names array.
    lpNamesArray = (char**)malloc( (dwBuffer+1) * sizeof(LPSTR) );

    if( lpNamesArray == NULL )
        return FALSE;

    // Open the key containing the counter and object names.
    if( RegOpenKeyEx( HKEY_LOCAL_MACHINE,
        "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Perflib\\009",
        0,
        KEY_READ,
        &hKeyPerflib009) != ERROR_SUCCESS
      )
        return FALSE;

    // Get the size of the largest value in the key (Counter or Help).
    if( RegQueryInfoKey( hKeyPerflib009,
        NULL,
        NULL,
        NULL,
        NULL,
        NULL,
        NULL,
        NULL,
        NULL,
        &dwMaxValueLen,
        NULL,
        NULL) != ERROR_SUCCESS
      )
        return FALSE;

    // Allocate memory for the counter and object names.
    dwBuffer = dwMaxValueLen + 1;

    lpNameStrings = (char*)malloc( dwBuffer * sizeof(CHAR) );

    if (lpNameStrings == NULL)
    {
        free( lpNamesArray );
        return FALSE;
    }

    // Read the counter value.
    lResult = RegQueryValueEx( hKeyPerflib009,
        "Counters",
        NULL,
        NULL,
        (unsigned char*)lpNameStrings, &dwBuffer );

    RegCloseKey( hKeyPerflib009 ); 

    if( lResult != ERROR_SUCCESS )
        return FALSE;

    // Load names into an array, by index.
    for( lpCurrentString = lpNameStrings; *lpCurrentString;
         lpCurrentString += (lstrlen(lpCurrentString)+1) )
    {
        dwCounter = atol( lpCurrentString );

        lpCurrentString += (lstrlen(lpCurrentString)+1);

        lpNamesArray[dwCounter] = (LPSTR) lpCurrentString;
    }

   return TRUE;
}

/*****************************************************************
 *                                                               *
 * Display the indexes and/or names for all performance objects, *
 * instances, and counters.                                      *
 *                                                               *
 *****************************************************************/

int ListCounters (bool in_allInstances)
{
    PPERF_DATA_BLOCK PerfData = NULL;
    PPERF_OBJECT_TYPE PerfObj;
    PPERF_INSTANCE_DEFINITION PerfInst;
    PPERF_COUNTER_DEFINITION PerfCntr, CurCntr;
    PPERF_COUNTER_BLOCK PtrToCntr;
    DWORD BufferSize = TOTALBYTES;
    DWORD i, j, k;

    printf("Requesting performance counters from registry...\n");

    // Get the name strings through the registry.
    if( !GetNameStrings( ) )
       return FALSE;

    printf("accessing counters...\n");

    // Allocate the buffer for the performance data.
    PerfData = (PPERF_DATA_BLOCK) malloc( BufferSize );

    if( PerfData == NULL )
       return FALSE;

    while( RegQueryValueEx( HKEY_PERFORMANCE_DATA,
                               "Global",
                               NULL,
                               NULL,
                               (LPBYTE) PerfData,
                               &BufferSize ) == ERROR_MORE_DATA )
    {
        // Get a buffer that is big enough.
        BufferSize += BYTEINCREMENT;
        PerfData = (PPERF_DATA_BLOCK) realloc( PerfData, BufferSize );
    }

    // Get the first object type.
    PerfObj = FirstObject( PerfData );

    // Process all objects.
    printf("\nAllocated bufferSize          : %ld\n",  BufferSize);
    printf("Number of counter objects     : %ld\n", PerfData->NumObjectTypes);
    printf("Instances per object to check : %s\n", 
        in_allInstances ? "all" : "only the first");

    for( i=0; i < PerfData->NumObjectTypes; i++ )
    {
        // Display the object by index and name.
        printf( "\nObject %ld: %s\n", PerfObj->ObjectNameTitleIndex,
            lpNamesArray[PerfObj->ObjectNameTitleIndex] );

        // Get the first counter.
        PerfCntr = FirstCounter( PerfObj );

        if( PerfObj->NumInstances > 0 )
        {
            // Get the first instance.
            PerfInst = FirstInstance( PerfObj );

            // Retrieve instances.
            long numInstancesToList = in_allInstances ? PerfObj->NumInstances : 1;
            for( k=0; k < (DWORD) numInstancesToList; k++ )
            {
                // Display the instance by name.
                printf( "\n\tInstance %S: \n", 
                    (char *)((PBYTE)PerfInst + PerfInst->NameOffset));
                CurCntr = PerfCntr;

                // Retrieve all counters of the instance.
                for( j=0; j < PerfObj->NumCounters; j++ )
                {
                    // Display the counter by index and name.
                    printf("\t\tCounter %ld: %s\n", 
                        CurCntr->CounterNameTitleIndex,
                        lpNamesArray[CurCntr->CounterNameTitleIndex]);

                    // Get the next counter.
                    CurCntr = NextCounter( CurCntr );
                }

                // Get the next instance.
                PerfInst = NextInstance( PerfInst );
            }
        }
        else
        {
            // Get the counter block.
            PtrToCntr = (PPERF_COUNTER_BLOCK) ((PBYTE)PerfObj +
                            PerfObj->DefinitionLength );

            // Retrieve all counters.
            for( j=0; j < PerfObj->NumCounters; j++ )
            {
                // Display the counter by index and name.
                printf( "\tCounter %ld: %s\n", PerfCntr->CounterNameTitleIndex,
                    lpNamesArray[PerfCntr->CounterNameTitleIndex] );

                // Get the next counter.
                PerfCntr = NextCounter( PerfCntr );
            }
        }

        // Get the next object type.
        PerfObj = NextObject( PerfObj );
    }

    free( lpNamesArray );
    free( lpNameStrings );
    free( PerfData );

    return TRUE;

} // ListCounters


//************************************************************************
//************************************************************************


int BrowseAndCheckSingleCounter(void) 
{
   HQUERY          hQuery;
   HCOUNTER        *pCounterHandle;
   PDH_STATUS      pdhStatus;
   PDH_FMT_COUNTERVALUE   fmtValue;
   DWORD           ctrType;
   SYSTEMTIME      stSampleTime;
   PDH_BROWSE_DLG_CONFIG  BrowseDlgData;
   CHAR            szPathBuffer[MAXPATH];
   int             nRetCode = 0;

   // Open the query object.
   pdhStatus = PdhOpenQuery (0, 0, &hQuery);

   // Allocate the counter handle array. Allocate room for
   //  one handle per command line arg, not including the
   //  executable file name.
   pCounterHandle = (HCOUNTER *)GlobalAlloc(GPTR, sizeof(HCOUNTER));
   // Zero the contents of the structure.
   memset (&BrowseDlgData, 0, sizeof(PDH_BROWSE_DLG_CONFIG));

   // Initialize the browser dialog window settings.
   BrowseDlgData.bIncludeInstanceIndex = FALSE;   
   BrowseDlgData.bSingleCounterPerAdd = TRUE;
   BrowseDlgData.bSingleCounterPerDialog = TRUE;  
   BrowseDlgData.bLocalCountersOnly = FALSE;      
   BrowseDlgData.bWildCardInstances = TRUE;
   BrowseDlgData.bHideDetailBox = TRUE;
   BrowseDlgData.bInitializePath = FALSE;     
   BrowseDlgData.bDisableMachineSelection = FALSE;
   BrowseDlgData.bIncludeCostlyObjects = FALSE;
   BrowseDlgData.bReserved = TRUE;
   BrowseDlgData.hWndOwner = NULL;   
   BrowseDlgData.bReserved = FALSE;           
   BrowseDlgData.szReturnPathBuffer = szPathBuffer;
   BrowseDlgData.cchReturnPathLength = MAXPATH;;
   BrowseDlgData.pCallBack = NULL;   
   BrowseDlgData.dwCallBackArg = 0;
   BrowseDlgData.CallBackStatus = ERROR_SUCCESS;
   BrowseDlgData.dwDefaultDetailLevel = PERF_DETAIL_WIZARD;
   BrowseDlgData.szDialogBoxCaption = "Select a counter to monitor.";

   // Display the counter browser window to select a single counter.
   printf("Opening browser window...\n");
   pdhStatus = PdhBrowseCounters (&BrowseDlgData);
   printf("Selected counter = %s\n",  szPathBuffer);

   pdhStatus = PdhAddCounter (hQuery,
                              szPathBuffer, 
                              0, 
                              pCounterHandle);

   // "Prime" counters that need two values to display a 
   // formatted value.
   pdhStatus = PdhCollectQueryData (hQuery);

   printf("\nRetrieving counter value every %d millisec (Stop with any <key>)\n", SAMPLE_INTERVAL_MS);

   // Print counter values until a key is pressed.
   while (!_kbhit()) {

     // Wait one interval.
     Sleep(SAMPLE_INTERVAL_MS);

     // Get the sample time.
     GetLocalTime (&stSampleTime);

     // Get the current data values.
     pdhStatus = PdhCollectQueryData (hQuery);

     // Print the time stamp for the sample.
     _tprintf (
           TEXT("\n\"%2.2d/%2.2d/%4.4d %2.2d:%2.2d:%2.2d.%3.3d\""),
               stSampleTime.wMonth, 
               stSampleTime.wDay, 
               stSampleTime.wYear,
               stSampleTime.wHour, 
               stSampleTime.wMinute, 
               stSampleTime.wSecond,
               stSampleTime.wMilliseconds);

     // Get the current value of this counter.
     pdhStatus = PdhGetFormattedCounterValue (*pCounterHandle,
                                              PDH_FMT_DOUBLE,
                                              &ctrType,
                                              &fmtValue);

     if (pdhStatus == ERROR_SUCCESS)
     {
         _tprintf (TEXT(",\"%.20g\""), fmtValue.doubleValue);
     }
     else
     {

         // Print the error value.
         _tprintf (TEXT(".\"-1\"")); 
     }
   }

   // Close the query.
   pdhStatus = PdhCloseQuery (hQuery);

   return nRetCode;
}


//************************************************************************
//************************************************************************

int _tmain(int argc, _TCHAR* argv[])
{
    bool listAllCounterInstances (false); // if false list only the first instance

    // Check command line
    if (argc > 1)
    {
        std::string argStr = argv[1];
        listAllCounterInstances = (argStr == "ALL_INSTANCES");
    }

    // List all available counter instances to stdout
    ListCounters(listAllCounterInstances);

    // Select a single counter instance via standard dialog and watch it
    // The name of the counter is written to standard out. This name can
    // be used within TestServerDll\PerformanceData.cpp.
    BrowseAndCheckSingleCounter();

    return 0;
}


//************************************************************************
//************************************************************************


///\endcond do not document

///@} // end group GrpScanPerfCounters
