///////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2011 ASM Assembly Systems GmbH & Co. KG
////////////////////////////////////////////////////////////////////////////////
//
// Description : Definition of class SimulatedObjectBase
//
////////////////////////////////////////////////////////////////////////////////

#pragma once

#include <functional>
#include <map>
#include "CommonSources/CommonDefinitions.h"
#include "WaitableCondition.h"
#include <boost/shared_ptr.hpp>

#ifdef USE_TTB_INSTANCE_DLL
#    include "TTB_InstanceDll.h"
#endif

//-------------------------------------------------------------------------
/// Macros to be used from a class derived from SimulatedObjectBase
//-------------------------------------------------------------------------

#define TTB_METHOD_CALL() _MethodCall(__FUNCTION__)
#define TTB_METHOD_CALL1(stringStreamExpression) _MethodCall(__FUNCTION__,TTB_STRING_S(stringStreamExpression))
#define TTB_METHOD_CALL2(stringStreamExpression,stringFuncBody) _MethodCall(__FUNCTION__,TTB_STRING_S(stringStreamExpression), ([&](void) -> std::string{stringFuncBody}))

#define TTB_METHOD_CALL_RET1(ReturnType) _MethodCallReturn<ReturnType>(__FUNCTION__);
#define TTB_METHOD_CALL_RET2(ReturnType,stringStreamExpression) _MethodCallReturn<ReturnType>(__FUNCTION__,TTB_STRING_S(stringStreamExpression));

//-------------------------------------------------------------------------

namespace TestToolBox
{

    typedef boost::shared_ptr<WaitableCondition> WaitableConditionSp;
    typedef std::map<std::string, WaitableConditionSp> WaitableConditions;

    class RegisteredSimulatedObject;

    //-------------------------------------------------------------------------
    // Controlling options
    //-------------------------------------------------------------------------

    namespace OptionType
    {
        enum Enum {
            NOT_ACTIVE,
            PERMANENTLY_ACTIVE,
            ACTIVE_ONCE,
            AFTER_N_CALLS_PERMANENTLY_ACTIVE,
            AFTER_N_CALLS_ACTIVE_ONCE,
            VALUE};
    }

    //-------------------------------------------------------------------------

    struct Option
    {
        /// commonly used options
        static std::string const oSILENT;
        static std::string const oERROR;
        static std::string const oBLOCKING;
        static std::string const oSYNC;
        static std::string const oOBJECT_NAME; // use object name when tracking method calls?

        /// scope to address all methods
        static std::string const ALL;

        Option()
            : m_optionType (OptionType::PERMANENTLY_ACTIVE)
            , m_numRemainingCalls (0)
        {
        }

        /// Check if option is activated
        /// Remark: The call to this function may change the internal
        /// state of the option.
        bool IsSet (void);

        std::string GetOptionTypeAsString() const;

        OptionType::Enum m_optionType;
        int              m_numRemainingCalls;
        std::string      m_valueStr;
    };

    //-------------------------------------------------------------------------

    /// Map of options
    /// Key is a string of the following structure: option identifier + scope identifier,
    /// Examples: "error:all", "silent:SomeFunction", sync:SomeClass::SomeMethod
    typedef std::map<std::string, Option> OptionData;

    //-------------------------------------------------------------------------

    /// Class storing any number of textual options each assignable for an
    /// arbitrary scope.
    ///
    /// Remark: Checking and changing of options may be done also
    /// within const member functions of an simulated object.
    /// To make this easy and to avoid ugly const casting operations
    /// all member functions are const and all data members are mutable.
    class Options
    {
    public:

        /// ---  Setting of options  ---

        void Set (
            std::string const & in_option,
            std::string const & in_scope             = Option::ALL,
            OptionType::Enum    in_optionType        = OptionType::PERMANENTLY_ACTIVE,
            int                 in_numRemainingCalls = 0) const;
        void SetValue(
            std::string const & in_name,
            std::string const & in_value,
            std::string const & in_scope = Option::ALL) const;
        void Reset (
            std::string const & in_option,
            std::string const & in_scope = Option::ALL) const;
        void ResetAll (void) const;

        /// ---  Checking of options  ---

        Option* FindOptionExact (
            std::string const & in_option,
            std::string const & in_scope = Option::ALL) const;

        /// If there is no option for the given scope then an existing option
        /// for scope Option::ALL is used instead
        Option* FindOption (
            std::string const & in_option,
            std::string const & in_scope = Option::ALL) const;

        /// Check whether the given option is set.
        ///
        /// If there is no option for the given scope then an existing option
        /// for scope Option::ALL is used instead.
        /// Remark: The call to this function may change the internal
        /// state of the optionto allow automatic control
        /// of temporarily active option typess like AFTER_N_CALLS_ACTIVE_ONCE
        bool IsSet (
            std::string const & in_option,
            std::string const & in_scope = Option::ALL) const;

        /// ---  Easy access to commonly used options  ---

        bool IsSilent (std::string const & in_scope = Option::ALL) const;
        bool IsError(std::string const & in_scope = Option::ALL) const;

        /// ---  Save and restore of options  ---

        void Dump(EventContext in_context = CTX_INFO) const;
        OptionData Save (void) const;
        void Restore (void) const;
        void Restore (OptionData const & in_optionData) const;

        OptionData GetOptionData(void) const;
        void SetOptionData(OptionData const & in_optionData) const;

    private:
        OptionData mutable m_optionData;
        OptionData mutable m_savedOptionData;

    }; // end class Options

    //-------------------------------------------------------------------------

    /// Singleton class to store options which are globally available within
    /// an test application.
    class GlobalOptions : public Options
    {
    private:

        /// \brief Constructor.
        /// Access/create instance via static method Get()
        explicit GlobalOptions (void);

    public:

        USE_AS_SINGLETON(GlobalOptions);

        /// Destructor
        ~GlobalOptions (void);

    }; // end class GlobalOptions

    DEFINE_SINGLETON_POINTER (GlobalOptions);


    //-------------------------------------------------------------------------

    /// Singleton class to store all instances of simulated objects derived from
    /// RegisteredSimulatedObject
    class RegisteredSimulatedObjects
    {
    private:

        /// \brief Constructor.
        /// Access/create instance via static method Get()
        explicit RegisteredSimulatedObjects (void);

    public:

        USE_AS_SINGLETON(RegisteredSimulatedObjects);

        typedef std::map<std::string, RegisteredSimulatedObject*> MapName2This;

        /// Destructor
        ~RegisteredSimulatedObjects (void);

        void Register (RegisteredSimulatedObject* in_pObject);
        void Unregister (RegisteredSimulatedObject* in_pObject);
        RegisteredSimulatedObject* Get (std::string const & in_name);
        void WriteObjectsToTestEvents(void);
        void Dump(EventContext in_context = CTX_INFO) const;

        MapName2This GetAllObjects() const;

    private:
        boost::mutex mutable m_registrationMutex;

        typedef std::map<RegisteredSimulatedObject*,std::string> MapThis2Name;

        MapThis2Name m_mapThis2Name;
        MapName2This m_mapName2This;

    }; // end class GlobalOptions

    DEFINE_SINGLETON_POINTER (RegisteredSimulatedObjects);


    //-------------------------------------------------------------------------
    // Formating test output
    //-------------------------------------------------------------------------

    // forward declaration;
    class Formatter;

    namespace Formating
    {
        /// The different kinds of information blocks availablle within output
        enum InfoType
        {
            CLASS_NAME,
            METHOD_NAME,
            OBJECT_NAME,
            USER_TEXT
        };

        //-------------------------------------------------------------------------

        /// Each information block consists of the type of information and the configured
        /// textual prefix and postfix.
        struct InfoBlock
        {
            InfoBlock (
                InfoType in_infoType,
                std::string const & in_preFix = " ",
                std::string const & in_postFix = "")
                : m_infoType (in_infoType)
                , m_preFix   (in_preFix)
                , m_postFix  (in_postFix)
            {
            }

            InfoType    m_infoType;
            std::string m_preFix;
            std::string m_postFix;
        };

        typedef std::vector<InfoBlock> InfoBlocks;

        //-------------------------------------------------------------------------

        /// Each instance of type FormatInstruction describes the order and contents
        /// of a textual information sequence.
        class FormatInstruction
        {
        public:
            void Reset (void);
            void AddInfoBlock (InfoBlock const & in_infoBlock);
        private:
            InfoBlocks m_infoBlocks;
            friend class ::TestToolBox::Formatter;

        }; // end of class FormatInstruction
    }

    //-------------------------------------------------------------------------

    /// This class is responsible to build the output format from a given
    /// FormatInstruction and a user provided set of specific textual informations.
    class Formatter
    {
    public:
        USE_AS_SINGLETON(Formatter);

        /// \brief Constructor.
        explicit Formatter (void);

        /// Destructor
        ~Formatter (void);


        void SetFormatInstructionForMethodCall (Formating::FormatInstruction const & in_formatInstruction);

        Formating::FormatInstruction SaveFormatInstructionForMethodCall (void);
        void RestoreFormatInstructionForMethodCall (void);

        std::string FormatMethodCall (
            std::string const & in_className,
            std::string const & in_methodName,
            std::string const & in_objectName,
            std::string const & in_userText);

    private:
        Formating::FormatInstruction m_methodFormatInstruction;
        Formating::FormatInstruction m_savedMethodFormatInstruction;

    }; // end of class Formatter

    DEFINE_SINGLETON_POINTER (Formatter);


    //-------------------------------------------------------------------------
    // Base class for simulated objectss (fake objects)
    //-------------------------------------------------------------------------

    /// \brief This base class supports some features useful for simulated objects:
    ///
    /// - setting/unsetting/checking of options
    /// - tracing
    /// - generating sync events
    ///
    /// Remarks:
    /// The derived class may have const methods which nevertheless
    /// want to make use of base class functionality. To avoid casting to non const
    /// this pointer, base class methods are set const, members are set mutable.
    class SimulatedObjectBase
    {
    public:

        /// Name of the (derived) class the object belongs to.
        /// Class name is given as second constructor argument and can be any user defined string
        /// When passing TTB_CLASS_NAME it will be identical to the name of the C++ class
        std::string m_className;

    public:
        /// Constructor
        explicit SimulatedObjectBase(
            std::string const & in_objectName,
            std::string const & in_className = "SimulatedObjectBase");

        // Destructor
        virtual ~SimulatedObjectBase();

        virtual void _Dump(EventContext in_context = CTX_INFO) const;

        virtual void SetObjectName (std::string const&  in_objectName)
        {m_objectName = in_objectName;}

        const std::string& GetClassName() const {return m_className;} 
        const std::string& GetObjectName() const {return m_objectName;} 
        const std::string& GetTraceName() const {return m_objectName;} 

        // -----  Setting and checking options  -----

        void _SetOption (
            std::string const & in_option,
            std::string const & in_scope             = Option::ALL,
            OptionType::Enum    in_optionType        = OptionType::PERMANENTLY_ACTIVE,
            int                 in_numRemainingCalls = 0) const;
        void _SetOptionValue(
            std::string const & in_name,
            std::string const & in_value,
            std::string const & in_scope = Option::ALL) const;
        void _ResetOption(
            std::string const & in_option,
            std::string const & in_scope = Option::ALL) const;
        void _ResetOptions (void) const;
        bool _IsOptionSet (
            std::string const & in_option,
            std::string const & in_scope = Option::ALL) const;
        std::string _GetOptionValueStr(
            std::string const & in_name,
            std::string const & in_scope = Option::ALL) const;
        /// Get the given option in numeric/any streamable format
        template<class T>
        T _GetOptionValue(
            const std::string& in_nameOption,
            std::string const & in_scope = Option::ALL) const
        {

            std::string optionValAsString = _GetOptionValueStr(in_nameOption, in_scope);
            std::istringstream iss(optionValAsString);
            T result = T();
            if ((iss >> result).fail())
            {
                result = T(); // reset to default value
                std::ostringstream oss;
                oss << "Could not convert option value >" << in_nameOption
                    << "< with value >" << optionValAsString << "<"
                    << " to type >" << typeid(T).name() << "<";
                TRCF("SimulatedObjectBase::_GetOptionValue");
                TRC(TL_ERROR, "WARNING: %s", oss.str().c_str());
            }
            return result;
        }

        bool _IsSilent (std::string const & in_scope = Option::ALL) const;
        bool _IsError(std::string const & in_scope = Option::ALL) const;

        Options& _GetOptions(void) const;


        // -----  Support for blocking calls  -----

        /// Continue the execution of a method call which has been blocked
        void _ContinueExecution (std::string const & in_methodName) const;

    protected:

        /// Generic method execution with support for writing to TestEvents,
        /// generating sync events and blocking.
        void _MethodCall (
            std::string const & in_methodName,
            std::string const & in_callInfo = "",
            StringFunc in_func = StringFuncEmpty) const;

        /// Same as _MethodCall with additional support for returning an
        /// (error) return value of any type.
        /// Assumptions:
        ///  - return type is default constructable (signalling "no error")
        ///  - there exists a function overload TestToolBox::MakeReturnValue(..)
        ///    for the specific return type
        template <typename RT>
        RT _MethodCallReturn(
            std::string const & in_methodName,
            std::string const & in_callInfo = ""
            ) const
        {
            typename RT retVal = RT();

            // in_methodName may come from __FUNCTION__ macro and may therefore
            // contain unwanted additional info about namespace and class name
            std::string methodName = ExtractFunctionName(in_methodName);

            _MethodCall(methodName,

                // Log input params
                in_callInfo,

                // Delayed processing of output/return values
                ([&](void) -> std::string
                {
                    bool simulateFailure = _IsError(methodName);

                    // Call user supplied function to generate return value.
                    // Context info about error flag, method name and called object
                    // is passed to the function.
                    std::string info = TestToolBox::MakeReturnValue(retVal,
                        simulateFailure, methodName, this);

                    if (info.empty() && simulateFailure)
                    {
                        info = "return error";
                    }
                    // info for output
                    return info;
                }));

            return retVal;
        };

        void _EnsureWaitableConditionExists(std::string const & methodName) const;

        Option* _FindOption(
            std::string const & in_option,
            std::string const & in_scope) const;

    private:

        boost::mutex mutable m_mutex;

        /// Object name used to identify an instance of simulated object
        /// Use private access to allow safe registration within RegisteredSimulatedObjects.
        std::string m_objectName;

        /// Object specific set of options
        Options mutable m_options;

        /// map with conditions to wait for;
        /// used for realizing blocking method calls which will return when they receive
        /// an explicit trigger from test environment
        WaitableConditions mutable m_waitableConditions;

    }; // class SimulatedObjectBase

    //-------------------------------------------------------------------------

    class RegisteredSimulatedObject : public SimulatedObjectBase
    {
    public:

        /// Constructor
        explicit RegisteredSimulatedObject (
            std::string const & in_objectName,
            std::string const & in_className = "SimulatedObjectBase");

        /// Destructor
        ~RegisteredSimulatedObject();

        virtual void SetObjectName (std::string const &  in_objectName);
    };

    //-------------------------------------------------------------------------

    /// \brief This class stores current test settings within constructor and
    /// restores them within destructor. All manipulations of test settings
    /// betweeen constructor and destructor call are restored.
    ///
    /// The following settings will be restored:
    /// - global settings
    /// - specifc settings (options) for all registtered simulated objects
    class AutoRestoreOptions
    {
    public:
        AutoRestoreOptions();
        ~AutoRestoreOptions();
    private:
        OptionData m_globalOptionData;
        std::map<std::string, OptionData> m_objectSpecificOptionData;
    };

    //-------------------------------------------------------------------------

    void DumpAllOptions(EventContext in_context = CTX_INFO);

    //-------------------------------------------------------------------------

    __declspec(selectany) std::string const Option::ALL = "all";
    __declspec(selectany) std::string const Option::oSILENT         = "silent:";
    __declspec(selectany) std::string const Option::oERROR          = "error:";
    __declspec(selectany) std::string const Option::oBLOCKING       = "blocking:";
    __declspec(selectany) std::string const Option::oSYNC           = "sync:";
    __declspec(selectany) std::string const Option::oOBJECT_NAME    = "objectName:";

    //-------------------------------------------------------------------------

}; // namespace TestToolBox
