/// \file
/// \author Gerald Fahrnholz

#pragma once

#include "SimulatedObjectBase.h"

#include "Synchronizer.h"
#include "TestEvents.h"

#include <boost/filesystem.hpp>
#include <boost/system/error_code.hpp>
#include <boost/algorithm/string.hpp> 

///\cond do not document


namespace TestToolBox
{
    //-----------------------------------------------------------------------------
    // Implementation of class Option
    //-----------------------------------------------------------------------------

    TTB_INLINE bool Option::IsSet (void)
    {
        bool retVal (false);

        switch (m_optionType)
        {

        case OptionType::NOT_ACTIVE:
            retVal = false;
            break;

        case OptionType::PERMANENTLY_ACTIVE:
            retVal = true;
            break;

        case OptionType::AFTER_N_CALLS_PERMANENTLY_ACTIVE:
            if (m_numRemainingCalls == 1)
            {
                retVal = true;
                m_optionType = OptionType::PERMANENTLY_ACTIVE;
                m_numRemainingCalls = 0;
            }
            else if (m_numRemainingCalls > 1)
            {
                --m_numRemainingCalls;
            }
            break;

        case OptionType::AFTER_N_CALLS_ACTIVE_ONCE:
            if (m_numRemainingCalls == 1)
            {
                retVal = true;
                // deactivate option by removing it
                m_optionType = OptionType::NOT_ACTIVE;
            }
            else if (m_numRemainingCalls > 1)
            {
                --m_numRemainingCalls;
            }
            break;
        }
        return retVal;
    }

    //-----------------------------------------------------------------------------
    // Implementation of class Options
    //-----------------------------------------------------------------------------

    TTB_INLINE void Options::Set (
        std::string const & in_option,
        std::string const & in_scope,
        OptionType::Enum    in_optionType,
        int                 in_numRemainingCalls) const
    {
         if (in_optionType == OptionType::ACTIVE_ONCE) // special case for easier usage by client
         {
             m_optionData[in_option+in_scope].m_optionType        = OptionType::AFTER_N_CALLS_ACTIVE_ONCE;
             m_optionData[in_option+in_scope].m_numRemainingCalls = 1;
         }
         else
         {
             m_optionData[in_option+in_scope].m_optionType = in_optionType;
             m_optionData[in_option+in_scope].m_numRemainingCalls = in_numRemainingCalls;
         }
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void Options::Reset (
        std::string const & in_option,
        std::string const & in_scope) const
    {
        m_optionData.erase(in_option+in_scope);
    }

    //-----------------------------------------------------------------------------
 
    TTB_INLINE void Options::ResetAll (void) const
    {
        m_optionData.clear();
    }

    //-------------------------------------------------------------------------

    TTB_INLINE Option* Options::FindOptionExact (
            std::string const & in_option,
            std::string const & in_scope) const
    {
        OptionData::iterator itOption = m_optionData.end();

        std::string key = in_option+in_scope;
        itOption = m_optionData.find(key);

        if (itOption == m_optionData.end()) // no option found
        {
            return 0;
        }
        else
        {
            return &itOption->second;
        }
    }

    //-------------------------------------------------------------------------

    TTB_INLINE Option* Options::FindOption (
            std::string const & in_option,
            std::string const & in_scope) const
    {
        Option* pOption = FindOptionExact(in_option, in_scope);

        if (!pOption) // no specific option found, now check for scope ALL
        {
            pOption = FindOptionExact(in_option, Option::ALL);
        }

        return pOption;
    }

    //-------------------------------------------------------------------------

    TTB_INLINE bool Options::IsSet (
            std::string const & in_option,
            std::string const & in_scope) const
    {
        bool retVal = false;

        Option* pOption = FindOption(in_option, in_scope);

        if (pOption) // option really exists
        {
            retVal = pOption->IsSet();
        }

        return retVal;
    }

    //-------------------------------------------------------------------------

    TTB_INLINE bool Options::IsSilent (std::string const & in_scope) const
    {
        return IsSet(Option::oSILENT, in_scope);
    }

    //-------------------------------------------------------------------------

    TTB_INLINE bool Options::IsError (std::string const & in_scope) const
    {
        return IsSet(Option::oERROR, in_scope);
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE OptionData Options::Save (void) const
    {
        m_savedOptionData = m_optionData;
        return m_savedOptionData;
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void Options::Restore (void) const
    {
        m_optionData = m_savedOptionData;
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void Options::Restore (OptionData const & in_optionData) const
    {
        m_optionData = in_optionData;
    }

    //-----------------------------------------------------------------------------
    // Implementation of class GlobalOptions
    //-----------------------------------------------------------------------------
    
    TTB_INLINE GlobalOptions::GlobalOptions(void)
    {
        TRCF("GlobalOptions::GlobalOptions");
        TRC(TL_DEBUG,"GlobalOptions: Constructor");

        // possible initial settings to come here
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE GlobalOptions::~GlobalOptions(void)
    {
        TRCF("GlobalOptions::~GlobalOptions");
        TRC(TL_DEBUG,"GlobalOptions: Destructor");
    }

    //-----------------------------------------------------------------------------
    // Implementation of class RegisteredSimulatedObjects
    //-----------------------------------------------------------------------------

    TTB_INLINE RegisteredSimulatedObjects::RegisteredSimulatedObjects(void)
    {
        TRCF("RegisteredSimulatedObjects::RegisteredSimulatedObjects");
        TRC(TL_DEBUG,"RegisteredSimulatedObjects:Constructor");
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE RegisteredSimulatedObjects::~RegisteredSimulatedObjects(void)
    {
        TRCF("RegisteredSimulatedObjects::~RegisteredSimulatedObjects");
        TRC(TL_DEBUG,"RegisteredSimulatedObjects:Destructor");
        if (m_mapName2This.size() > 0)
        {
            TRC(TL_DEBUG,"RegisteredSimulatedObjects:Destructor: WARNING! There are unregistered objects:");
        }
        for (auto it = m_mapName2This.begin();it!=m_mapName2This.end(); ++it)
        {
            TRC(TL_DEBUG,"RegisteredSimulatedObjects:Destructor unregistered: %s", it->first.c_str());
        }
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void RegisteredSimulatedObjects::Register (RegisteredSimulatedObject* in_pObject)
    {
        boost::mutex::scoped_lock lock(m_registrationMutex);
        TRCF("RegisteredSimulatedObjects::Register");
        TRC(TL_DEBUG,"RegisteredSimulatedObjects: Register %s", in_pObject->GetObjectName().c_str());
        m_mapThis2Name[in_pObject] = in_pObject->GetObjectName();
        m_mapName2This[in_pObject->GetObjectName()] = in_pObject;
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void RegisteredSimulatedObjects::Unregister (RegisteredSimulatedObject* in_pObject)
    {
        boost::mutex::scoped_lock lock(m_registrationMutex);
        std::string objectName = in_pObject->GetObjectName();
        TRCF("RegisteredSimulatedObjects::Unregister");
        TRC(TL_DEBUG,"RegisteredSimulatedObjects: Unregister %s", objectName.c_str());
        m_mapThis2Name.erase(in_pObject);
        m_mapName2This.erase(objectName);
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE RegisteredSimulatedObject* RegisteredSimulatedObjects::Get (std::string const & in_objectName)
    {
        boost::mutex::scoped_lock lock(m_registrationMutex);
        auto pos = m_mapName2This.find(in_objectName);
        if (pos != m_mapName2This.end())
        {
            return pos->second;
        }
        else
        {
            return 0;
        }
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void RegisteredSimulatedObjects::WriteObjectsToTestEvents (void)
    {
        for (MapName2This::iterator it = m_mapName2This.begin(); it != m_mapName2This.end(); ++it)
        {
            TTB_ACT_S (it->first); // object name
        }
    }

    //-----------------------------------------------------------------------------
    // Implementation of class FormatInstruction
    //-----------------------------------------------------------------------------

    namespace Formating
    {
        TTB_INLINE void FormatInstruction::Reset (void)
        {
            m_infoBlocks.clear();
        }

        //-----------------------------------------------------------------------------

        TTB_INLINE void FormatInstruction::AddInfoBlock (InfoBlock const & in_infoBlock)
        {
            m_infoBlocks.push_back(in_infoBlock);
        }

    } // namespace Formating

    //-----------------------------------------------------------------------------
    // Implementation of class Formatter
    //-----------------------------------------------------------------------------
    
    TTB_INLINE Formatter::Formatter(void)
    {
        TRCF("Formatter::Formatter");
        TRC(TL_DEBUG,"Formatter: Constructor");

        // default settings for output format
        Formating::FormatInstruction instruction;
        instruction.AddInfoBlock(Formating::InfoBlock(Formating::METHOD_NAME));
        instruction.AddInfoBlock(Formating::InfoBlock(Formating::USER_TEXT));
        instruction.AddInfoBlock(Formating::InfoBlock(Formating::OBJECT_NAME," (",")"));

        SetFormatInstructionForMethodCall(instruction);
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE Formatter::~Formatter(void)
    {
        TRCF("Formatter::~Formatter");
        TRC(TL_DEBUG,"Formatter: Destructor");
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void Formatter::SetFormatInstructionForMethodCall (
        Formating::FormatInstruction const & in_formatInstruction)
    {
        m_methodFormatInstruction = in_formatInstruction;
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE Formating::FormatInstruction Formatter::SaveFormatInstructionForMethodCall (void)
    {
        m_savedMethodFormatInstruction = m_methodFormatInstruction;
        return m_savedMethodFormatInstruction;
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void Formatter::RestoreFormatInstructionForMethodCall (void)
    {
        m_methodFormatInstruction = m_savedMethodFormatInstruction;
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE std::string Formatter::FormatMethodCall (
        std::string const & in_className,
        std::string const & in_methodName,
        std::string const & in_objectName,
        std::string const & in_userText)
    {
        std::string formattedText;

        for (Formating::InfoBlocks::iterator it = m_methodFormatInstruction.m_infoBlocks.begin();
            it!= m_methodFormatInstruction.m_infoBlocks.end();++it)
        {
            // Get content for given info type
            std::string value;
            switch (it->m_infoType)
            {
            case Formating::CLASS_NAME:
                value = in_className;
                break;
            case Formating::METHOD_NAME:
                value = in_methodName;
                break;
            case Formating::OBJECT_NAME:
                value = in_objectName;
                break;
            case Formating::USER_TEXT:
                value = in_userText;
                break;
            default:
                break;
            }

            // Add content to output text
            if (!value.empty()) // there is something for output
            {
                if (!formattedText.empty()) // something preceds the text
                {
                    formattedText += it->m_preFix;
                }
                formattedText += value;
                formattedText += it->m_postFix;
            }
        }
        return formattedText;
    }


    //-----------------------------------------------------------------------------
    // Implementation of class SimulatedObjectBase
    //-----------------------------------------------------------------------------

    TTB_INLINE SimulatedObjectBase::SimulatedObjectBase(
        std::string const & in_objectName,
        std::string const & in_className)
        : m_objectName (in_objectName)
        , m_className (in_className)

    {
        if (m_className.empty())
        {
            m_className = TTB_CLASS_NAME;
        }

        TRCF("SimulatedObjectBase::SimulatedObjectBase");
        TRC(TL_DEBUG,"%s: Constructor, class=%s", m_objectName.c_str(), m_className.c_str());
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE SimulatedObjectBase::~SimulatedObjectBase(void)
    {
        TRCF("SimulatedObjectBase::~SimulatedObjectBase");
        TRC(TL_DEBUG,"%s: Destructor", m_objectName.c_str());
    }

    //-------------------------------------------------------------------------

    TTB_INLINE void SimulatedObjectBase::_SetOption (
        std::string const & in_option,
        std::string const & in_scope,
        OptionType::Enum    in_optionType,
        int                 in_numRemainingCalls) const
    {
        m_options.Set(in_option, in_scope, in_optionType, in_numRemainingCalls);
    }

    //-------------------------------------------------------------------------

    TTB_INLINE void SimulatedObjectBase::_ResetOption (
        std::string const & in_option,
        std::string const & in_scope) const
    {
        m_options.Reset(in_option, in_scope);
    }

    //-------------------------------------------------------------------------

    TTB_INLINE void SimulatedObjectBase::_ResetOptions (void) const
    {
        m_options.ResetAll();
    }

    //-------------------------------------------------------------------------

    TTB_INLINE bool SimulatedObjectBase::_IsOptionSet (
            std::string const & in_option,
            std::string const & in_scope) const
    {
        bool retVal (false);

        // When called from _MethodCall scope is identical to the name of the calling function
        // "SomeMethod". In all other cases scope may be any user defined description.

        // Search within instance specific options
        // for scope "SomeMethod" or if not existing for scope ALL
        Option* pOption = m_options.FindOption(in_option, in_scope);

        // If nothing found search within global options

        if (!pOption) // Search for option "SomeClass::SomeMethod"
        {
            pOption = TheGlobalOptions()->FindOptionExact(in_option, m_className + "::" + in_scope);
        }
        if (!pOption) // Search for option "SomeClass::all"
        {
            pOption = TheGlobalOptions()->FindOptionExact(in_option, m_className + "::" + Option::ALL);
        }
        if (!pOption) // Search for option "SomeClass"
        {
            pOption = TheGlobalOptions()->FindOptionExact(in_option, m_className);
        }
        if (!pOption) // Search for option "SimulatedObjectBase"
        {
            pOption = TheGlobalOptions()->FindOptionExact(in_option, "SimulatedObjectBase");
        }
        if (!pOption) // Search for option "SomeMethod" or "all"
        {
            pOption = TheGlobalOptions()->FindOption(in_option, in_scope);
        }

        if (pOption)
        {
            retVal = pOption->IsSet();
        }

        return retVal;
    }

    //-------------------------------------------------------------------------

    TTB_INLINE bool SimulatedObjectBase::_IsSilent (std::string const & in_scope) const
    {
        return _IsOptionSet(Option::oSILENT, in_scope);
    }

    //-------------------------------------------------------------------------

    TTB_INLINE bool SimulatedObjectBase::_IsError (std::string const & in_scope) const
    {
        return _IsOptionSet(Option::oERROR, in_scope);
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void SimulatedObjectBase::_ContinueExecution (std::string const & in_methodName) const
    {
        boost::mutex::scoped_lock lock(m_mutex);
        WaitableConditions::iterator it = m_waitableConditions.find(in_methodName);
        if (it != m_waitableConditions.end())
        {
            m_waitableConditions[in_methodName]->Set(true);
        }
        else
        {
            TTB_ACT_S("_ContinueExecution: ERROR: waitable condition not found for "
                << in_methodName << " (" << m_objectName << ")");
        }
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void SimulatedObjectBase::_MethodCall (
        std::string const & in_methodName,
        std::string const & in_callInfo,
        StringFunc          in_func) const
    {
        // in_methodName may come from __FUNCTION__ macro and may therefore
        // contain unwanted additional info about namespace and class name
        std::string methodName = ExtractFunctionName(in_methodName);

        // usage of object name may be switched off for a specific instance
        std::string objectName = _IsOptionSet(Option::oOBJECT_NAME, methodName) ? m_objectName : "";

        bool waitNeeded = false;

        if (_IsOptionSet(Option::oBLOCKING, methodName))
        {
            _EnsureWaitableConditionExists(methodName);
            waitNeeded = !m_waitableConditions[methodName]->IsTrue();
        }

        if (!waitNeeded) // -----  simple synchronous call  -----
        {
            // call lambda function given by client
            std::string infoFromFuncCall = in_func();

            if (!_IsOptionSet(Option::oSILENT, methodName))
            {
                // Build formatted output text
                std::string text = TheFormatter()->FormatMethodCall (
                    m_className, methodName, objectName,
                    in_callInfo
                    + (!in_callInfo.empty() && !infoFromFuncCall.empty() ? " " : "")
                    + infoFromFuncCall);

                // Pass to TestEvents
                TTB_ACT_S(text);

                // Generate sync event if needed
                if (_IsOptionSet(Option::oSYNC, methodName))
                {
                    TTB_SYNC();
                }
            }
        }
        else // -----  blocking call to be continued by trigger from test script  -----
        {
            // In any case ignore silent and noSync options to make proper synchronization
            // with test script easier.

            // Build formatted output text
            std::string text = TheFormatter()->FormatMethodCall(
                m_className, methodName+"-Start", objectName,
                in_callInfo);

            TTB_ACT_S(text);
            TTB_SYNC();

            // Wait here until a trigger for continue arrives (usually from the test script)
            m_waitableConditions[methodName]->WaitUntilTrue();

            // -----  continue a blocked call  -----
            {
                // Settings may have changed during wait
                 objectName = _IsOptionSet(Option::oOBJECT_NAME, methodName) ? m_objectName : "";

                // delayed execution of lambda function gives the tester a chance of controlling
                // the reaction (e.g. when function is blocked the tester can set some (error)
                // condition which may influence the result of the lamda function)
                std::string infoFromFuncCall = in_func();

                // Build formatted output text
                std::string text = TheFormatter()->FormatMethodCall(
                    m_className, methodName+"-Stop", objectName,
                    infoFromFuncCall);

                TTB_ACT_S(text);
                TTB_SYNC();
            }
        }
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void SimulatedObjectBase::_EnsureWaitableConditionExists(
        std::string const & in_methodName) const
    {
        boost::mutex::scoped_lock lock(m_mutex);
        WaitableConditions::iterator it = m_waitableConditions.find(in_methodName);
        if (it == m_waitableConditions.end())
        {
            // create waitable condition which is initially not fulfilled
            m_waitableConditions[in_methodName].reset(new WaitableCondition);
        }
    }

    //-----------------------------------------------------------------------------

    //-----------------------------------------------------------------------------
    // Implementation of class RegisteredSimulatedObject
    //-----------------------------------------------------------------------------

    TTB_INLINE RegisteredSimulatedObject::RegisteredSimulatedObject(
        std::string const & in_objectName,
        std::string const & in_className)
        : SimulatedObjectBase(in_objectName, in_className)
    {
        TheRegisteredSimulatedObjects()->Register(this);
        TRCF("RegisterdSimulatedObject::RegisterdSimulatedObject");
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE RegisteredSimulatedObject::~RegisteredSimulatedObject(void)
    {
        TRCF("RegisteredSimulatedObject::~RegisteredSimulatedObject");
        TRC(TL_DEBUG,"%s: Destructor", GetObjectName().c_str());
        TheRegisteredSimulatedObjects()->Unregister(this);
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void RegisteredSimulatedObject::SetObjectName (std::string const &  in_objectName)
    {

        TheRegisteredSimulatedObjects()->Unregister(this);
        SimulatedObjectBase::SetObjectName(in_objectName);
        TheRegisteredSimulatedObjects()->Register(this);
    }

    //-----------------------------------------------------------------------------

}; // namespace TestToolBox

///\endcond
