#pragma once

////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \defgroup GrpTTB_InstanceDll Using TTB_InstanceDll
/// @{
/// \ingroup GrpTestUtils 
/// \brief
/// The Win32 DLL "TTB_instanceDll.DLL" is responsible for creating and storing
/// singleton instances of framework objects from TestToolBox. This is important when
/// running a test environment which wants to make use of the test utilities also within some
/// DLLs.
/// 
/// <b>\n Motivation:</b>
///
/// Most of the test utilities are used as singletons in order to give simple access
/// to the same object from any piece of your test code.
/// Per default the following simple singleton pattern is used:
/// \code
/// class TestEvents
/// {...
/// static TestEvents* Get (void)
/// {if (!s_pTestEvents) s_pTestEvents = new TestEvents;
///  return s_pTestEvents;}
/// static TestEvents* s_pTestEvents;
/// ...}
/// \endcode
///
/// This works well as long as you use the test utilities only within your main application.
/// When making use of them also in user written DLL code the problem will arise, that each DLL
/// gets its own singleton instance when calling TestEvents::Get().
/// In most cases this is not what you want. Usually you are interested to use singleton instances
/// being unique within your whole Win32 process.
///
/// <b>\n Solution:</b>
///
/// Instead of storing simple static class members, instances are created and stored within
/// the separate TTB_instanceDll.
///
/// \n To leave the client code unaffected the access to the instance DLL is encapsulated
/// in the Get() methods and the changed behaviour is activated via preprocessor define
/// USE_TTB_INSTANCE_DLL: 
/// 
/// \code
/// // file TestEvents.h:
///
/// // Declaration of methods within TTB_InstanceDll
/// #ifdef USE_TTB_INSTANCE_DLL
/// #    include "TestToolBox/TTB_InstanceDll.h"
/// #endif
/// 
/// class TestEvents
/// {...
/// #ifdef USE_TTB_INSTANCE_DLL
///     static TestEvents* CreateNewInstanceForUsageInDll(void){return new TestEvents;};
///     static TestEvents* Get (void)
///     {return TTB_GetInstance_TestEvents();}
/// #else
///     static TestEvents* Get (void)
///     {if (!s_pTestEvents) s_pTestEvents = new TestEvents;
///      return s_pTestEvents;}
///     static TestEvents* s_pTestEvents;
/// #endif
/// ...}
///
/// // Use static memeber only when running without TTB_InstanceDll
/// #ifndef USE_TTB_INSTANCE_DLL
/// __declspec(selectany) TestEvents* TestEvents::s_pTestEvents = 0;
/// #endif
/// \endcode
///
/// <b>\n Necessary steps for using test utilities within DLLs:</b>
///
/// - generate TTB_InstanceDll and / or create dependency of your test project from it
/// - set define USE_TTB_INSTANCE_DLL within all your test projects 
/// - currently the test utilities are used as explicit classes (and not only via virtual
///   member functions). Therefore all projects have to include the implementation code
///   (e.g. TestEventsImpl.h)
///   and/or link with the used test libraries.
/// - access singleton instances via static class methods Get() (e.g. TestEvents::Get());
///   this is no difference compared to usage within a single main app!
///
/// \sa TTB_InstanceDll.h (file documentation)
/// \sa GrpTestEnvWithMultipleDlls (test app)
///
////////////////////////////////////////////////////////////////////////////////

/// \cond

#ifdef TTB_INSTANCE_DLL_EXPORTS
#define TTB_INSTANCE_DLL_API __declspec(dllexport)
#else
#define TTB_INSTANCE_DLL_API __declspec(dllimport)
#ifdef NDEBUG
#pragma comment( lib, "TTB_InstanceDll" )
#else
#pragma comment( lib, "TTB_InstanceDllD" )
#endif
#endif

#include "IProtocol.h"

namespace TestToolBox
{
    class Environment;
    class Protocol;
    class Synchronizer;
    class TestEvents;
    class TestTime;
    class GlobalOptions;
    class RegisteredSimulatedObjects;
    class Formatter;
}


TTB_INSTANCE_DLL_API TestToolBox::Environment* TTB_GetInstance_Environment (void);
TTB_INSTANCE_DLL_API void TTB_CleanupInstance_Environment (void);

TTB_INSTANCE_DLL_API TestToolBox::TestEvents* TTB_GetInstance_TestEvents (void);
TTB_INSTANCE_DLL_API void TTB_CleanupInstance_TestEvents (void);

TTB_INSTANCE_DLL_API TestToolBox::TestTime* TTB_GetInstance_TestTime (void);
TTB_INSTANCE_DLL_API void TTB_CleanupInstance_TestTime (void);

TTB_INSTANCE_DLL_API TestToolBox::Synchronizer* TTB_GetInstance_Synchronizer (void);
TTB_INSTANCE_DLL_API void TTB_CleanupInstance_Synchronizer (void);

TTB_INSTANCE_DLL_API TestToolBox::Protocol* TTB_GetInstance_Protocol (unsigned int in_allowedStreams, TestToolBox::OutputLevel::Enum in_outputLevel);
TTB_INSTANCE_DLL_API void TTB_CleanupInstance_Protocol (void);

TTB_INSTANCE_DLL_API TestToolBox::GlobalOptions* TTB_GetInstance_GlobalOptions (void);
TTB_INSTANCE_DLL_API void TTB_CleanupInstance_GlobalOptions (void);

TTB_INSTANCE_DLL_API TestToolBox::RegisteredSimulatedObjects* TTB_GetInstance_RegisteredSimulatedObjects (void);
TTB_INSTANCE_DLL_API void TTB_CleanupInstance_RegisteredSimulatedObjects (void);

TTB_INSTANCE_DLL_API TestToolBox::Formatter* TTB_GetInstance_Formatter (void);
TTB_INSTANCE_DLL_API void TTB_CleanupInstance_Formatter (void);

///\endcond

///@} // end group GrpTestInstanceDll
