////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \defgroup GrpDemoApp DemoApp
/// @{
/// \ingroup GrpHelperApps
/// \brief
/// Win32 console application "DemoApp.exe" is a simple program
/// which can simulate erraneous program behaviour (exit error code,
/// endless loops, invalid memory access). The required
/// behaviour can be controlled via command line args.
///
/// \n
/// The program may represent the behaviour of a typical test
/// executable during sw development.
///
/// <b>Syntax:</b>\n
/// \code
/// DemoApp EXIT_CODE|MEMORY_ACCESS_FAILURE|DIVIDE_BY_ZERO|ENDLESS_LOOP  [exitCode]
/// \endcode
///
/// EXIT_CODE NN\n
/// correct program function with simply returning the specified
/// code NN (may be 0 or any other value)
///
/// MEMORY_ACCESS_FAILURE\n
/// performs an illegal access to a not existing element of std::vector;
/// typical reaction: assert message box
///
/// DIVIDE_BY_ZERO\n
/// division of zero typically causes a Microsoft system message box
///
/// ENDLESS_LOOP\n
/// program will hang and consume 100% of processor capacity
///
/// \sa DemoApp.cpp
///
////////////////////////////////////////////////////////////////////////////////

///\cond do not document

#define WIN32_LEAN_AND_MEAN		// Selten verwendete Teile der Windows-Header nicht einbinden.
#include <windows.h>
#include <tchar.h>

#include <iostream>
#include <sstream>
#include <vector>

using namespace std;


//*****************************************************************************
//*****************************************************************************


void DisplayStartArgs (int argc, _TCHAR* argv[])
{
    cout << "[DemoApp] ----- Args Begin -----" << endl;
    cout << "[DemoApp] Commandline : " << ::GetCommandLine() << endl;
    cout << "[DemoApp] argc = " << argc << endl;
    for (int i = 0; i <argc; ++i)
    {
        cout << "[DemoApp] arg " << i << ": >" << argv[i] << "<" << endl;
    }
    cout << "[DemoApp] ----- Args End -----" << endl;

} // DisplayStartArgs


//*****************************************************************************
//*****************************************************************************


void DisplayEnvironment (void)
{
    LPSTR lpszVariable; 
    LPVOID lpvEnv; 
     
    cout << "[DemoApp] ----- Environment Begin -----" << endl;
 
    // Get a pointer to the environment block. 
    lpvEnv = ::GetEnvironmentStrings();

    // If the returned pointer is NULL, exit.
    if (lpvEnv == NULL)
    {
       cout << "[DemoApp]: GetEnvironmentStrings failed with error"
            << GetLastError() << endl; 
    }
    else
    {
        // Variable strings are separated by NULL byte, and the block is 
        // terminated by a NULL byte. 
        for (lpszVariable = (LPTSTR) lpvEnv; *lpszVariable; lpszVariable++) 
        { 
           cout << "[DemoApp]: ";    
           while (*lpszVariable) 
              cout << *lpszVariable++;
           cout << endl;
        }

        ::FreeEnvironmentStrings((LPCH)lpvEnv);
    }

    cout << "[DemoApp] ----- Environment End -----" << endl;

} // DisplayEnvironment


//*****************************************************************************
//*****************************************************************************


int _tmain(int argc, _TCHAR* argv[])
{
    int exitCode = 0; // assume correct execution

    // wait some time before proceeding and writing to stdout;
    // there is no synchronization used when writing to stdout;
    // otherwise the output of this process and
    // the calling process would completely get mixed.
    // (Would printf instead of cout provide better synchronization?)
    ::Sleep(100); 

    DisplayStartArgs(argc, argv);
    // DisplayEnvironment();

    // ----- React according to start arguments -----

    if (argc > 1)
    {
        std::string arg = argv[1];
        if (arg=="ENDLESS_LOOP")
        {
            cout << "[DemoApp]: starting endless loop..." << endl;
            for (int i = 1; i < 10; i)
            {
            }
        }
        else if (arg=="MEMORY_ACCESS_FAILURE")
        {
            cout << "[DemoApp]: performing invalid memory access..." << endl;
            std::vector<int> v;
            v[5] = 1;
        }
        else if (arg=="DIVIDE_BY_ZERO")
        {
            cout << "[DemoApp]: performing divide by zero..." << endl;
            int j = 1;
            int i = 4 / (j-1);
        }
        else if (arg=="EXIT_CODE")
        {
            // Set exit code
            if (argc > 2)
            {
                std::istringstream is (argv[2]);
                is >> exitCode;
            }
        }
    }
    
    cout << "[DemoApp]: returns exit code " << exitCode << endl;
	return exitCode;
}

///\endcond do not document

///@} // end group GrpDemoApp
