////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \defgroup GrpTestEnvironment TestEnvironment
/// @{
/// \ingroup GrpTestApps 
/// \brief
/// Win32 console application "TestEnvironment.exe" demonstrates
/// the usage of class Environment which allows access to several aspects
/// of the runtime environment (directories, command line, environment variable).
/// 
/// <b>\n Log: Run with specific command line arguments (see below):</b>
///
/// \code
/// Initial settings after default construction
///   GetNameOfComputer      : SIRIUS
///   GetNameOfUser          : Gerald
///   GetCommandLineStr      : TestEnvironment.exe -trace TestEnvironment.6 -verbous 2.15 ThisWillBeSkipped -optWithoutParam -lastOpt lastValue 
///   GetExeFileFullPath     : C:\My Target\debug\TestEnvironment.exe
///   GetExeDir              : C:\My Target\debug
///   GetExeDrive            : C:
///   GetSourceDir           : c:\userdata\gerald\sw\c\testtoolbox
///   GetSourceDrive         : c:
///   GetOutputDir           : C:\My Target\debug
///   GetReferenceDir        : C:\My Target\debug
/// 
/// Cut Source dir (CutSourceDirAfter), explicitly set out/ref dirs (SetOutputDir, SetReferenceDir)
///   GetSourceDir           : c:\userdata\gerald\sw
///   GetOutputDir           : W:\MyDirs\TestOutput
///   GetReferenceDir        : Z:\MyArchive\Expected Test Reactions
/// 
/// Explicitly set source directory (SetSourceDir)
///   GetSourceDir           : Q:\SomeImaginaryDir\MyProject
///   GetSourceDrive         : Q:
/// 
/// Aceessing command line
/// GetCommandLine         : TestEnvironment.exe -trace TestEnvironment.6 -verbous 2.15 ThisWillBeSkipped -optWithoutParam -lastOpt lastValue 
/// Existence Check (IsExistingCommandLineOption)
///   -trace               : true
///   -notExisting         : false
/// Access as string (GetCommandLineOption)
///   -verbous             : >2.15<
///   -trace               : >TestEnvironment.6<
///   -notExisting         : ><
/// Access as numeric value (GetCommandLineOptionVal<>, ignoring conversion errors)
///   -verbous (int)       : 2
///   -verbous (float)     : 2.15
///   -trace (long)        : 0
///   -notExisting (double): 0
/// Detecting conversion errors by catching exceptions
///   Could not convert command line option >-trace< with value >TestEnvironment.6< to type >long<
///   Could not convert command line option >-notExisting< with value >< to type >double<
/// 
/// Accessing system environment variables
/// Existence Check (IsExistingSystemEnvVariable,GetSystemEnvVariable)
///   TTB_TestVariable                 : [not existing]
///   TTB_TestVariable2                : false
///   HOMEPATH                         : true
/// Access as string (GetSystemEnvVariable)
///   TTB_TestVariable                 : ><
///   TTB_TestVariable2                : ><
///   HOMEPATH                         : >\Dokumente und Einstellungen\Gerald<
/// Access as string after having set some values (SetSystemEnvVariable)
///   TTB_TestVariable                 : >17.345 ABC 0123<
///   TTB_TestVariable2                : >ABCD<
/// Reading as numeric value (GetSystemEnvVariableVal<>, ignoring conversion errors)
///   TTB_TestVariable (long)          : 17
///   TTB_TestVariable (double)        : 17.345
///   TTB_TestVariable2 (long)         : 0
///   TTB_NotExistingVariable (double) : 0
/// Detecting conversion errors by catching exceptions
///   Could not convert environment variable >TTB_TestVariable2< with value >ABCD< to type >long<
///   Could not convert environment variable >TTB_NotExistingVariable< with value >< to type >double<
/// 
/// Starting not existing exefile (StartExeFile,GetErrorDescription)
///   Start of executable d:\NotExistingDir\notepad.exe failed (details (GetErrorDescription) see trace output)
/// 
/// Starting second instance of this application to verify changed environment variables
///
/// Second instance starts here:
///
/// Only checking environment variables
///   TTB_TestVariable                 : >17.345 ABC 0123<
///   TTB_TestVariable2                : >ABCD<
///   TTB_LongVariable                 : >12345<
///   TTB_DoubleVariable               : >3.14<
///   TTB_BoolVariable                 : >1<
///   HOMEPATH                         : >\Dokumente und Einstellungen\Gerald<
///
/// Second instance terminates now
///
///   Success starting second instance C:\My Target\debug\TestEnvironment.exe
///
/// \endcode
///
/// \sa TestEnvironment.cpp (file documentation)
/// \sa TestToolBox::Environment (class documentation)
/// \sa GrpEnvironment (usage documentation)
///
////////////////////////////////////////////////////////////////////////////////

///\cond do not document

#include <tchar.h>
#include <assert.h>

#include "TestToolBox\EnvironmentImpl.h"

namespace TTB = TestToolBox;
using namespace std;

// Define name of this component for simple trace utility and set initial trace level
TTB_TRC_DEF_GLOBALS ("TestEnvironment", TL_VAL_DEBUG);


//************************************************************************
//************************************************************************


void TestManipulatePathSettings (void)
{
    TTB::Environment* pEnv = TTB::TheEnvironment();

    // Cut source directory
    pEnv->CutSourceDirAfter("\\Gerald\\");

    // Change output and reference directories
    pEnv->SetOutputDir("W:\\MyDirs\\TestOutput");
    pEnv->SetReferenceDir("Z:\\MyArchive\\Expected Test Reactions");

    assert(pEnv->GetSourceDir() == "c:\\userdata\\gerald");
    assert(pEnv->GetOutputDir() == "W:\\MyDirs\\TestOutput");
    assert(pEnv->GetReferenceDir() == "Z:\\MyArchive\\Expected Test Reactions");

    cout
        << "\nCut Source dir (CutSourceDirAfter), explicitly set out/ref dirs (SetOutputDir, SetReferenceDir)" << endl
        << "  GetSourceDir           : "
        << pEnv->GetSourceDir().c_str() << endl
        << "  GetOutputDir           : "
        << pEnv->GetOutputDir().c_str() << endl
        << "  GetReferenceDir        : "
        << pEnv->GetReferenceDir().c_str() << endl;

    pEnv->SetSourceDir("Q:\\SomeImaginaryDir\\MyProject\\Sources\\ModuleX");
    assert(pEnv->GetSourceDir() == "Q:\\SomeImaginaryDir\\MyProject\\Sources\\ModuleX");
    assert(pEnv->GetSourceDrive() == "Q:");
    pEnv->CutSourceDirAfter("MyProject");
    cout
        << "\nExplicitly set source directory (SetSourceDir)" << endl
        << "  GetSourceDir           : "
        << pEnv->GetSourceDir().c_str() << endl
        << "  GetSourceDrive         : "
        << pEnv->GetSourceDrive().c_str() << endl;

} // TestManipulatePathSettings


//************************************************************************
//************************************************************************


void TestAccessToCommandLine (void)
{
    TTB::Environment* pEnv = TTB::TheEnvironment();
    cout
        << "\nAceessing command line" << endl
        << "GetCommandLineStr      : "
        << pEnv->GetCommandLineStr().c_str() << endl

        // existence check
        << "Existence Check (IsExistingCommandLineOption)\n"
        << "  -trace               : "
        << std::boolalpha
        << pEnv->IsExistingCommandLineOption("-trace") << "\n"
        << "  -notExisting         : "
        << pEnv->IsExistingCommandLineOption("-notExisting") << "\n"
 
        // reading option as string
        << "Access as string (GetCommandLineOption)\n"
        << "  -verbous             : >"
        << pEnv->GetCommandLineOption("-verbous").c_str() << "<\n"
        << "  -trace               : >"
        << pEnv->GetCommandLineOption("-trace").c_str() << "<\n"
        << "  -notExisting         : >"
        << pEnv->GetCommandLineOption("-notExisting").c_str() << "<\n"

        // reading numeric value
        << "Access as numeric value (GetCommandLineOptionVal<>, ignoring conversion errors)\n"
        << "  -verbous (int)       : "
        << pEnv->GetCommandLineOptionVal<int>("-verbous") << "\n"
        << "  -verbous (float)     : "
        << pEnv->GetCommandLineOptionVal<float>("-verbous") << "\n"
        << "  -trace (long)        : "
        << pEnv->GetCommandLineOptionVal<long>("-trace") << "\n"
        << "  -notExisting (double): "
        << pEnv->GetCommandLineOptionVal<double>("-notExisting") << "\n";


    // Detect conversion error
    cout 
        << "Detecting conversion errors by catching exceptions" << endl;
    {
        std::string errorReason = "No error";
        try
        {
            long val = pEnv->GetCommandLineOptionVal<long>("-trace",TTB::Environment::CM_THROW);
        }
        catch (std::exception& e)
        {
            errorReason = e.what();
            cout << "  " << errorReason << endl;
        }
        assert (errorReason == "Could not convert command line option >-trace< with value >TestEnvironment.6< to type >long<");
    }
    {
        std::string errorReason = "No error";
        try
        {
            double val = pEnv->GetCommandLineOptionVal<double>("-notExisting",TTB::Environment::CM_THROW);
        }
        catch (std::exception& e)
        {
            errorReason = e.what();
            cout << "  " << errorReason << endl;
        }
        assert (errorReason == "Could not convert command line option >-notExisting< with value >< to type >double<");
    }

    // Set command line options
    // Remark: the command line itself remains unchanged, changes apply
    // to the internal map of stored command line options
    long   longVal   = 12345;
    double doubleVal = 3.14;
    bool   boolVal   = true;

    pEnv->SetCommandLineOptionVal("-verbous", longVal);
    pEnv->SetCommandLineOptionVal("-trace", doubleVal);
    pEnv->SetCommandLineOptionVal("-notExisting", boolVal);

    long   longVal2   = pEnv->GetCommandLineOptionVal<long>("-verbous");
    double doubleVal2 = pEnv->GetCommandLineOptionVal<double>("-trace");
    bool   boolVal2   = pEnv->GetCommandLineOptionVal<bool>("-notExisting");

    assert(longVal   == longVal2);
    assert(doubleVal == doubleVal2);
    assert(boolVal   == boolVal2);

    pEnv->SetCommandLineOptionVal("-notExisting", false);
    assert (pEnv->GetCommandLineOptionVal<bool>("-notExisting") == false);

} // TestAccessToCommandLine

 
//************************************************************************
//************************************************************************


void TestCheckEnvironmentVariables (void)
{
    TTB::Environment* pEnv = TTB::TheEnvironment();

    cout
        << "\nOnly checking environment variables\n"
        << "  TTB_TestVariable                 : >"
        << pEnv->GetSystemEnvVariable("TTB_TestVariable") << "<\n"
        << "  TTB_TestVariable2                : >"
        << pEnv->GetSystemEnvVariable("TTB_TestVariable2") << "<\n"
        << "  TTB_LongVariable                 : >"
        << pEnv->GetSystemEnvVariable("TTB_LongVariable") << "<\n"
        << "  TTB_DoubleVariable               : >"
        << pEnv->GetSystemEnvVariable("TTB_DoubleVariable") << "<\n"
        << "  TTB_BoolVariable                 : >"
        << pEnv->GetSystemEnvVariable("TTB_BoolVariable") << "<\n"
        << "  HOMEPATH                         : >"
        << pEnv->GetSystemEnvVariable("HOMEPATH") << "<\n";

} // TestCheckEnvironmentVariables


//************************************************************************
//************************************************************************


void TestAccessToEnvironmentVariables (void)
{
    TTB::Environment* pEnv = TTB::TheEnvironment();

    std::string envVarValue;
    bool exists = pEnv->GetSystemEnvVariable("TTB_TestVariable", envVarValue);
    cout
        << "\nAccessing system environment variables\n"

        // existence check
        << "Existence Check (IsExistingSystemEnvVariable,GetSystemEnvVariable)\n"
        << "  TTB_TestVariable                 : "
        << (exists ? envVarValue : "<not existing>\n")
        << "  TTB_TestVariable2                : "
        << pEnv->IsExistingSystemEnvVariable("TTB_TestVariable2") << endl
        << "  HOMEPATH                         : "
        << pEnv->IsExistingSystemEnvVariable("HOMEPATH") << endl

        // access as string
        << "Access as string (GetSystemEnvVariable)\n"
        << "  TTB_TestVariable                 : >"
        << pEnv->GetSystemEnvVariable("TTB_TestVariable") << "<\n"
        << "  TTB_TestVariable2                : >"
        << pEnv->GetSystemEnvVariable("TTB_TestVariable2") << "<\n"
        << "  HOMEPATH                         : >"
        << pEnv->GetSystemEnvVariable("HOMEPATH") << "<\n";

    // Set some values and read again
    pEnv->SetSystemEnvVariable("TTB_TestVariable", "17.345 ABC 0123");
    pEnv->SetSystemEnvVariable("TTB_TestVariable2", "ABCD");
    cout 
        << "Access as string after having set some values (SetSystemEnvVariable)\n"
        << "  TTB_TestVariable                 : >"
        << pEnv->GetSystemEnvVariable("TTB_TestVariable") << "<\n"
        << "  TTB_TestVariable2                : >"
        << pEnv->GetSystemEnvVariable("TTB_TestVariable2") << "<\n";

    cout
        << "Reading as numeric value (GetSystemEnvVariableVal<>, ignoring conversion errors)" << endl
        << "  TTB_TestVariable (long)          : "
        << pEnv->GetSystemEnvVariableVal<long>("TTB_TestVariable") << endl
        << "  TTB_TestVariable (double)        : "
        << pEnv->GetSystemEnvVariableVal<double>("TTB_TestVariable") << endl
        << "  TTB_TestVariable2 (long)         : "
        << pEnv->GetSystemEnvVariableVal<long>("TTB_TestVariable2") << endl
        << "  TTB_NotExistingVariable (double) : "
        << pEnv->GetSystemEnvVariableVal<double>("TTB_NotExistingVariable") << endl;

    // Detect conversion error
    cout 
        << "Detecting conversion errors by catching exceptions" << endl;
    {
        std::string errorReason = "No error";
        try
        {
            long val = pEnv->GetSystemEnvVariableVal<long>("TTB_TestVariable2",TTB::Environment::CM_THROW);
        }
        catch (std::exception& e)
        {
            errorReason = e.what();
            cout << "  " << errorReason << endl;
        }
        assert (errorReason == "Could not convert environment variable >TTB_TestVariable2< with value >ABCD< to type >long<");
    }
    {
        std::string errorReason = "No error";
        try
        {
            double val = pEnv->GetSystemEnvVariableVal<double>("TTB_NotExistingVariable",TTB::Environment::CM_THROW);
        }
        catch (std::exception& e)
        {
            errorReason = e.what();
            cout << "  " << errorReason << endl;
        }
        assert (errorReason == "Could not convert environment variable >TTB_NotExistingVariable< with value >< to type >double<");
    }

    // Set some environment variables
    // Remark: the changes within the environment are available for all processes
    // started from this environment
    long   longVal   = 12345;
    double doubleVal = 3.14;
    bool   boolVal   = false;

    pEnv->SetSystemEnvVariableVal("TTB_LongVariable", longVal);
    pEnv->SetSystemEnvVariableVal("TTB_DoubleVariable", doubleVal);
    pEnv->SetSystemEnvVariableVal("TTB_BoolVariable", boolVal);

    long   longVal2   = pEnv->GetSystemEnvVariableVal<long>("TTB_LongVariable");
    double doubleVal2 = pEnv->GetSystemEnvVariableVal<double>("TTB_DoubleVariable");
    bool   boolVal2   = pEnv->GetSystemEnvVariableVal<bool>("TTB_BoolVariable");

    assert(longVal   == longVal2);
    assert(doubleVal == doubleVal2);
    assert(boolVal   == boolVal2);

    pEnv->SetSystemEnvVariableVal("TTB_BoolVariable", true);
    assert (pEnv->GetSystemEnvVariableVal<bool>("TTB_BoolVariable") == true);

} // TestAccessToEnvironmentVariables


//************************************************************************
//************************************************************************

 
void TestStartingExeFiles (void)
{
    TTB::Environment* pEnv = TTB::TheEnvironment();

    cout << "\nStarting not existing exefile (StartExeFile,GetErrorDescription)\n";
    std::string exeToStart = "d:\\NotExistingDir\\notepad.exe";
    if (!TTB::Environment::StartExeFile(exeToStart))
    {
        cout << "  Start of executable " << exeToStart.c_str()
             << " failed (details (GetErrorDescription) see trace output)" << endl;
    }

    // Start second instance of this executable to verify that changes
    // to environment variables are available within a second process
    cout << "\nStarting second instance of this application to verify changed environment variables" << endl;
    exeToStart = pEnv->GetExeFileFullPath();
    bool started = TTB::Environment::StartExeFile(pEnv->GetExeFileFullPath(),"-OnlyCheckEnvironmentVariables");
    ::Sleep(500); // wait some time to avoid mixing of outputs from the two processes
    cout << "  " <<(started ? "Success" : "ERROR") << " starting second instance " << exeToStart << endl;
    assert (started==true);

} // TestStartingExeFiles


//************************************************************************
//************************************************************************

 
int _tmain(int argc, _TCHAR* argv[])
{
    TTB::SetTraceLevelFromCommandLine();

    TTB::Environment* pEnv = TTB::TheEnvironment();

    // A second instance of this executable will be called from the first one
    // to check if manipulations on the environment variables where successful
    if (pEnv->IsExistingCommandLineOption("-OnlyCheckEnvironmentVariables"))
    {
        cout << "\nSecond instance starts here:" << endl;
        TestCheckEnvironmentVariables();
        cout << "\nSecond instance terminates now\n" << endl;
        exit(0);
    }

    cout
        << "\nInitial settings after default construction" << endl
        << "  GetNameOfComputer      : "
        << pEnv->GetNameOfComputer().c_str() << endl
        << "  GetNameOfUser          : "
        << pEnv->GetNameOfUser().c_str() << endl
        << "  GetCommandLineStr      : "
        << pEnv->GetCommandLineStr().c_str() << endl
        << "  GetExeFileFullPath     : "
        << pEnv->GetExeFileFullPath().c_str() << endl
        << "  GetExeDir              : "
        << pEnv->GetExeDir().c_str() << endl
        << "  GetExeDrive            : "
        << pEnv->GetExeDrive().c_str() << endl
        << "  GetSourceDir           : "
        << pEnv->GetSourceDir().c_str() << endl
        << "  GetSourceDrive         : "
        << pEnv->GetSourceDrive().c_str() << endl
        << "  GetOutputDir           : "
        << pEnv->GetOutputDir().c_str() << endl
        << "  GetReferenceDir        : "
        << pEnv->GetReferenceDir().c_str() << endl;

    TestManipulatePathSettings();
    TestAccessToCommandLine();
    TestAccessToEnvironmentVariables();
    TestStartingExeFiles();

	return 0;
}

///\endcond do not document

///@} // end group GrpTestEnvironment
