////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \defgroup GrpTestPerformanceDataSet TestPerformanceDataSet
/// @{
/// \ingroup GrpTestApps 
/// \brief
/// Win32 console application "TestPerformanceDataSet.exe" demonstrates
/// the usage of class PerformanceDataSet which allows access to arbitrary
/// performance counter values.
/// 
/// This demo application especially examines the consumption of private bytes of
/// its own process.  As can be seen in the following logs, the operating system
/// (or the runtime library?) has its own strategy when to really allocate/deallocate
/// memory.
/// 
/// <b>Example 1 - Run with big memory allocation block:</b>
///
/// \code
/// Trying to access the specified performance counters.
/// In case of incorrect names it may need some time to detect the error...
///
/// Configured size of memory block: 1048576
///
/// Processortime in % Threads notepad.exe       Private bytes    Bytes Changed by
///               0.00                   1             2768896             2768896
///               0.00                   1             3825664             1056768
///               0.00                   1             4878336             1052672
///               0.00                   1             5935104             1056768
///               0.00                   1             6987776             1052672
/// Allocated memory has been deleted
///               1.54                   1             2768896            -4218880
///               0.00                   1             2768896                   0
///               0.00                   1             2768896                   0
/// 
/// \endcode
///
/// <b>Example 2 - Run with small memory allocation block:</b>
///
/// \code
/// Trying to access the specified performance counters.
/// In case of incorrect names it may need some time to detect the error...
///
/// Configured size of memory block: 4096
/// 
/// Processortime in % Threads notepad.exe       Private bytes    Bytes Changed by
///               0.00                   1             2768896             2768896
///               1.54                   1             2768896                   0
///              37.50                   1             2777088                8192
///               0.00                   1             2777088                   0
///               0.00                   1             2785280                8192
/// Allocated memory has been deleted
///               0.00                   1             2785280                   0
///               1.56                   1             2785280                   0
///               1.56                   1             2785280                   0
/// \endcode
///
/// Known problems with performance counters: http://support.microsoft.com/kb/300956/en-us
///
/// \sa TestPerformanceDataSet.cpp  (file documentation)
/// \sa TestToolBox::PerformanceDataSet (class documentation)
/// \sa GrpPerformanceDataSet (Usage documentation)
///
////////////////////////////////////////////////////////////////////////////////

///\cond do not document

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <tchar.h>
#include <iostream>
#include <iomanip>

#include "PerformanceDataSetImpl.h" // use as inline code

#pragma warning( disable : 4996 ) // vsprintf declared deprecated

namespace TTB = TestToolBox;


// Define name of this component for trace utility and set initial trace level
TTB_TRC_DEF_GLOBALS ("TestPerfDataSet", TL_VAL_DEBUG);

// Number of requests for performance data executed within a loop
const int  NUM_CHECKS = 5;

// Delay between two requests for performance data
const int  DELAY_BETWEEN_CHECKS_MS = 1000;


// To watch behaviour of counter "private bytes" of this process change
// the following constant value.
// In the small range allocation of memory will not immediately be visible
// through a changed performance counter value (memory allocation policy of
// OS/runtime library)
const long MEM_BLOCK_SIZE   = 1024 * 1024;
//const long MEM_BLOCK_SIZE = 1024*4;

// Output formating: width of a column
const int  COL_WIDTH = 20;


//************************************************************************
//************************************************************************
    

struct MemoryBlock
{
    char m_array [MEM_BLOCK_SIZE];
};


//************************************************************************
//************************************************************************
    

// Simple class which demonstrates access to some performance counter values
// using TestToolBox::PerfomanceDataSet
class MyPerformanceChecker
{
public:

    // Constructor.
    // Builds array of performance counter names which will be examined
    // later
    MyPerformanceChecker()
        :
        m_lastValPrivateBytes (0)
    {
        // assuming Windows XP with German language pack
        m_counterNames.push_back("\\\\SIRIUS\\Prozessor(_Total)\\Prozessorzeit (%)");
        m_counterNames.push_back("\\\\SIRIUS\\Prozess(notepad)\\Threadanzahl");
        m_counterNames.push_back("\\\\SIRIUS\\Prozess(TestPerformanceDataSet)\\Private Bytes");

        // assuming Windows XP with English language pack
        // TODO: add valid counter names here

        std::cout << "Trying to access the specified performance counters.\n"
                  << "In case of incorrect names it may need some time to detect the error...\n\n";

        if (!m_perfDataSet.InitPerfCounters (m_counterNames))
        {
            std::cout << "ERROR: Check the name of your computer and the names of the performance counters\n"
                      << "on your system (e.g. with utility ScanPerfCounters), correct the sample code\n"
                      << "and try again! (see also the trace hints for invalid counter names) \n\n";
            exit (1);
        }
        else // initialization succeeded, we can write the header of the output
        {
            std::cout
                << "Configured size of memory block: " << MEM_BLOCK_SIZE
                << std::endl << std::endl;

            std::cout
                << std::setw(COL_WIDTH) << "Processortime in %"
                << std::setw(COL_WIDTH) << "Threads notepad.exe"
                << std::setw(COL_WIDTH) << "Private bytes"
                << std::setw(COL_WIDTH) << "Bytes Changed by" 
                << std::endl;
        }
    }

    // Repeatedly check the configured performance counter values within a loop.
    // If in_releaseMemory is false allocate a predefined memory block to watch
    // effects on counter "private bytes".
    // If in_releaseMemory is true free the alocated memory and request the counter
    // values in a timed loop of 3 executions.
    void Check (bool in_releaseMemory = false)
    {
        if (in_releaseMemory)
        {
            // Release allocated memory
            for (unsigned int i = 0; i< m_pAllocatedMemory.size(); ++i)
            {
                delete (m_pAllocatedMemory[i]);
            }
            std::cout << "Allocated memory has been deleted" << std::endl;
        }

        int numLoops = in_releaseMemory ? 3 : NUM_CHECKS;
        for (int i = 0; i < numLoops; ++i)
        {
            // Get actual counter values
            m_counterValues = m_perfDataSet.GetCurrentCounterValues();

            // Access single value within array
            long curNumPrivateBytes = static_cast<long>(m_counterValues[2]);

            // Write formatted output
            std::cout
                 << std::setw(COL_WIDTH) << std::setprecision(2) << std::fixed
                     << m_counterValues[0]
                 << std::setw(COL_WIDTH) << std::setprecision(0) << m_counterValues[1]
                 << std::setw(COL_WIDTH) << std::setprecision(0) << m_counterValues[2]
                 << std::setw(COL_WIDTH) << std::setprecision(0) 
                     <<  curNumPrivateBytes-m_lastValPrivateBytes
                 << std::endl;

            m_lastValPrivateBytes = curNumPrivateBytes;

            ::Sleep(DELAY_BETWEEN_CHECKS_MS);

            if (!in_releaseMemory) // allocate new memory
            {
                m_pAllocatedMemory.push_back(new MemoryBlock);
            }
        }

    }

private:

    // Utility class which encapsulates access to system performance counters
    TTB::PerformanceDataSet m_perfDataSet;

    // vector of performance counter names we are interested in
    std::vector<std::string> m_counterNames;

    // returned vector of counter values (in double format)
    std::vector<double> m_counterValues;

    // some data to specifically watch changes in memory consumption of this process
    long m_lastValPrivateBytes;
    std::vector<MemoryBlock*> m_pAllocatedMemory; // stores pointers to allocated memory

}; // class MyPerformanceChecker


//************************************************************************
//************************************************************************


int _tmain(int argc, _TCHAR* argv[])
{
    TRCF("main");
    TTB::SetTraceLevelFromCommandLine();

    TRC(TL_PROD, "main-Begin");

    // Set the debug-heap flag so that memory leaks are reported when
    // the process terminates.
    SET_CRT_DEBUG_FIELD( _CRTDBG_LEAK_CHECK_DF );


    // Construct an object with several performance counter names (one of them are the
    // the private bytes of this process which are watched)
    MyPerformanceChecker myChecker;

    // Execute within loop: allocate memory and check perf counter values
    myChecker.Check();

    // Release allocated memory and check perf counter values again
    myChecker.Check(true);

    TRC(TL_PROD, "main-End");
	return 0;
}

///\endcond do not document

///@} // end group GrpTestPerformanceDataSet
