////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \defgroup GrpTestPollingCheck TestPollingCheck
/// @{
/// \ingroup GrpTestApps 
/// \brief
/// Win32 console application "TestPollingCheck.exe" demonstrates
/// the usage of class PollingCheck which allows to wait until a condition
/// comes true.
/// 
/// <b>\n Log: Run with successful checks</b>
///
/// \code
/// 
/// --- free boolean check function without arguments  ---
/// VerifyCounterIsZero succeeded!
/// 
/// <VerifyCounterIsZero  [polling]
/// 
/// <VerifyCounterIsZero  [polling]
/// 
/// --- free boolean check function using arguments  ---
/// 
/// <VerifyCounterHasValue 96 [polling]
/// 
/// --- boolean member function without argument  ---
/// 
/// <CounterOfTestObjectIsZero  [polling]
/// 
/// --- boolean member function with single argument  ---
/// 
/// <CounterOfTestObjectHasValue -3 [polling]
/// 
/// 
/// TEST SUCCEEDED!
/// \endcode
///
/// <b>\n Log: Run with failed checks</b>
/// \code
///
/// --- free boolean check function without arguments  ---
/// VerifyCounterIsZero failed!
/// INFO: Poll-Timeout after 516 ms
/// POLL-ERROR: verification failed
/// 
/// <VerifyCounterIsZero  [polling]
/// INFO: Poll-Timeout after 516 ms
/// c:\userdata\gerald\sw\c\testtoolbox\test\testpollingcheck\testpollingcheck.cpp (257): POLL-ERROR: verification failed
/// 
/// <VerifyCounterIsZero  [polling]
/// INFO: Poll-Timeout after 515 ms
/// c:\userdata\gerald\sw\c\testtoolbox\test\testpollingcheck\testpollingcheck.cpp (262): POLL-ERROR: verification failed
/// 
/// --- free boolean check function using arguments  ---
/// INFO: Poll-Timeout after 516 ms
/// POLL-ERROR: verification failed
/// 
/// <VerifyCounterHasValue 96 [polling]
/// INFO: Poll-Timeout after 516 ms
/// c:\userdata\gerald\sw\c\testtoolbox\test\testpollingcheck\testpollingcheck.cpp (272): POLL-ERROR: verification failed
/// INFO: Poll-Timeout after 500 ms
/// POLL-ERROR: verification failed
/// 
/// --- boolean member function without argument  ---
/// 
/// <CounterOfTestObjectIsZero  [polling]
/// INFO: Poll-Timeout after 515 ms
/// c:\userdata\gerald\sw\c\testtoolbox\test\testpollingcheck\testpollingcheck.cpp (191): POLL-ERROR: verification failed
/// 
/// --- boolean member function with single argument  ---
/// 
/// <CounterOfTestObjectHasValue -3 [polling]
/// INFO: Poll-Timeout after 516 ms
/// c:\userdata\gerald\sw\c\testtoolbox\test\testpollingcheck\testpollingcheck.cpp (202): POLL-ERROR: verification failed
/// 
/// 
/// TEST FAILED (8 errors detected)
/// 
/// \endcode
///
/// \sa TestPollingCheck.cpp (file documentation)
/// \sa TestToolBox::PollingCheck (class documentation)
/// \sa GrpPollingCheck (usage documentation)
///
////////////////////////////////////////////////////////////////////////////////


#include "TestToolBox\PollingCheckImpl.h"
#include "TestToolBox\EventReceiverStdOut.h"

#include <tchar.h>
#include <iostream>
#include <boost/bind.hpp>

/// \cond

namespace TTB = TestToolBox;
using namespace std;

// With activated define SHOW_ERROR_REACTIONS several error reactions are reported
// #define SHOW_ERROR_REACTIONS

// Define name of this component for simple trace utility and set initial trace level
TTB_TRC_DEF_GLOBALS ("TestPollingCheck", TL_VAL_DEBUG);


//************************************************************************
//************************************************************************


// Simple test object with method GetCounter().
// A real world application may have several threads and an expected state change
// will occure asynchronously within a separate thread. For simplicity multithreading
// is here omitted and replaced by a simple decrementing schema.
class SomeTestObject
{
public:
    int GetCounter (void)
    {
        TRCF("SomeTestObject::GetCounter");
#ifndef SHOW_ERROR_REACTIONS
        --m_counter;
#endif
        TRC(TL_DEBUG, "m_counter=%d", m_counter);
        return m_counter;
    }

    void SetCounter (int in_counter)
    {m_counter = in_counter;}

private:
    int m_counter;
};

static SomeTestObject* s_pSomeTestObject;


// Free verification functions

bool VerifyCounterIsZero ()
{return s_pSomeTestObject->GetCounter() == 0;}

bool VerifyCounterHasValue (
    int             in_value)
{return s_pSomeTestObject->GetCounter() == in_value;}


//************************************************************************
//************************************************************************


// Demo class which runs a test of SomeTestObject
class MyTestExecutor
{
public:

    MyTestExecutor()
    {m_pTestObject = new SomeTestObject;}

    ~MyTestExecutor()
    {delete m_pTestObject;}

    void RunTestCases (void);

private:
    SomeTestObject* m_pTestObject;

private:
    bool CounterOfTestObjectHasValue (int in_value)
    {return m_pTestObject->GetCounter() == in_value;}

    bool CounterOfTestObjectIsZero (void)
    {return CounterOfTestObjectHasValue(0);}

}; // class MyTestExecutor

//************************************************************************

void MyTestExecutor::RunTestCases(void)
{
    m_pTestObject->SetCounter(4);

    TTB::ThePollingCheck()->SetSleepTimeMs(100);
    TTB::ThePollingCheck()->SetTimeoutMs(500);

    //-----  boolean member function without argument  -----
    cout << "\n--- boolean member function without argument  ---\n";

    POLL_VERIFY(MyTestExecutor::CounterOfTestObjectIsZero);
    

    //-----  boolean member function with single argument  -----
    cout << "\n--- boolean member function with single argument  ---\n";

    // Checking with assert
    // assert(TTB::ThePollingCheck()->Run(boost::bind(&MyTestExecutor::CounterOfTestObjectHasValue,this,51)));

    // Builtin check with macro which enables file/line information for
    // automatic jump to source code location in case of error
    POLL_VERIFY1(MyTestExecutor::CounterOfTestObjectHasValue,-3);

} // MyTestExecutor::RunTestCases


//************************************************************************
//************************************************************************


int _tmain(int argc, _TCHAR* argv[])
{
    TTB::SetTraceLevelFromCommandLine();

    // Create simple object used for testing
    s_pSomeTestObject = new SomeTestObject;

    // Configure timeouts used for polling
    TTB::ThePollingCheck()->SetSleepTimeMs(100);
    TTB::ThePollingCheck()->SetTimeoutMs(500);


    //-----  Free boolean check function without arguments  -----
    cout << "\n--- free boolean check function without arguments  ---\n";

    // simple check with if
    s_pSomeTestObject->SetCounter(3);
    if (TTB::ThePollingCheck()->Run(&VerifyCounterIsZero))
    {
        cout << "VerifyCounterIsZero succeeded!" << endl;
    }
    else
    {
        cout << "VerifyCounterIsZero failed!" << endl;
    }

    // Check with assert
    s_pSomeTestObject->SetCounter(4);
    // assert(TTB::ThePollingCheck()->Run(&VerifyCounterIsZero));

    // Check with Boost verification macro
    // BOOST_CHECK(TTB::ThePollingCheck()->Run(&VerifyCounterIsZero));

    // Test events shall be written to stdout
    TTB::EventReceiverStdOut eventReceiver;
    TTB::ThePollingCheck()->SetEventReceiver(&eventReceiver);
    // optional: configure amount of infos writen to stdout
    // eventReceiver.SetOutputFilter(TTB::CTX_ONLY_ERROR);

    // in case of error a message is written to stdout
    s_pSomeTestObject->SetCounter(2);
    TTB::ThePollingCheck()->Run(&VerifyCounterIsZero);

    // same but additional info about function name and file location
    // which enables automatic jump to source code location in case of error
    s_pSomeTestObject->SetCounter(3);
    TTB::ThePollingCheck()->Run(&VerifyCounterIsZero,"VerifyCounterIsZero","",__FILE__,__LINE__);

    // Builtin check with macro (output of called function and possible error
    // results via EventReceiver)
    s_pSomeTestObject->SetCounter(4);
    POLL_VERIFY_F(VerifyCounterIsZero);


    //-----  Free bool function with arguments -----
    cout << "\n--- free boolean check function using arguments  ---\n";

    s_pSomeTestObject->SetCounter(100);
    TTB::ThePollingCheck()->Run(boost::bind(&VerifyCounterHasValue,98));

    // Use convenience macro for free function with single argument
    POLL_VERIFY_F1(VerifyCounterHasValue, 96);

    // directly using the member function to check
    TTB::ThePollingCheck()->Run(boost::bind(
        &SomeTestObject::GetCounter,
        s_pSomeTestObject)==94);


    //-----  Assuming test cases are executed within an execution class  -----

    MyTestExecutor executor;
    executor.RunTestCases();


    //-----  Checking for total test result  -----

    // Check total result of the test
    if (eventReceiver.GetNumErrors() >0)
    {
        cout << "\n\nTEST FAILED (" << eventReceiver.GetNumErrors() << " errors detected)\n";
    }
    else
    {
        cout << "\n\nTEST SUCCEEDED!\n";
    }

    // return test result as exit code
    return eventReceiver.GetNumErrors();
}

///\endcond

///@} // end group GrpTestPollingCheck
