////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \defgroup GrpTestEnvWithMultipleDlls TestEnvWithMultipleDlls
/// @{
/// \ingroup GrpTestApps 
/// \brief
/// Win32 console application "TestEnvWithMultipleDlls.exe" demonstrates
/// how to use the test utilities also from one or more user DLLs.
/// 
/// <b>\n Short description:</b>
///
/// To use TestUtilities within DLLs
/// - generate TestInstanceDLL and/or make your test project dependent of it
/// - Set define USE_TEST_INSTANCE_DLL within all your test projects
///
/// <b>\n Run-Log:</b>
///
/// \code
/// 
/// _tmain-Begin
/// Creating Environment
/// Environment=003B65B0
/// GetSourceDir=c:\userdata\gerald\sw\c\testtoolbox
/// GetOutputDir=C:\My Target\debug
/// Change path settings of environment
///
/// Now calling AccessEnvironmentWithinExampleDll
/// AccessEnvironmentWithinExampleDll-Begin
/// Environment=003B65B0
/// GetSourceDir=Y:/Path/Was/Set/ByMainFunction
/// GetOutputDir=Z:/AlsoSetFromMain
/// Change path settings of environment
/// AccessEnvironmentWithinExampleDll-End
///
/// AccessEnvironmentWithinExampleDll() returned, now check again path settings
/// Environment=003B65B0
/// GetSourceDir=Q:/SomePathSetFromExampleDll
/// GetOutputDir=R:/AnotherPathSetFromExampleDll
///
/// Now calling AccessTestEventsWithinExampleDll
/// AccessTestEventsWithinExampleDll-Begin
/// INFO: List of stored TestEvents:
/// ?? Something happened in main
/// ?? Also in Dll an event occured
///
/// AccessTestEventsWithinExampleDll-End
/// _tmain-End
///
/// \endcode
///
/// \sa TestEnvWithMultipleDlls.cpp (file documentation)
/// \sa GrpTestInstanceDll (detailed usage documentation)
///
////////////////////////////////////////////////////////////////////////////////


#include "stdafx.h"


// Example DLL showing interaction between main program and DLL
#include "TestToolBox\Test\TestWithDlls\ExampleDll/ExampleDll.h"

// Test utilities used within this test application
// The listed utilities are usable in all DLLs of a test application
// (the instance is stored automatically within TestInstanceDLL
// if preprocessor define USE_TTB_INSTANCE_DLL is defined for all DLLs
// and the main application)
#include "TestToolBox\EnvironmentImpl.h"
#include "TestToolBox\EventReceiverStdOut.h"
#include "TestToolBox\ProtocolImpl.h"
#include "TestToolBox\SynchronizerImpl.h"
#include "TestToolBox\TestEventsImpl.h"
#include "TestToolBox\SimulatedObjectBaseImpl.h"

#include <assert.h>

/// \cond

namespace TTB = TestToolBox;
using namespace std;

// Define name of this component for simple trace utility and set initial trace level
TTB_TRC_DEF_GLOBALS ("TestEnvWithMultipleDlls", TL_VAL_DEBUG);


//*****************************************************************************
//*****************************************************************************


int _tmain(int argc, _TCHAR* argv[])
{

    cout << "\n_tmain-Begin\n";
    TRCF("_tmain");
    TRC(TL_DEBUG,"Begin");

    //----- Access to TestToolBox::Environment -----

    cout << "Creating Environment\n";
    TTB::Environment* pEnvironment = TTB::TheEnvironment();

    cout << "Environment="  << pEnvironment << endl;
    cout << "GetSourceDir=" << pEnvironment->GetSourceDir() << endl;  
    cout << "GetOutputDir=" << pEnvironment->GetOutputDir() << endl;

    cout << "Change path settings of environment" << endl;
    pEnvironment->SetSourceDir("Y:/Path/Was/Set/ByMainFunction");
    pEnvironment->SetOutputDir("Z:/AlsoSetFromMain");

    cout << "\nNow calling AccessEnvironmentWithinExampleDll" << endl;
    AccessEnvironmentWithinExampleDll();

    cout << "\nAccessEnvironmentWithinExampleDll() returned, now check again path settings" << endl;
    cout << "Environment="  << pEnvironment << endl;
    cout << "GetSourceDir=" << pEnvironment->GetSourceDir() << endl;  
    cout << "GetOutputDir=" << pEnvironment->GetOutputDir() << endl;
    assert(pEnvironment->GetSourceDir() == "Q:/SomePathSetFromExampleDll");
    assert(pEnvironment->GetOutputDir() == "R:/AnotherPathSetFromExampleDll");

    pEnvironment->SetOutputDir("C:/My Target/Debug");

    //----- Access to TestToolBox::TestEvents/Synchronizer -----

    // Prepare environment for testEvents and Synchronizer
    //TTB::EventReceiverStdOut rcvStdOut;
    //TTB::Synchronizer::Get()->ConnectWithEventReceiver(&rcvStdOut);
    //TTB::TestEvents* pEvents = TTB::TestEvents::Get();
    //pEvents->ConnectWithEventReceiver(&rcvStdOut);

    TTB::TestEvents* pEvents = TTB::TestEvents::Get();
    pEvents->ConnectWithProtocol(TTB::Protocol::Get());
    pEvents->SetDetailedLog(true);
    TTB::Synchronizer::Get()->ConnectWithEventReceiver(TTB::TestEvents::Get());

    pEvents->Act("Something happened in main");

    // Prepare to receive 1 sync event
    TTB_SET_SYNC_TIMEOUT(5000);
  	TTB_INIT_SYNC(1);

    TTB::TheGlobalOptions()->Set("SomeOptionForDll");

    cout << "\nNow calling AccessTestEventsWithinExampleDll" << endl;
    AccessTestEventsWithinExampleDll();

    TTB_WAIT_SYNC(); // SyncEvent is created within called DLL
    TTB_EXP("Something happened in main");
    TTB_EXP("Also in Dll an event occured");
    TTB_EXP("SomeOptionForDll is set: true");
    TTB_EXP("Dll will now rise a sync event...");

    TTB_CHECK_FOR_UNEXPECTED_EVENTS();

    TTB::TestEvents::Get()->ListAllStoredEvents();

    int retCode = TTB::TestEvents::Get()->AllTestsHaveSucceeded() ? 0 : 1;

    cout << "_tmain-End\n";
    TRC(TL_DEBUG, "End retCode=%d", retCode);

    // Cleanup to avoid memory leaks, force closing of protocol if not yet done
    TTB::Synchronizer::Cleanup();
    TTB::TestEvents::Cleanup();
    TTB::Protocol::Cleanup();
    TTB::TestTime::Cleanup();
    TTB::Environment::Cleanup();
    TTB::Formatter::Cleanup();
    TTB::GlobalOptions::Cleanup();
    TTB::RegisteredSimulatedObjects::Cleanup();

    return retCode;

} // _tmain


//*****************************************************************************
//*****************************************************************************


///\endcond

///@} // end group GrpTestEnvWithMultipleDlls
