////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \defgroup GrpTestXmlCheckWithBoost TestXmlCheckWithBoost
/// @{
/// \ingroup GrpTestApps 
/// \brief
/// Win32 console application "TestXmlCheckWithBoost.exe" demonstrates
/// the running of a unit test with Boost.Test lib and performing of XML checks
/// using TestToolBox::XmlCheck.
/// 
/// <b>\n Supported features</b>
///
///  - check of XML trees and subtrees
///
///  - support for XPath navigation
///
///  - skipping specific attributes/nodes
///
///  - treating float attributes/nodes with configurable precision
///
/// <b>Depends on</b>
///
///  - TinyXpath (includes TinyXml)
///
///  - Boost.Test lib
///
///  - TestToolBox::XmlCheck
///
/// <b>Typical code snippets</b>
///
///  - comparing complete XML trees
///    <pre>
///
///    BOOST_TEST_MESSAGE("\nComparing two XML trees");
///    BOOST_CHECK(TTB::TheXmlCheck()->Compare(&docA,&docB));
///    BOOST_TEST_MESSAGE(TTB::TheXmlCheck()->GetCompareResult());
///    </pre>
///
///  - comparing XML subtrees and ignoring attributes/nodes
///    <pre>
///
///    BOOST_TEST_MESSAGE("\nComparing two XML subtrees using XPath, ignoring time");
///    TTB::TheXmlCheck()->IgnoreAttribute("time");
///    BOOST_CHECK(TTB::TheXmlCheck()->Compare(
///    TTB::XmlCheck::SubNodeFromXPath(docA.RootElement(), "/MyContainer/Group_B"),
///    TTB::XmlCheck::SubNodeFromXPath(docB.RootElement(), "/MyContainer/Group_B")));
///    BOOST_TEST_MESSAGE(TTB::TheXmlCheck()->GetCompareResult());
///    </pre>
///
///
/// <b>Project settings</b>
///
///     - C/C++ \ additional includes (for TestToolBox, TinyXPath, boost)<br>
///       <code>../../..;C:/;C:/boost/boost_actual</code>
///
///     - Linker \ Input<br>
///       <code>tinyxpathd.lib</code>
///
///     - Linker \ additional lib dirs (for TinyXPath, boost)<br>
///       <code>C:/TinyXPath/tinyxpath_lib/Debug;C:/boost/boost_actual/lib;</code>
///
///     - Linker \ Command line<br>
///       <code>/NODEFAULTLIB:LIBCMTD.lib</code>
///
/// <b>Hints for execution</b>
///
/// For better error detection set command line option "--log_level=test_suite"
///
/// <b>\n Log: Detecting some differences when comparing XML trees/subtrees</b>
///
/// \code
///
/// Running 3 test cases...
/// Entering test suite "Master Test Suite"
/// Entering test case "TestCheckNode"
///
/// CheckNode does not produce any output
/// see also projects TestXmlCheckBasic,TestXmlCheckWithBoostExtensions
/// for more productive examples using EventReceiver concept
/// Test case TestCheckNode doesn't include any assertions
/// Leaving test case "TestCheckNode"
/// Entering test case "TestCompare"
///
/// Comparing two XML trees
/// c:/userdata/gerald/sw/c/testtoolbox/test/testxmlcheckwithboost/testxmlcheckwithboost.cpp(193): error in "TestCompare": check TTB::TheXmlCheck()->Compare(&docA,&docB) failed
/// Difference detected
/// A: time: 09:15:12.234 (line: 3 col: 34, path: //MyContainer/Group_A/Vector)
/// B: time: 09:30:12.234 (line: 5 col: 34, path: //MyContainer/Group_A/Vector)
///
/// Comparing two XML trees without differences
/// XML-Compare: no differences found
///
/// Comparing two XML trees without differences - now suppressing OK message
///
/// Comparing two XML trees, ignoring time
/// c:/userdata/gerald/sw/c/testtoolbox/test/testxmlcheckwithboost/testxmlcheckwithboost.cpp(211): error in "TestCompare": check TTB::TheXmlCheck()->Compare(&docA,&docB) failed
/// Difference detected
/// A: tolerance: 0.0258 (line: 7 col: 54, path: //MyContainer/Group_A/Vector)
/// B: tolerance: 0.8258 (line: 9 col: 54, path: //MyContainer/Group_A/Vector)
///
/// Comparing two XML subtrees using XPath
/// c:/userdata/gerald/sw/c/testtoolbox/test/testxmlcheckwithboost/testxmlcheckwithboost.cpp(217): error in "TestCompare": check TTB::TheXmlCheck()->Compare( TTB::XmlCheck::SubNodeFromXPath(docA.RootElement(), "/MyContainer/Group_B"), TTB::XmlCheck::SubNodeFromXPath(docB.RootElement(), "/MyContainer/Group_B")) failed
/// Difference detected
/// A: 3.132437 (line: 18 col: 8, path: //MyContainer/Group_B/x/3.132437/)
/// B: 12.132437 (line: 20 col: 8, path: //MyContainer/Group_B/x/12.132437/)
///
/// For more test cases see also projects TestXmlCheckBasic,TestXmlCheckWithExtensionsForBoost
/// Leaving test case "TestCompare"; testing time: 31ms
/// Entering test case "Cleanup"
/// Test case Cleanup doesn't include any assertions
/// Leaving test case "Cleanup"
/// Leaving test suite "Master Test Suite"
///
/// *** 3 failures detected in test suite "Master Test Suite"
///
/// \endcode
///
/// \sa TestXmlCheckWithBoost.cpp (file documentation)
/// \sa GrpXmlCheck (usage documentation)
///
////////////////////////////////////////////////////////////////////////////////

///\cond do not document

#include "stdafx.h"
#include "TinyXPath\TinyXml.h"
#pragma warning(push)
#pragma warning(disable:4267)
#pragma warning(disable:4996)
#pragma warning(disable:4100)
#include "TinyXPath\xpath_static.h"
#pragma warning(pop) 
#include "TestToolBox\XmlCheckImpl.h"
#include <iostream>

namespace TTB = TestToolBox;
using namespace std;

#include <boost/bind.hpp>


// Test cases shoild be automatically executed by Boost framework
//#define BOOST_AUTO_TEST_MAIN (here defined within project settings)
#include <boost/test/auto_unit_test.hpp>

//*** Use existing boost test framework implementation ***

// Possibility 1:
// using test framework by explicitly linking with boost framework library
// Important: use the library version which is built with the same C/C++ runtime
// as the test executable is built with
//   mt : multithreaded
//    s : static runtime
//    d : debug build
//    g : debug/diagnostic runtime
//#pragma comment(lib, "boost_unit_test_framework-vc80-mt-gd-1_36")

// Possibility 2:
// Use the boost auto link feature which automatically chooses the correct library version
#define BOOST_LIB_NAME boost_unit_test_framework
#include <boost/config/auto_link.hpp>

// Possibility 3:
// instead of linking with the unit test library you can compile all required sources
// through the following include;
// disadvantage: longer compilation time
// #include <boost/test/included/unit_test_framework.hpp>




using boost::unit_test::test_suite;


//************************************************************************
//************************************************************************


static const char* DEMO_XML_STRING =
    "<MyContainer>\n"
    "  <Group_A>\n"
    "    <Vector id=1 date=\"01.07.09\" time=\"09:15:12.234\" tolerance=0.0258>\n"
    "      This is the text of the first vector element\n"
    "      <x>3.14235</x>\n"
    "    </Vector>\n"
    "    <Vector id=2 date=\"01.07.09\" time=\"09:15:12.234\" tolerance=0.0258>\n"
    "      This is the text of the second vector element\n"
    "      <x>3.14235</x>\n"
    "    </Vector>\n"
    "    <Vector id=3 tolerance=1.357>\n"
    "      <x>6.28532</x>\n"
    "    </Vector>\n"
    "  </Group_A>\n"
    "  <Group_B>\n"
    "    <x>1.23456</x>\n"
    "    <x>2.3581</x>\n"
    "    <x>3.132437</x>\n"
    "    <x>4.987654</x>\n"
    "    <x>5.29741</x>\n"
    "  </Group_B>\n"
    "  <Group_C>\n"
    "    <Interior>\n"
    "      <Something>Some text</Something>\n"
    "      <date>28.09.09</date>\n"
    "    </Interior>\n"
    "  </Group_C>\n"
    "</MyContainer>\n";

// Group_A\Vector\time, Group_A\Vector\tolerance, Group_A\Vector\x and
// Group_B\ third x is changed
static const char* DEMO_XML_STRING_FOR_COMPARE =
    "<MyContainer>\n"
    "  <Group_A>\n\n\n"
    "    <Vector id=1 date=\"01.07.09\" time=\"09:30:12.234\" tolerance=0.0258>\n"
    "      This is the text of the first vector element\n"
    "      <x>3.14235</x>\n"
    "    </Vector>\n"
    "    <Vector id=2 date=\"01.07.09\" time=\"09:15:12.234\" tolerance=0.8258>\n"
    "      This is the text of the second vector element\n"
    "      <x>3.14235</x>\n"
    "    </Vector>\n"
    "    <Vector id=3 tolerance=1.357>\n"
    "      <x>6.28532</x>\n"
    "    </Vector>\n"
    "  </Group_A>\n"
    "  <Group_B>\n"
    "    <x>1.23456</x>\n"
    "    <x>2.3581</x>\n"
    "    <x>12.132437</x>\n"
    "    <x>4.987654</x>\n"
    "    <x>5.29741</x>\n"
    "  </Group_B>\n"
    "  <Group_C>\n"
    "    <Interior>\n"
    "      <Something>Some text</Something>\n"
    "      <date>28.09.09</date>\n"
    "    </Interior>\n"
    "  </Group_C>\n"
    "</MyContainer>\n";


//************************************************************************
//************************************************************************

class MyFixture
{
	// do soemthin within constructor/ddestructor
};

BOOST_FIXTURE_TEST_SUITE(MyFixtureTestSuite, MyFixture)

//BOOST_AUTO_TEST_CASE_TEMPLATE(SomeTemplateTestCase,T, asdf)
//{
//}

BOOST_AUTO_TEST_CASE (TestCheckNode)
{
    TiXmlDocument doc;
    doc.Parse(DEMO_XML_STRING);

    BOOST_TEST_MESSAGE ("\nCheckNode does not produce any output\n"
        "see also projects TestXmlCheckBasic,TestXmlCheckWithBoostExtensions\n"
        "for more productive examples using EventReceiver concept");
    TTB::TheXmlCheck()->CheckNode(&doc);

} // TestCheckNode

BOOST_AUTO_TEST_SUITE_END();

//************************************************************************
//************************************************************************
BOOST_AUTO_TEST_SUITE(SomeOtherTestSuite)
BOOST_AUTO_TEST_CASE(DoSomething)
{

}   
BOOST_AUTO_TEST_SUITE(NestedSuite)

BOOST_AUTO_TEST_CASE(DoSomethingElse)
{

}
BOOST_AUTO_TEST_SUITE_END();

BOOST_AUTO_TEST_SUITE_END();


BOOST_AUTO_TEST_SUITE(MyAutoTestSuite)

BOOST_AUTO_TEST_CASE (TestCompare)
{
    TiXmlDocument doc;
    doc.Parse(DEMO_XML_STRING);
//assert(!"HALLOOOOOOO!");
    TiXmlDocument docA;
    docA.Parse(DEMO_XML_STRING);

    TiXmlDocument docB;
    docB.Parse(DEMO_XML_STRING_FOR_COMPARE);

	BOOST_TEST_MESSAGE("\nComparing two XML trees");
    BOOST_CHECK(TTB::TheXmlCheck()->Compare(&docA,&docB));
    BOOST_TEST_MESSAGE(TTB::TheXmlCheck()->GetCompareResult());

    BOOST_TEST_MESSAGE("\nComparing two XML trees without differences");
    BOOST_CHECK(TTB::TheXmlCheck()->Compare(&docA,&docA));
    BOOST_TEST_MESSAGE(TTB::TheXmlCheck()->GetCompareResult());

    BOOST_TEST_MESSAGE("\nComparing two XML trees without differences"
                  " - now suppressing OK message");
    bool checkOk = TTB::TheXmlCheck()->Compare(&docA,&docA);
    BOOST_CHECK(checkOk);
    if (!checkOk)
    {
        BOOST_TEST_MESSAGE(TTB::TheXmlCheck()->GetCompareResult());
    }

    BOOST_TEST_MESSAGE("\nComparing two XML trees, ignoring time");
    TTB::TheXmlCheck()->IgnoreAttribute("time");
    BOOST_CHECK(TTB::TheXmlCheck()->Compare(&docA,&docB));
    BOOST_TEST_MESSAGE(TTB::TheXmlCheck()->GetCompareResult());

    BOOST_TEST_MESSAGE("\nComparing two XML subtrees using XPath");
    BOOST_CHECK(TTB::TheXmlCheck()->Compare(
        TTB::XmlCheck::SubNodeFromXPath(docA.RootElement(), "/MyContainer/Group_B"),
        TTB::XmlCheck::SubNodeFromXPath(docB.RootElement(), "/MyContainer/Group_B")));
    BOOST_TEST_MESSAGE(TTB::TheXmlCheck()->GetCompareResult());

    BOOST_TEST_MESSAGE("\nFor more test cases see also projects TestXmlCheckBasic,TestXmlCheckWithExtensionsForBoost");

} // TestCompare

BOOST_AUTO_TEST_SUITE_END();

//************************************************************************
//************************************************************************


//BOsOST_AUTO_TEST_CASE (Cleanup)
//{
//    TTB::XmlCheck::Cleanup();
//
//} // Cleanup


//************************************************************************
//************************************************************************

///\endcond

///@} // end group GrpTestApps
