////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \defgroup GrpTestXmlCheckWithBoostExtensions TestXmlCheckWithBoostExtensions
/// @{
/// \ingroup GrpTestApps 
/// \brief
/// Win32 console application "TestXmlCheckWithBoostExtensions.exe" demonstrates
/// the running of a unit test with Boost.Test lib and performing of XML checks
/// using TestToolBox::XmlCheck, TestToolBox::TestEvents.
/// 
/// <b>\n Supported features</b>
///
///  - check of XML trees and subtrees
///
///  - support for XPath navigation
///
///  - skipping specific attributes/nodes
///
///  - treating float attributes/nodes with configurable precision
///
/// <b>Depends on</b>
///
///  - TinyXpath (includes TinyXml)
///
///  - Boost.Test lib
///
///  - TestToolBox::XmlCheck, TestToolBox::TestEvents
///
/// <b>Typical code snippets</b>
///
///  - checking XML contents, ignoring some attributes/nodes
///    <pre>
///
///    TTB::TheXmlCheck()->IgnoreAttribute("date");
///    TTB::TheXmlCheck()->IgnoreAttribute("time");
///    TTB::TheXmlCheck()->IgnoreNode("date");
///    TTB::TheXmlCheck()->UseAsDoubleAttribute("tolerance", true, 3);
///    TTB::TheXmlCheck()->UseAsDoubleNode("x", true, 4);
///    TTB::TheXmlCheck()-><strong>CheckNode(&doc)</strong>;
///
///    TTB_EXP("MyContainer");
///    TTB_EXP("  Group_A");
///    TTB_EXP("    Vector");
///    TTB_EXP("      id: 1");
///    TTB_EXP("      tolerance: 0.026");
///    TTB_EXP("      This is the text of the first vector element");
///    TTB_EXP("      x");
///    TTB_EXP("        3.1424");
///    TTB_EXP("    Vector");
///    TTB_EXP("      id: 2");
///    TTB_EXP("      tolerance: 0.026");
///    TTB_EXP("      This is the text of the second vector element");
///    TTB_EXP("      x");
///    TTB_EXP("        1.3264");
///    </pre>
///
///  - checking XML contents, accessing subtree using XPath
///    <pre>
///
///    TTB::TheXmlCheck()-><b>CheckSubNode(doc.RootElement(),""/MyContainer/Group_A/Vector[last()]")</b>;
///
///    TTB_EXP("Vector");
///    TTB_EXP("  id: 3");
///    TTB_EXP("  tolerance: 1.357");
///    TTB_EXP("  x");
///    TTB_EXP("    6.28532");
///    </pre>
///
///  - comparing two XML trees
///    <pre>
///
///    TTB::TheXmlCheck()->IgnoreAttribute("time");
///    TTB::TheXmlCheck()-><b>Compare(&docA,&docB)</b>;
///
///    TTB_EXP("Difference detected");
///    TTB_EXP("A: tolerance: 0.0258 (line: 7 col: 54, path: //MyContainer/Group_A/Vector)");
///    TTB_EXP("B: tolerance: 0.8258 (line: 9 col: 54, path: //MyContainer/Group_A/Vector)");
///
///
///    TTB::TheXmlCheck()->Compare(
///        TTB::XmlCheck::<b>SubNodeFromXPath</b>(docA.RootElement(), "/MyContainer/Group_B"),
///        TTB::XmlCheck::<b>SubNodeFromXPath</b>(docB.RootElement(), "/MyContainer/Group_B"));
///
///    TTB_EXP("Difference detected");
///    TTB_EXP("A: 3.132437 (line: 18 col: 8, path: //MyContainer/Group_B/x/3.132437/)");
///    TTB_EXP("B: 12.132437 (line: 20 col: 8, path: //MyContainer/Group_B/x/12.132437/)");
///    </pre>
///
///
/// <b>Project settings</b>
///
///     - C/C++ \ additional includes (for TestToolBox, TinyXPath, boost)<br>
///       <code>../../..;C:/;C:/boost/boost_actual</code>
///
///     - C/C++ \ preprocessor<br>
///       BOOST_AUTO_TEST_MAIN<br>
///       (Test cases should be automatically executed by Boost framework)
///
///     - Linker \ Input<br>
///       <code>tinyxpathd.lib</code>
///
///     - Linker \ additional lib dirs (for TinyXPath, boost)<br>
///       <code>C:/TinyXPath/tinyxpath_lib/Debug;C:/boost/boost_actual/lib;</code>
///
///     - Linker \ Command line<br>
///       <code>/NODEFAULTLIB:LIBCMTD.lib</code>
///
/// <b>Hints for execution</b>
///
/// - set command line option "--log_level=test_suite" to activate output of BOOST_TEST
///
/// - to watch test eventrs in the moment of their creation activate logging
///   via <code>TTB::TestEvents::Get()->LogCallsToAct(true);</code>
///
///
/// <b>\n Log: Running without "--log_level=test_suite"</b>
///
/// \code
///
/// Running 4 test cases...
///
/// *** No errors detected
///
/// \endcode
///
///
/// <b>\n Log: Running with "--log_level=test_suite" and default setting "LogCallsToAct(false)"</b>
///
/// \code
/// Running 4 test cases...
/// Entering test suite "Master Test Suite"
/// Entering test case "SetupTestEnv"
/// Test case SetupTestEnv doesn't include any assertions
/// Leaving test case "SetupTestEnv"
/// Entering test case "TestCheckNode"
/// 
/// ______________________________________________
/// Testcases for CheckNode
/// ______________________________________________
/// 
/// 
/// ---- TEST_EVENT next section: "Examples how to get an XmlNode from an XML string or an XML file"
/// 
/// now doc and docFromFile should contain the same XML data
/// 
/// ---- TEST_EVENT next section: "CheckNode"
/// 
/// Remark:
/// Within this test environment XmlCheck is connected with TestEvents via XmlCheck::SetEventReceiver(TTB::TestEvents::Get()).
/// CheckNode() writes all entries to TestEvents, but this class doesn't use the context entry.
/// Therefore the context set by XmlCheck()::SetOutputModeForWritingXmlContents is not relevant.
/// 
/// 
/// ---- TEST_EVENT next section: CheckNode - detailed output
/// 
/// 
/// ---- TEST_EVENT next section: "CheckNode - ignore attributes and nodes, define double values with precision"
/// 
/// Ignore attributes 'date' and 'time' and node 'date'
/// Attribute 'tolerance' is a double with precision 3
/// Node 'x' is a double with precision 4
///
/// 
/// ---- TEST_EVENT next section: "CheckNode - ignore complete subtree 'Group_A' "
///
/// Furthermore set precision for 'x' to 1
///
/// 
/// ---- TEST_EVENT next section: "CheckSubNode with XPath - only Group_B (/MyContainer/Group_B)"
/// 
/// 
/// ---- TEST_EVENT next section: "CheckSubNode with XPath - first x element within Group_B (/MyContainer/Group_B/x)"
/// 
/// 
/// ---- TEST_EVENT next section: "CheckSubNode with XPath - third x element within Group_B (/MyContainer/Group_B/x[3])"
/// 
/// 
/// ---- TEST_EVENT next section: "CheckSubNode with XPath - second vector element of Group_A (/MyContainer/Group_A/Vector[2])"
/// 
/// 
/// ---- TEST_EVENT next section: "CheckSubNode with XPath - last vector element of Group_A (/MyContainer/Group_A/Vector[last()])"
/// 
/// 
/// ---- TEST_EVENT next section: "CheckSubNode with XPath - relative to subtree via SubNodeFromXPath (/MyContainer/Group_A, Vector[last())"
/// 
/// 
/// ---- TEST_EVENT check section
/// 
/// Leaving test case "TestCheckNode"; testing time: 31ms
/// Entering test case "TestCompare"
/// 
/// ---- TEST_EVENT next section: "Compare - with/without use of TestEvents)"
/// 
/// Comparing two XML trees without differences (with TestEvents)
/// 
/// Comparing two XML trees without differences (without TestEvents)
/// XML-Compare: no differences found
///
/// Comparing two XML trees without differences - now suppressing OK message
///
/// ---- TEST_EVENT next section: "Compare - detection of first difference"
/// 
/// 
/// ---- TEST_EVENT next section: "Compare - ignoring 'time'"
/// 
/// 
/// ---- TEST_EVENT next section: "Compare - using XPath to compare subtrees (/MyContainer/Group_B)"
/// 
/// 
/// For more test cases see also projects TestXmlCheckBasic,TestXmlCheckWithExtensionsForBoost
/// 
/// ---- TEST_EVENT check section
/// 
/// Leaving test case "TestCompare"; testing time: 47ms
/// Entering test case "Cleanup"
/// Test case Cleanup doesn't include any assertions
/// Leaving test case "Cleanup"
/// Leaving test suite "Master Test Suite"
/// 
/// *** No errors detected
/// 
///
/// \endcode
///
///
/// <b>\n Log: Running with "LogCallsToAct(true)" and "--log_level=test_suite"</b>
///
/// \code
///
/// Running 4 test cases...
/// Entering test suite "Master Test Suite"
/// Entering test case "SetupTestEnv"
/// Test case SetupTestEnv doesn't include any assertions
/// Leaving test case "SetupTestEnv"
/// Entering test case "TestCheckNode"
/// 
/// ______________________________________________
/// Testcases for CheckNode
/// ______________________________________________
/// 
/// 
/// ---- TEST_EVENT next section: "Examples how to get an XmlNode from an XML string or an XML file"
/// 
/// now doc and docFromFile should contain the same XML data
/// 
/// ---- TEST_EVENT next section: "CheckNode"
/// 
/// Remark:
/// Within this test environment XmlCheck is connected with TestEvents via XmlCheck::SetEventReceiver(TTB::TestEvents::Get()).
/// CheckNode() writes all entries to TestEvents, but this class doesn't use the context entry.
/// Therefore the context set by XmlCheck()::SetOutputModeForWritingXmlContents is not relevant.
/// 
/// TEST_EVENT: INFO: Act: TTB_EXP("MyContainer");
/// TEST_EVENT: INFO: Act: TTB_EXP("  Group_A");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Vector");
/// TEST_EVENT: INFO: Act: TTB_EXP("      id: 1");
/// TEST_EVENT: INFO: Act: TTB_EXP("      date: 01.07.09");
/// TEST_EVENT: INFO: Act: TTB_EXP("      time: 09:15:12.234");
/// TEST_EVENT: INFO: Act: TTB_EXP("      tolerance: 0.0258");
/// TEST_EVENT: INFO: Act: TTB_EXP("      This is the text of the first vector element");
/// TEST_EVENT: INFO: Act: TTB_EXP("      x");
/// TEST_EVENT: INFO: Act: TTB_EXP("        3.14235");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Vector");
/// TEST_EVENT: INFO: Act: TTB_EXP("      id: 2");
/// TEST_EVENT: INFO: Act: TTB_EXP("      date: 01.07.09");
/// TEST_EVENT: INFO: Act: TTB_EXP("      time: 09:15:12.234");
/// TEST_EVENT: INFO: Act: TTB_EXP("      tolerance: 0.0258");
/// TEST_EVENT: INFO: Act: TTB_EXP("      This is the text of the second vector element");
/// TEST_EVENT: INFO: Act: TTB_EXP("      x");
/// TEST_EVENT: INFO: Act: TTB_EXP("        3.14235");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Vector");
/// TEST_EVENT: INFO: Act: TTB_EXP("      id: 3");
/// TEST_EVENT: INFO: Act: TTB_EXP("      tolerance: 1.357");
/// TEST_EVENT: INFO: Act: TTB_EXP("      x");
/// TEST_EVENT: INFO: Act: TTB_EXP("        6.28532");
/// TEST_EVENT: INFO: Act: TTB_EXP("  Group_B");
/// TEST_EVENT: INFO: Act: TTB_EXP("    x");
/// TEST_EVENT: INFO: Act: TTB_EXP("      1.23456");
/// TEST_EVENT: INFO: Act: TTB_EXP("    x");
/// TEST_EVENT: INFO: Act: TTB_EXP("      2.3581");
/// TEST_EVENT: INFO: Act: TTB_EXP("    x");
/// TEST_EVENT: INFO: Act: TTB_EXP("      3.132437");
/// TEST_EVENT: INFO: Act: TTB_EXP("    x");
/// TEST_EVENT: INFO: Act: TTB_EXP("      4.987654");
/// TEST_EVENT: INFO: Act: TTB_EXP("    x");
/// TEST_EVENT: INFO: Act: TTB_EXP("      5.29741");
/// TEST_EVENT: INFO: Act: TTB_EXP("  Group_C");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Interior");
/// TEST_EVENT: INFO: Act: TTB_EXP("      Something");
/// TEST_EVENT: INFO: Act: TTB_EXP("        Some text");
/// TEST_EVENT: INFO: Act: TTB_EXP("      date");
/// TEST_EVENT: INFO: Act: TTB_EXP("        28.09.09");
/// 
/// ---- TEST_EVENT next section: CheckNode - detailed output
/// 
/// TEST_EVENT: INFO: Act: TTB_EXP("Document (line: 1, col: 1)");
/// TEST_EVENT: INFO: Act: TTB_EXP("Element: MyContainer (line: 1, col: 1)");
/// TEST_EVENT: INFO: Act: TTB_EXP("  Element: Group_A (line: 2, col: 5)");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Element: Vector (line: 3, col: 9)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      id: 1, as int: 1, as double: 1 (line: 3, col: 17)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      date: 01.07.09, as int: 1, as double: 1.07 (line: 3, col: 24)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      time: 09:15:12.234, as int: 9, as double: 9 (line: 3, col: 40)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      tolerance: 0.0258, as int: 0, as double: 0.0258 (line: 3, col: 60)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      4 attributes");
/// TEST_EVENT: INFO: Act: TTB_EXP("      Text: This is the text of the first vector element (line: 3, col: 79)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      Element: x (line: 4, col: 13)");
/// TEST_EVENT: INFO: Act: TTB_EXP("        Text: 3.14235 (line: 4, col: 16)");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Element: Vector (line: 6, col: 9)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      id: 2, as int: 2, as double: 2 (line: 6, col: 17)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      date: 01.07.09, as int: 1, as double: 1.07 (line: 6, col: 24)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      time: 09:15:12.234, as int: 9, as double: 9 (line: 6, col: 40)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      tolerance: 0.0258, as int: 0, as double: 0.0258 (line: 6, col: 60)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      4 attributes");
/// TEST_EVENT: INFO: Act: TTB_EXP("      Text: This is the text of the second vector element (line: 6, col: 79)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      Element: x (line: 7, col: 13)");
/// TEST_EVENT: INFO: Act: TTB_EXP("        Text: 3.14235 (line: 7, col: 16)");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Element: Vector (line: 9, col: 9)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      id: 3, as int: 3, as double: 3 (line: 9, col: 17)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      tolerance: 1.357, as int: 1, as double: 1.357 (line: 9, col: 24)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      2 attributes");
/// TEST_EVENT: INFO: Act: TTB_EXP("      Element: x (line: 10, col: 13)");
/// TEST_EVENT: INFO: Act: TTB_EXP("        Text: 6.28532 (line: 10, col: 16)");
/// TEST_EVENT: INFO: Act: TTB_EXP("  Element: Group_B (line: 13, col: 5)");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Element: x (line: 14, col: 9)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      Text: 1.23456 (line: 14, col: 12)");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Element: x (line: 15, col: 9)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      Text: 2.3581 (line: 15, col: 12)");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Element: x (line: 16, col: 9)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      Text: 3.132437 (line: 16, col: 12)");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Element: x (line: 17, col: 9)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      Text: 4.987654 (line: 17, col: 12)");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Element: x (line: 18, col: 9)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      Text: 5.29741 (line: 18, col: 12)");
/// TEST_EVENT: INFO: Act: TTB_EXP("  Element: Group_C (line: 20, col: 5)");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Element: Interior (line: 21, col: 9)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      Element: Something (line: 22, col: 13)");
/// TEST_EVENT: INFO: Act: TTB_EXP("        Text: Some text (line: 22, col: 24)");
/// TEST_EVENT: INFO: Act: TTB_EXP("      Element: date (line: 23, col: 13)");
/// TEST_EVENT: INFO: Act: TTB_EXP("        Text: 28.09.09 (line: 23, col: 19)");
/// 
/// ---- TEST_EVENT next section: "CheckNode - ignore attributes and nodes, define double values with precision"
/// 
/// Ignore attributes 'date' and 'time' and node 'date'
/// Attribute 'tolerance' is a double with precision 3
/// Node 'x' is a double with precision 4
/// 
/// TEST_EVENT: INFO: Act: TTB_EXP("MyContainer");
/// TEST_EVENT: INFO: Act: TTB_EXP("  Group_A");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Vector");
/// TEST_EVENT: INFO: Act: TTB_EXP("      id: 1");
/// TEST_EVENT: INFO: Act: TTB_EXP("      tolerance: 0.026");
/// TEST_EVENT: INFO: Act: TTB_EXP("      This is the text of the first vector element");
/// TEST_EVENT: INFO: Act: TTB_EXP("      x");
/// TEST_EVENT: INFO: Act: TTB_EXP("        3.1424");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Vector");
/// TEST_EVENT: INFO: Act: TTB_EXP("      id: 2");
/// TEST_EVENT: INFO: Act: TTB_EXP("      tolerance: 0.026");
/// TEST_EVENT: INFO: Act: TTB_EXP("      This is the text of the second vector element");
/// TEST_EVENT: INFO: Act: TTB_EXP("      x");
/// TEST_EVENT: INFO: Act: TTB_EXP("        3.1424");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Vector");
/// TEST_EVENT: INFO: Act: TTB_EXP("      id: 3");
/// TEST_EVENT: INFO: Act: TTB_EXP("      tolerance: 1.357");
/// TEST_EVENT: INFO: Act: TTB_EXP("      x");
/// TEST_EVENT: INFO: Act: TTB_EXP("        6.2853");
/// TEST_EVENT: INFO: Act: TTB_EXP("  Group_B");
/// TEST_EVENT: INFO: Act: TTB_EXP("    x");
/// TEST_EVENT: INFO: Act: TTB_EXP("      1.2346");
/// TEST_EVENT: INFO: Act: TTB_EXP("    x");
/// TEST_EVENT: INFO: Act: TTB_EXP("      2.3581");
/// TEST_EVENT: INFO: Act: TTB_EXP("    x");
/// TEST_EVENT: INFO: Act: TTB_EXP("      3.1324");
/// TEST_EVENT: INFO: Act: TTB_EXP("    x");
/// TEST_EVENT: INFO: Act: TTB_EXP("      4.9877");
/// TEST_EVENT: INFO: Act: TTB_EXP("    x");
/// TEST_EVENT: INFO: Act: TTB_EXP("      5.2974");
/// TEST_EVENT: INFO: Act: TTB_EXP("  Group_C");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Interior");
/// TEST_EVENT: INFO: Act: TTB_EXP("      Something");
/// TEST_EVENT: INFO: Act: TTB_EXP("        Some text");
/// 
/// ---- TEST_EVENT next section: "CheckNode - ignore complete subtree 'Group_A' "
/// 
/// Furthermore set precision for 'x' to 1
///
/// TEST_EVENT: INFO: Act: TTB_EXP("MyContainer");
/// TEST_EVENT: INFO: Act: TTB_EXP("  Group_B");
/// TEST_EVENT: INFO: Act: TTB_EXP("    x");
/// TEST_EVENT: INFO: Act: TTB_EXP("      1.2");
/// TEST_EVENT: INFO: Act: TTB_EXP("    x");
/// TEST_EVENT: INFO: Act: TTB_EXP("      2.4");
/// TEST_EVENT: INFO: Act: TTB_EXP("    x");
/// TEST_EVENT: INFO: Act: TTB_EXP("      3.1");
/// TEST_EVENT: INFO: Act: TTB_EXP("    x");
/// TEST_EVENT: INFO: Act: TTB_EXP("      5.0");
/// TEST_EVENT: INFO: Act: TTB_EXP("    x");
/// TEST_EVENT: INFO: Act: TTB_EXP("      5.3");
/// TEST_EVENT: INFO: Act: TTB_EXP("  Group_C");
/// TEST_EVENT: INFO: Act: TTB_EXP("    Interior");
/// TEST_EVENT: INFO: Act: TTB_EXP("      Something");
/// TEST_EVENT: INFO: Act: TTB_EXP("        Some text");
/// 
/// ---- TEST_EVENT next section: "CheckSubNode with XPath - only Group_B (/MyContainer/Group_B)"
/// 
/// TEST_EVENT: INFO: Act: TTB_EXP("Group_B");
/// TEST_EVENT: INFO: Act: TTB_EXP("  x");
/// TEST_EVENT: INFO: Act: TTB_EXP("    1.23456");
/// TEST_EVENT: INFO: Act: TTB_EXP("  x");
/// TEST_EVENT: INFO: Act: TTB_EXP("    2.3581");
/// TEST_EVENT: INFO: Act: TTB_EXP("  x");
/// TEST_EVENT: INFO: Act: TTB_EXP("    3.132437");
/// TEST_EVENT: INFO: Act: TTB_EXP("  x");
/// TEST_EVENT: INFO: Act: TTB_EXP("    4.987654");
/// TEST_EVENT: INFO: Act: TTB_EXP("  x");
/// TEST_EVENT: INFO: Act: TTB_EXP("    5.29741");
/// 
/// ---- TEST_EVENT next section: "CheckSubNode with XPath - first x element within Group_B (/MyContainer/Group_B/x)"
/// 
/// TEST_EVENT: INFO: Act: TTB_EXP("x");
/// TEST_EVENT: INFO: Act: TTB_EXP("  1.23456");
/// 
/// ---- TEST_EVENT next section: "CheckSubNode with XPath - third x element within Group_B (/MyContainer/Group_B/x[3])"
/// 
/// TEST_EVENT: INFO: Act: TTB_EXP("x");
/// TEST_EVENT: INFO: Act: TTB_EXP("  3.132437");
/// 
/// ---- TEST_EVENT next section: "CheckSubNode with XPath - second vector element of Group_A (/MyContainer/Group_A/Vector[2])"
/// 
/// TEST_EVENT: INFO: Act: TTB_EXP("Vector");
/// TEST_EVENT: INFO: Act: TTB_EXP("  id: 2");
/// TEST_EVENT: INFO: Act: TTB_EXP("  date: 01.07.09");
/// TEST_EVENT: INFO: Act: TTB_EXP("  time: 09:15:12.234");
/// TEST_EVENT: INFO: Act: TTB_EXP("  tolerance: 0.0258");
/// TEST_EVENT: INFO: Act: TTB_EXP("  This is the text of the second vector element");
/// TEST_EVENT: INFO: Act: TTB_EXP("  x");
/// TEST_EVENT: INFO: Act: TTB_EXP("    3.14235");
/// 
/// ---- TEST_EVENT next section: "CheckSubNode with XPath - last vector element of Group_A (/MyContainer/Group_A/Vector[last()])"
/// 
/// TEST_EVENT: INFO: Act: TTB_EXP("Vector");
/// TEST_EVENT: INFO: Act: TTB_EXP("  id: 3");
/// TEST_EVENT: INFO: Act: TTB_EXP("  tolerance: 1.357");
/// TEST_EVENT: INFO: Act: TTB_EXP("  x");
/// TEST_EVENT: INFO: Act: TTB_EXP("    6.28532");
/// 
/// ---- TEST_EVENT next section: "CheckSubNode with XPath - relative to subtree via SubNodeFromXPath (/MyContainer/Group_A, Vector[last())"
/// 
/// TEST_EVENT: INFO: Act: TTB_EXP("Vector");
/// TEST_EVENT: INFO: Act: TTB_EXP("  id: 3");
/// TEST_EVENT: INFO: Act: TTB_EXP("  tolerance: 1.357");
/// TEST_EVENT: INFO: Act: TTB_EXP("  x");
/// TEST_EVENT: INFO: Act: TTB_EXP("    6.28532");
/// 
/// ---- TEST_EVENT check section
/// 
/// Leaving test case "TestCheckNode"; testing time: 47ms
/// Entering test case "TestCompare"
/// 
/// ---- TEST_EVENT next section: "Compare - with/without use of TestEvents)"
/// 
/// Comparing two XML trees without differences (with TestEvents)
/// TEST_EVENT: INFO: Act: TTB_EXP("XML-Compare: no differences found");
/// 
/// Comparing two XML trees without differences (without TestEvents)
/// XML-Compare: no differences found
/// 
/// Comparing two XML trees without differences - now suppressing OK message
/// 
/// ---- TEST_EVENT next section: "Compare - detection of first difference"
/// 
/// TEST_EVENT: INFO: Act: TTB_EXP("Difference detected");
/// TEST_EVENT: INFO: Act: TTB_EXP("A: time: 09:15:12.234 (line: 3 col: 34, path: //MyContainer/Group_A/Vector)");
/// TEST_EVENT: INFO: Act: TTB_EXP("B: time: 09:30:12.234 (line: 5 col: 34, path: //MyContainer/Group_A/Vector)");
/// 
/// ---- TEST_EVENT next section: "Compare - ignoring 'time'"
/// 
/// TEST_EVENT: INFO: Act: TTB_EXP("Difference detected");
/// TEST_EVENT: INFO: Act: TTB_EXP("A: tolerance: 0.0258 (line: 7 col: 54, path: //MyContainer/Group_A/Vector)");
/// TEST_EVENT: INFO: Act: TTB_EXP("B: tolerance: 0.8258 (line: 9 col: 54, path: //MyContainer/Group_A/Vector)");
/// 
/// ---- TEST_EVENT next section: "Compare - using XPath to compare subtrees (/MyContainer/Group_B)"
/// 
/// TEST_EVENT: INFO: Act: TTB_EXP("Difference detected");
/// TEST_EVENT: INFO: Act: TTB_EXP("A: 3.132437 (line: 18 col: 8, path: //MyContainer/Group_B/x/3.132437/)");
/// TEST_EVENT: INFO: Act: TTB_EXP("B: 12.132437 (line: 20 col: 8, path: //MyContainer/Group_B/x/12.132437/)");
/// 
/// For more test cases see also projects TestXmlCheckBasic,TestXmlCheckWithExtensionsForBoost
/// 
/// ---- TEST_EVENT check section
/// 
/// Leaving test case "TestCompare"; testing time: 47ms
/// Entering test case "Cleanup"
/// Test case Cleanup doesn't include any assertions
/// Leaving test case "Cleanup"
/// Leaving test suite "Master Test Suite"
/// 
/// *** No errors detected
///
/// \endcode
///
/// \sa TestXmlCheckWithBoostExtensions.cpp (file documentation)
/// \sa GrpXmlCheck (usage documentation)
///
////////////////////////////////////////////////////////////////////////////////

///\cond do not document

#include "stdafx.h"

#include "TinyXPath\TinyXml.h"
#pragma warning(push)
#pragma warning(disable:4267)
#pragma warning(disable:4996)
#pragma warning(disable:4100)
#include "TinyXPath\xpath_static.h"
#pragma warning(pop) 

#include "TestToolBox\XmlCheck.h"
#include "TestToolBox\ExtensionsForBoostTest/BoostDefaultTestEnvironment.h"
#include "TinyXPath\TinyXml.h"
#include <iostream>
//#include "TestToolBox\TestToolBoxImpl.h"

namespace TTB = TestToolBox;
using namespace std;

#include <boost/bind.hpp>


//#include <boost/test/unit_test.hpp>

// Test cases should be automatically executed by Boost framework
//#define BOOST_AUTO_TEST_MAIN (here defined within project settings)
#include <boost/test/auto_unit_test.hpp>

//*** Use existing boost test framework implementation ***

// Possibility 1:
// using test framework by explicitly linking with boost framework library
// Important: use the library version which is built with the same C/C++ runtime
// as the test executable is built with
//   mt : multithreaded
//    s : static runtime
//    d : debug build
//    g : debug/diagnostic runtime
//#pragma comment(lib, "boost_unit_test_framework-vc80-mt-gd-1_36")

// Possibility 2:
// Use the boost auto link feature which automatically chooses the correct library version
#define BOOST_LIB_NAME boost_unit_test_framework
#include <boost/config/auto_link.hpp>

// Possibility 3:
// instead of linking with the unit test library you can compile all required sources
// through the following include;
// disadvantage: longer compilation time
// #include <boost/test/included/unit_test_framework.hpp>


using boost::unit_test::test_suite;


static const char* DEMO_XML_STRING =
    "<MyContainer>\n"
    "  <Group_A>\n"
    "    <Vector id=1 date=\"01.07.09\" time=\"09:15:12.234\" tolerance=0.0258>\n"
    "      This is the text of the first vector element\n"
    "      <x>3.14235</x>\n"
    "    </Vector>\n"
    "    <Vector id=2 date=\"01.07.09\" time=\"09:15:12.234\" tolerance=0.0258>\n"
    "      This is the text of the second vector element\n"
    "      <x>3.14235</x>\n"
    "    </Vector>\n"
    "    <Vector id=3 tolerance=1.357>\n"
    "      <x>6.28532</x>\n"
    "    </Vector>\n"
    "  </Group_A>\n"
    "  <Group_B>\n"
    "    <x>1.23456</x>\n"
    "    <x>2.3581</x>\n"
    "    <x>3.132437</x>\n"
    "    <x>4.987654</x>\n"
    "    <x>5.29741</x>\n"
    "  </Group_B>\n"
    "  <Group_C>\n"
    "    <Interior>\n"
    "      <Something>Some text</Something>\n"
    "      <date>28.09.09</date>\n"
    "    </Interior>\n"
    "  </Group_C>\n"
    "</MyContainer>\n";

static const char* DEMO_XML_STRING_WITH_POINTER_VALUES =
    "<MyContainer>\n"
    "  <Group_A>\n"
    "    <Vector id=1 pointerAttr=0>\n"
    "      This is the text of the first vector element\n"
    "      <x>3.14235</x>\n"
    "    </Vector>\n"
    "    <Vector id=2  pointerAttr=0abcdeff>\n"
    "      This is the text of the second vector element\n"
    "      <x>3.14235</x>\n"
    "    </Vector>\n"
    "    <Vector id=3 pointerAttr=\"\">\n"
    "      This is the text of the third vector element\n"
    "      <x>3.14235</x>\n"
    "    </Vector>\n"
    "    <Vector id=4 pointerAttr=\"00000000\">\n"
    "      This is the text of the fourth vector element\n"
    "      <x>3.14235</x>\n"
    "    </Vector>\n"
    "  </Group_A>\n"
    "  <Group_B>\n"
    "    <pX>0</pX>\n"
    "    <pY>abcdef0f</pY>\n"
    "    <pZ></pZ>\n"
    "    <pPhi>00000000</pPhi>\n"
    "  </Group_B>\n"
    "</MyContainer>\n";


// Group_A\x and Group_B\ 3th x is changed
static const char* DEMO_XML_STRING_FOR_COMPARE =
    "<MyContainer>\n"
    "  <Group_A>\n\n\n"
    "    <Vector id=1 date=\"01.07.09\" time=\"09:30:12.234\" tolerance=0.0258>\n"
    "      This is the text of the first vector element\n"
    "      <x>3.14235</x>\n"
    "    </Vector>\n"
    "    <Vector id=2 date=\"01.07.09\" time=\"09:15:12.234\" tolerance=0.8258>\n"
    "      This is the text of the second vector element\n"
    "      <x>3.14235</x>\n"
    "    </Vector>\n"
    "    <Vector id=3 tolerance=1.357>\n"
    "      <x>6.28532</x>\n"
    "    </Vector>\n"
    "  </Group_A>\n"
    "  <Group_B>\n"
    "    <x>1.23456</x>\n"
    "    <x>2.3581</x>\n"
    "    <x>12.132437</x>\n"
    "    <x>4.987654</x>\n"
    "    <x>5.29741</x>\n"
    "  </Group_B>\n"
    "  <Group_C>\n"
    "    <Interior>\n"
    "      <Something>Some text</Something>\n"
    "      <date>28.09.09</date>\n"
    "    </Interior>\n"
    "  </Group_C>\n"
    "</MyContainer>\n";

static const std::string MY_TEST_DIR = "c:/MyTemp";


//************************************************************************
//************************************************************************


class MyTestEnv : TTB::AutoSetupDefaultTestEnv, TTB::FinalCleanup
{
public:
    MyTestEnv()
    {
	    TTB::XmlCheck::Get()->SetEventReceiver(TTB::TestEvents::Get());
        m_eventReceiverForTestEvents.SetSplitAtNewLine(true);

        // Write TestEvents as informational message to test output
        // in the moment when they occur.
        // Optionally write events as 'TTB_EXP("here is the event text");' to test output
        // to allow simple copying to test source code when writing test script.
        TTB::TestEvents::Get()->SetDetailedLog(true, "TTB_EXP");
    }
    virtual ~MyTestEnv()
    {
        TTB::XmlCheck::Cleanup();
    }
};

MyTestEnv* global_pMyTestEnv = 0;

BOOST_GLOBAL_FIXTURE(MyTestEnv);

//************************************************************************
//************************************************************************


BOOST_AUTO_TEST_CASE (SetupTestEnv)
{
    global_pMyTestEnv = new MyTestEnv;

} // SetupTestEnv


//************************************************************************
//************************************************************************


BOOST_AUTO_TEST_CASE (TestCheckNode)
{
    BOOST_MESSAGE( "\n______________________________________________\n"
                   "Testcases for CheckNode"
                   "\n______________________________________________\n");

    TTB_NEXT_SECTION("Examples how to get an XmlNode from an XML string or an XML file");

    // Read XML string into XML document structure
    TiXmlDocument doc;
    doc.Parse(DEMO_XML_STRING);

    // Save XML document to file
    std::string filePath = MY_TEST_DIR + "/GeneratedTinyXmlFile.xml";
    doc.SaveFile(filePath.c_str());

    // Read XML file into XML document structure
    TiXmlDocument docFromFile;
    docFromFile.LoadFile(filePath.c_str());

    BOOST_MESSAGE("now doc and docFromFile should contain the same XML data");


    //-----------------------------------------------------
    TTB_NEXT_SECTION("CheckNode");
    //-----------------------------------------------------

    BOOST_MESSAGE("Remark:\n"
        "Within this test environment XmlCheck is connected with TestEvents via XmlCheck::SetEventReceiver(TTB::TestEvents::Get()).\n"
        "CheckNode() writes all entries to TestEvents, but this class doesn't use the context entry.\n"
        "Therefore the context set by XmlCheck()::SetOutputModeForWritingXmlContents is not relevant.\n");

    TTB::TheXmlCheck()->CheckNode(&doc);
    TTB_EXP("MyContainer");
    TTB_EXP("  Group_A");
    TTB_EXP("    Vector");
    TTB_EXP("      id: 1");
    TTB_EXP("      date: 01.07.09");
    TTB_EXP("      time: 09:15:12.234");
    TTB_EXP("      tolerance: 0.0258");
    TTB_EXP("      This is the text of the first vector element");
    TTB_EXP("      x");
    TTB_EXP("        3.14235");
    TTB_EXP("    Vector");
    TTB_EXP("      id: 2");
    TTB_EXP("      date: 01.07.09");
    TTB_EXP("      time: 09:15:12.234");
    TTB_EXP("      tolerance: 0.0258");
    TTB_EXP("      This is the text of the second vector element");
    TTB_EXP("      x");
    TTB_EXP("        3.14235");
    TTB_EXP("    Vector");
    TTB_EXP("      id: 3");
    TTB_EXP("      tolerance: 1.357");
    TTB_EXP("      x");
    TTB_EXP("        6.28532");
    TTB_EXP("  Group_B");
    TTB_EXP("    x");
    TTB_EXP("      1.23456");
    TTB_EXP("    x");
    TTB_EXP("      2.3581");
    TTB_EXP("    x");
    TTB_EXP("      3.132437");
    TTB_EXP("    x");
    TTB_EXP("      4.987654");
    TTB_EXP("    x");
    TTB_EXP("      5.29741");
    TTB_EXP("  Group_C");
    TTB_EXP("    Interior");
    TTB_EXP("      Something");
    TTB_EXP("        Some text");
    TTB_EXP("      date");
    TTB_EXP("        28.09.09");


    //-----------------------------------------------------
    TTB_NEXT_SECTION(CheckNode - detailed output);
    //-----------------------------------------------------

    TTB::TheXmlCheck()->SetDetailedOutput(true);
    TTB::TheXmlCheck()->CheckNode(&docFromFile);

    TTB_EXP("Document (line: 1, col: 1)");
    TTB_EXP("Element: MyContainer (line: 1, col: 1)");
    TTB_EXP("  Element: Group_A (line: 2, col: 5)");
    TTB_EXP("    Element: Vector (line: 3, col: 9)");
    TTB_EXP("      id: 1, as int: 1, as double: 1 (line: 3, col: 17)");
    TTB_EXP("      date: 01.07.09, as int: 1, as double: 1.07 (line: 3, col: 24)");
    TTB_EXP("      time: 09:15:12.234, as int: 9, as double: 9 (line: 3, col: 40)");
    TTB_EXP("      tolerance: 0.0258, as int: 0, as double: 0.0258 (line: 3, col: 60)");
    TTB_EXP("      4 attributes");
    TTB_EXP("      Text: This is the text of the first vector element (line: 3, col: 79)");
    TTB_EXP("      Element: x (line: 4, col: 13)");
    TTB_EXP("        Text: 3.14235 (line: 4, col: 16)");
    TTB_EXP("    Element: Vector (line: 6, col: 9)");
    TTB_EXP("      id: 2, as int: 2, as double: 2 (line: 6, col: 17)");
    TTB_EXP("      date: 01.07.09, as int: 1, as double: 1.07 (line: 6, col: 24)");
    TTB_EXP("      time: 09:15:12.234, as int: 9, as double: 9 (line: 6, col: 40)");
    TTB_EXP("      tolerance: 0.0258, as int: 0, as double: 0.0258 (line: 6, col: 60)");
    TTB_EXP("      4 attributes");
    TTB_EXP("      Text: This is the text of the second vector element (line: 6, col: 79)");
    TTB_EXP("      Element: x (line: 7, col: 13)");
    TTB_EXP("        Text: 3.14235 (line: 7, col: 16)");
    TTB_EXP("    Element: Vector (line: 9, col: 9)");
    TTB_EXP("      id: 3, as int: 3, as double: 3 (line: 9, col: 17)");
    TTB_EXP("      tolerance: 1.357, as int: 1, as double: 1.357 (line: 9, col: 24)");
    TTB_EXP("      2 attributes");
    TTB_EXP("      Element: x (line: 10, col: 13)");
    TTB_EXP("        Text: 6.28532 (line: 10, col: 16)");
    TTB_EXP("  Element: Group_B (line: 13, col: 5)");
    TTB_EXP("    Element: x (line: 14, col: 9)");
    TTB_EXP("      Text: 1.23456 (line: 14, col: 12)");
    TTB_EXP("    Element: x (line: 15, col: 9)");
    TTB_EXP("      Text: 2.3581 (line: 15, col: 12)");
    TTB_EXP("    Element: x (line: 16, col: 9)");
    TTB_EXP("      Text: 3.132437 (line: 16, col: 12)");
    TTB_EXP("    Element: x (line: 17, col: 9)");
    TTB_EXP("      Text: 4.987654 (line: 17, col: 12)");
    TTB_EXP("    Element: x (line: 18, col: 9)");
    TTB_EXP("      Text: 5.29741 (line: 18, col: 12)");
    TTB_EXP("  Element: Group_C (line: 20, col: 5)");
    TTB_EXP("    Element: Interior (line: 21, col: 9)");
    TTB_EXP("      Element: Something (line: 22, col: 13)");
    TTB_EXP("        Text: Some text (line: 22, col: 24)");
    TTB_EXP("      Element: date (line: 23, col: 13)");
    TTB_EXP("        Text: 28.09.09 (line: 23, col: 19)");



    //-----------------------------------------------------
    TTB_NEXT_SECTION("CheckNode - ignore attributes and nodes, define double values with precision");
    //-----------------------------------------------------

    BOOST_MESSAGE("Ignore attributes 'date' and 'time' and node 'date'\n"
                  "Attribute 'tolerance' is a double with precision 3\n"
                  "Node 'x' is a double with precision 4\n");

    TTB::TheXmlCheck()->SetDetailedOutput(false);
    TTB::TheXmlCheck()->IgnoreAttribute("date");
    TTB::TheXmlCheck()->IgnoreAttribute("time");
    TTB::TheXmlCheck()->IgnoreNode("date");
    TTB::TheXmlCheck()->UseAsDoubleAttribute("tolerance", true, 3);
    TTB::TheXmlCheck()->UseAsDoubleNode("x", true, 4);
    TTB::TheXmlCheck()->CheckNode(&doc);

    TTB_EXP("MyContainer");
    TTB_EXP("  Group_A");
    TTB_EXP("    Vector");
    TTB_EXP("      id: 1");
    TTB_EXP("      tolerance: 0.026");
    TTB_EXP("      This is the text of the first vector element");
    TTB_EXP("      x");
    TTB_EXP("        3.1424");
    TTB_EXP("    Vector");
    TTB_EXP("      id: 2");
    TTB_EXP("      tolerance: 0.026");
    TTB_EXP("      This is the text of the second vector element");
    TTB_EXP("      x");
    TTB_EXP("        3.1424");
    TTB_EXP("    Vector");
    TTB_EXP("      id: 3");
    TTB_EXP("      tolerance: 1.357");
    TTB_EXP("      x");
    TTB_EXP("        6.2853");
    TTB_EXP("  Group_B");
    TTB_EXP("    x");
    TTB_EXP("      1.2346");
    TTB_EXP("    x");
    TTB_EXP("      2.3581");
    TTB_EXP("    x");
    TTB_EXP("      3.1324");
    TTB_EXP("    x");
    TTB_EXP("      4.9877");
    TTB_EXP("    x");
    TTB_EXP("      5.2974");
    TTB_EXP("  Group_C");
    TTB_EXP("    Interior");
    TTB_EXP("      Something");
    TTB_EXP("        Some text");

    //-----------------------------------------------------
    TTB_NEXT_SECTION("CheckNode - ignore complete subtree 'Group_A' ");
    //-----------------------------------------------------

    BOOST_MESSAGE("Furthermore set precision for 'x' to 1\n");

    TTB::TheXmlCheck()->IgnoreNode("Group_A");
    TTB::TheXmlCheck()->UseAsDoubleNode("x", true, 1);
    TTB::TheXmlCheck()->CheckNode(&doc);

    TTB_EXP("MyContainer");
    TTB_EXP("  Group_B");
    TTB_EXP("    x");
    TTB_EXP("      1.2");
    TTB_EXP("    x");
    TTB_EXP("      2.4");
    TTB_EXP("    x");
    TTB_EXP("      3.1");
    TTB_EXP("    x");
    TTB_EXP("      5.0");
    TTB_EXP("    x");
    TTB_EXP("      5.3");
    TTB_EXP("  Group_C");
    TTB_EXP("    Interior");
    TTB_EXP("      Something");
    TTB_EXP("        Some text");


    //-----------------------------------------------------
    TTB_NEXT_SECTION("CheckSubNode with XPath - only Group_B (/MyContainer/Group_B)");
    //-----------------------------------------------------

    TTB::TheXmlCheck()->SetDefaults();

    TTB::TheXmlCheck()->CheckSubNode(doc.RootElement(),"/MyContainer/Group_B");

    TTB_EXP("Group_B");
    TTB_EXP("  x");
    TTB_EXP("    1.23456");
    TTB_EXP("  x");
    TTB_EXP("    2.3581");
    TTB_EXP("  x");
    TTB_EXP("    3.132437");
    TTB_EXP("  x");
    TTB_EXP("    4.987654");
    TTB_EXP("  x");
    TTB_EXP("    5.29741");


    //-----------------------------------------------------
    TTB_NEXT_SECTION("CheckSubNode with XPath - first x element within Group_B (/MyContainer/Group_B/x)");
    //-----------------------------------------------------

    TTB::TheXmlCheck()->CheckSubNode(doc.RootElement(),"/MyContainer/Group_B/x");

    TTB_EXP("x");
    TTB_EXP("  1.23456");


    //-----------------------------------------------------
    TTB_NEXT_SECTION("CheckSubNode with XPath - third x element within Group_B (/MyContainer/Group_B/x[3])");
    //-----------------------------------------------------

    TTB::TheXmlCheck()->CheckSubNode(doc.RootElement(),"/MyContainer/Group_B/x[3]");

    TTB_EXP("x");
    TTB_EXP("  3.132437");


    //-----------------------------------------------------
    TTB_NEXT_SECTION("CheckSubNode with XPath - second vector element of Group_A (/MyContainer/Group_A/Vector[2])");
    //-----------------------------------------------------

    TTB::TheXmlCheck()->CheckSubNode(doc.RootElement(),"/MyContainer/Group_A/Vector[2]");

    TTB_EXP("Vector");
    TTB_EXP("  id: 2");
    TTB_EXP("  date: 01.07.09");
    TTB_EXP("  time: 09:15:12.234");
    TTB_EXP("  tolerance: 0.0258");
    TTB_EXP("  This is the text of the second vector element");
    TTB_EXP("  x");
    TTB_EXP("    3.14235");


    //-----------------------------------------------------
    TTB_NEXT_SECTION("CheckSubNode with XPath - last vector element of Group_A (/MyContainer/Group_A/Vector[last()])");
    //-----------------------------------------------------

    TTB::TheXmlCheck()->CheckSubNode(doc.RootElement(),"/MyContainer/Group_A/Vector[last()]");

    TTB_EXP("Vector");
    TTB_EXP("  id: 3");
    TTB_EXP("  tolerance: 1.357");
    TTB_EXP("  x");
    TTB_EXP("    6.28532");


    //-----------------------------------------------------
    TTB_NEXT_SECTION("CheckSubNode with XPath - relative to subtree via SubNodeFromXPath (/MyContainer/Group_A, Vector[last())");
    //-----------------------------------------------------

    TTB::TheXmlCheck()->CheckSubNode(
        TTB::XmlCheck::SubNodeFromXPath(doc.RootElement(),"/MyContainer/Group_A"),
        "Vector[last()]");

    TTB_EXP("Vector");
    TTB_EXP("  id: 3");
    TTB_EXP("  tolerance: 1.357");
    TTB_EXP("  x");
    TTB_EXP("    6.28532");

    TTB_CHECK_SECTION();

} // TestCheckNode


//************************************************************************
//************************************************************************

BOOST_AUTO_TEST_CASE (TestCheckNodeWithPointerValues)
{
    // Read XML string into XML document structure
    TiXmlDocument doc;
    doc.Parse(DEMO_XML_STRING_WITH_POINTER_VALUES);

   TTB::TheXmlCheck()->UseAsPointerAttribute("pointerAttr", true);
   TTB::TheXmlCheck()->UseAsPointerNode("pX", true);
   TTB::TheXmlCheck()->UseAsPointerNode("pY", true);
   TTB::TheXmlCheck()->UseAsPointerNode("pZ", true);
   TTB::TheXmlCheck()->UseAsPointerNode("pPhi", true);

    TTB::TheXmlCheck()->CheckNode(&doc);
    TTB_EXP("MyContainer");
    TTB_EXP("  Group_A");
    TTB_EXP("    Vector");
    TTB_EXP("      id: 1");
    TTB_EXP("      pointerAttr: 0");
    TTB_EXP("      This is the text of the first vector element");
    TTB_EXP("      x");
    TTB_EXP("        3.14235");
    TTB_EXP("    Vector");
    TTB_EXP("      id: 2");
    TTB_EXP("      pointerAttr: exists");
    TTB_EXP("      This is the text of the second vector element");
    TTB_EXP("      x");
    TTB_EXP("        3.14235");
    TTB_EXP("    Vector");
    TTB_EXP("      id: 3");
    TTB_EXP("      pointerAttr: not set");
    TTB_EXP("      This is the text of the third vector element");
    TTB_EXP("      x");
    TTB_EXP("        3.14235");
    TTB_EXP("    Vector");
    TTB_EXP("      id: 4");
    TTB_EXP("      pointerAttr: 0");
    TTB_EXP("      This is the text of the fourth vector element");
    TTB_EXP("      x");
    TTB_EXP("        3.14235");
    TTB_EXP("  Group_B");
    TTB_EXP("    pX");
    TTB_EXP("      0");
    TTB_EXP("    pY");
    TTB_EXP("      exists");
    TTB_EXP("    pZ");
    TTB_EXP("    pPhi");
    TTB_EXP("      0");
}

//************************************************************************
//************************************************************************


BOOST_AUTO_TEST_CASE (TestCompare)
{
    //-----------------------------------------------------
    TTB_NEXT_SECTION("Compare - with/without use of TestEvents)");
    //-----------------------------------------------------

    TiXmlDocument doc;
    doc.Parse(DEMO_XML_STRING);

    TiXmlDocument docA;
    docA.Parse(DEMO_XML_STRING);

    TiXmlDocument docB;
    docB.Parse(DEMO_XML_STRING_FOR_COMPARE);

    BOOST_MESSAGE("Comparing two XML trees without differences (with TestEvents)");
    TTB::TheXmlCheck()->Compare(&docA,&docA);
    TTB_EXP("XML-Compare: no differences found");


    BOOST_MESSAGE("\nComparing two XML trees without differences (without TestEvents)");

    // Compare does not produce any output if no EventReceiver (here class TestEvents)
    // is set but the result can be explicitly checked
    TTB::IEventReceiver* pOldReceiver = TTB::TheXmlCheck()->SetEventReceiver(0);

    BOOST_CHECK(TTB::TheXmlCheck()->Compare(&docA,&docA));
    BOOST_MESSAGE(TTB::TheXmlCheck()->GetCompareResult());

    BOOST_MESSAGE("\nComparing two XML trees without differences"
                  " - now suppressing OK message");
    bool checkOk = TTB::TheXmlCheck()->Compare(&docA,&docA);
    BOOST_CHECK(checkOk);
    if (!checkOk)
    {
        BOOST_MESSAGE(TTB::TheXmlCheck()->GetCompareResult());
    }

    // Reactivate EventReceiver (i.e. class TestEvents will now be used again)
    TTB::TheXmlCheck()->SetEventReceiver(pOldReceiver);


    //-----------------------------------------------------
    TTB_NEXT_SECTION("Compare - detection of first difference");
    //-----------------------------------------------------

    TTB::TheXmlCheck()->Compare(&docA,&docB);

    TTB_EXP("Difference detected");
    TTB_EXP("A: time: 09:15:12.234 (line: 3 col: 34, path: //MyContainer/Group_A/Vector)");
    TTB_EXP("B: time: 09:30:12.234 (line: 5 col: 34, path: //MyContainer/Group_A/Vector)");


    //-----------------------------------------------------
    TTB_NEXT_SECTION("Compare - ignoring 'time'");
    //-----------------------------------------------------

    TTB::TheXmlCheck()->IgnoreAttribute("time");
    TTB::TheXmlCheck()->Compare(&docA,&docB);

    TTB_EXP("Difference detected");
    TTB_EXP("A: tolerance: 0.0258 (line: 7 col: 54, path: //MyContainer/Group_A/Vector)");
    TTB_EXP("B: tolerance: 0.8258 (line: 9 col: 54, path: //MyContainer/Group_A/Vector)");


    //-----------------------------------------------------
    TTB_NEXT_SECTION("Compare - using XPath to compare subtrees (/MyContainer/Group_B)");
    //-----------------------------------------------------

    TTB::TheXmlCheck()->Compare(
        TTB::XmlCheck::SubNodeFromXPath(docA.RootElement(), "/MyContainer/Group_B"),
        TTB::XmlCheck::SubNodeFromXPath(docB.RootElement(), "/MyContainer/Group_B"));

    TTB_EXP("Difference detected");
    TTB_EXP("A: 3.132437 (line: 18 col: 8, path: //MyContainer/Group_B/x/3.132437/)");
    TTB_EXP("B: 12.132437 (line: 20 col: 8, path: //MyContainer/Group_B/x/12.132437/)");

    BOOST_MESSAGE("\nFor more test cases see also projects TestXmlCheckBasic,TestXmlCheckWithExtensionsForBoost");

    TTB_CHECK_SECTION();

} // TestCompare


//************************************************************************
//************************************************************************


BOOST_AUTO_TEST_CASE (Cleanup)
{
    delete global_pMyTestEnv;
    global_pMyTestEnv = 0;

} // Cleanup


//************************************************************************
//************************************************************************


///\endcond

///@} // end group GrpTestApps

