////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \defgroup GrpTestTime Using TestTime
/// @{
/// \ingroup GrpTestUtils 
/// \brief
/// Class TestTime gives access to start/stop time of test and offers
/// simple access to current time and date.
///
/// \sa TestTime.h (File documentation)
/// \sa TestToolBox::TestTime (class documentation)
/// 
////////////////////////////////////////////////////////////////////////////////

#pragma once

#ifdef USE_TTB_INSTANCE_DLL
#    include "TTB_InstanceDll.h"
#endif

#include <string>

namespace TestToolBox
{

/// \brief This class stores start/stop time of test and supports time
/// and date formatting
///
/// \nosubgrouping 
/// \sa TestTime.h (File documentation)
/// \sa GrpTestTime (Usage documentation)
class TestTime
{
private:
///\cond
    // Constructor.
    explicit TestTime (void);
///\endcond
public:
///\cond

    // Destructor
    ~TestTime(){};

    // Get the pointer to the only instance of this class
#ifdef USE_TTB_INSTANCE_DLL
    static TestTime* CreateNewInstanceForUsageInDll(void){return new TestTime;};
    static TestTime* Get (void)
    {return TTB_GetInstance_TestTime();}
    static void Cleanup (void)
    {TTB_CleanupInstance_TestTime();}
#else
    static TestTime* Get (void)
    {if (!s_pTestTime) s_pTestTime = new TestTime;
     return s_pTestTime;}
    static TestTime* s_pTestTime;
    static void Cleanup (void)
    {delete s_pTestTime; s_pTestTime = 0;}
#endif

///\endcond


    /// Retrieve current time and store as start/stop time.
    void StoreStartTime();
    void StoreStopTime();

    /// \brief Reset the time already stored for stopping (e.g. we are
    /// running a loop test and want to get actualized stop times
    /// on each cycle)
    void ResetStopTime();

    /// Get the stored time information
    std::string GetStartTimeDateStr (void);
    std::string GetStopTimeDateStr (void);

    /// Get the test duration in format hh:mm::ss,ms
    std::string GetTestDurationStr (void);

    /// Get the current test time in milliseconds
    long GetCurTimeMs (void);

    /// Get the current test time in seconds
    long GetCurTimeSec (void);

    /// Get the current testtime in format hh:mm::ss,ms
    std::string GetCurTimeStr (void);

    /// Get the current date and time in format dd.mm.yyyy  hh:mm::ss,ms
    std::string GetCurTimeDateStr (void);

    /// Constructs formatted time and date string
    static std::string FormatTimeAndDate (
        struct tm* in_pTime);

    /// Constructs time string of format hh:mm:ss,ms
    static std::string FormatTestDuration (
        long in_durationMs);

private:

    /// TickCount at start of test
    long m_startTimeMs;

    /// Stored time in string format for start and end of test
    std::string m_startTimeDateStr;
    std::string m_stopTimeDateStr;
    std::string m_testDurationStr;

}; // class TestTime

///\cond do not document
#ifndef USE_TTB_INSTANCE_DLL
__declspec(selectany) TestTime* TestTime::s_pTestTime = 0;
#endif
///\endcond

}; // namespace TestToolBox

/// @} // endGrpTestTime
