////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
////////////////////////////////////////////////////////////////////////////////

#include "CaptureOutput.h"
#include "Services.h"

#include <iostream>
#include <string>
#include <iomanip>

//-----------------------------------------------------------------------------

CaptureOutput::CaptureOutput(
	std::string in_fullPathToExecutable,
	bool        in_captureIsActive,
	int         in_traceLevel)

	: m_captureIsActive        (in_captureIsActive)
	, m_traceLevel             (in_traceLevel)
	, m_hStdOutRead            (NULL)
	, m_hStdOutWrite           (NULL) 
	, m_fileNameCapturedOutput (in_fullPathToExecutable + ".CapturedOutput.log")
	, m_hPipe(NULL)
	, m_pipeOverlapped()
{

}

//-----------------------------------------------------------------------------

CaptureOutput::~CaptureOutput(void)
{
}

//-----------------------------------------------------------------------------

bool CaptureOutput::IsActive() const
{
	return m_captureIsActive;
}

//-----------------------------------------------------------------------------

DWORD CaptureOutput::GetWaitTime () const
{
	return PIPE_TIMEOUT_MS;
}

//-----------------------------------------------------------------------------

std::string CaptureOutput::GetFileNameCapturedOutput() const
{
	return m_fileNameCapturedOutput;
}

//-----------------------------------------------------------------------------

bool CaptureOutput::Prepare(
	int &          io_numEventsToWaitFor,
	EventInfoMap & io_eventInfos,
	HANDLE &       out_hStdOut,
	HANDLE &       out_hStdErr)
{
	out_hStdOut = 0;
	out_hStdErr = 0;

	if (!m_captureIsActive)
	{
		// do nothing
		return true;
	}

	SECURITY_ATTRIBUTES saAttr; 
	// Set the bInheritHandle flag so pipe handles are inherited. 
	saAttr.nLength = sizeof(SECURITY_ATTRIBUTES); 
	saAttr.bInheritHandle = TRUE; 
	saAttr.lpSecurityDescriptor = NULL; 

	LPTSTR lpszPipename = TEXT("\\\\.\\pipe\\TimedExec_pipeToCaptureStdOut2");
	HANDLE hEventReadFromPipe = CreateEvent( 
		NULL,    // default security attribute 
		TRUE,    // manual-reset event 
		TRUE,    // initial state = signaled 
		NULL);   // unnamed event object 

	if (hEventReadFromPipe == NULL) 
	{
		WriteSystemErrorMsg (m_traceLevel > 0,
			"[TimedExec::CaptureOutput::Prepare]: CreateEvent failed!");
		return false;
	}

	++io_numEventsToWaitFor;
	io_eventInfos[hEventReadFromPipe]=EventInfo(HandleType::PIPE_STD_OUT, hEventReadFromPipe);
	m_pipeOverlapped.hEvent = hEventReadFromPipe; 

	//m_hPipe = CreateNamedPipe( 
	//	lpszPipename,             // pipe name 
	//	PIPE_ACCESS_DUPLEX |      // read/write access 
	//	FILE_FLAG_OVERLAPPED,     // overlapped mode 
	//	PIPE_TYPE_MESSAGE |       // message-type pipe 
	//	PIPE_READMODE_MESSAGE |   // message-read mode 
	//	PIPE_WAIT,                // blocking mode 
	//	PIPE_UNLIMITED_INSTANCES, // number of instances 
	//	BUFSIZE*sizeof(TCHAR),    // output buffer size 
	//	BUFSIZE*sizeof(TCHAR),    // input buffer size 
	//	PIPE_TIMEOUT_MS,          // client time-out 
	//	NULL);                    // default security attributes 
	m_hPipe = CreateNamedPipe( 
		lpszPipename,             // pipe name 
		PIPE_ACCESS_INBOUND |     // The flow of data in the pipe goes from client to server only. 
		FILE_FLAG_OVERLAPPED,     // overlapped mode 
		PIPE_TYPE_BYTE |          // Data is written to the pipe as a stream of bytes. 
		PIPE_READMODE_BYTE |      // Data is read from the pipe as a stream of bytes
		PIPE_NOWAIT,
		PIPE_UNLIMITED_INSTANCES, // number of instances 
		BUFSIZE*sizeof(TCHAR),    // output buffer size 
		BUFSIZE*sizeof(TCHAR),    // input buffer size 
		PIPE_TIMEOUT_MS,          // client time-out 
		NULL);                    // default security attributes 
	if (m_hPipe == INVALID_HANDLE_VALUE) 
	{
		WriteSystemErrorMsg (m_traceLevel > 0,
			"[TimedExec::CaptureOutput::Prepare]: CreateNamedPipe failed!");
		return false;
	}

	if ( ! CreatePipe(&m_hStdOutRead, &m_hStdOutWrite, &saAttr, 0) )
	{
		WriteSystemErrorMsg (m_traceLevel > 0,
			"[TimedExec::CaptureOutput::Prepare]: CreatePipe failed!");
		return false;
	}

	// Ensure the read handle to the pipe for STDOUT is not inherited.
	if ( ! SetHandleInformation(m_hStdOutRead, HANDLE_FLAG_INHERIT, 0) )
	{
		WriteSystemErrorMsg (m_traceLevel > 0,
			"[TimedExec::CaptureOutput::Prepare]: SetHandleInformation failed!");
		return false;
	}

	m_fsCapturedOut.open (m_fileNameCapturedOutput, std::fstream::out);

	out_hStdOut = m_hStdOutWrite;
	out_hStdErr = m_hStdOutWrite;

	return true;
}

//-----------------------------------------------------------------------------

bool CaptureOutput::ReleaseResourcesAfterStartOfExeFile()
{
	if (m_captureIsActive)
	{
		// Close write handle to allow correct recognition of a terminated pipe connection
		// when reading from the pipe
		CloseHandle(m_hStdOutWrite);
	}

	return true;
}

//-----------------------------------------------------------------------------

void  CaptureOutput::HandleEventReadPipeData(HANDLE handle, EventInfoMap& eventInfoMap)
{
	if (!m_captureIsActive)
	{
		eventInfoMap[handle].SetFinished();
		return;
	}
	const int MY_BUFSIZE = 200;
	char buffer[MY_BUFSIZE];

	// Check if there are data available (check should not be necessary)
	DWORD numBytesRead=0;
	DWORD numBytesAvailable=0;
	BOOL succeeded = PeekNamedPipe(m_hStdOutRead,&buffer,MY_BUFSIZE,&numBytesRead,&numBytesAvailable,NULL);
	if (!succeeded)
	{
		DWORD errCode = GetLastError();
		if (errCode == ERROR_BROKEN_PIPE)  // regular end of pipe connnection
		{
			std::cout << "[TimedExec]: End of captured output" << std::endl;
			eventInfoMap[handle].SetFinished();
			return;
		}

		WriteSystemErrorMsg (m_traceLevel >= 0, errCode,
			"[TimedExec-CaptureOutput::HandleEventReadPipeData] PeekNamedPipe-Error");
		return;
	}
	if (numBytesAvailable<=0)
	{
		WriteSystemErrorMsg (m_traceLevel >= 0,
			"[TimedExec-CaptureOutput::HandleEventReadPipeData] PeekNamedPipe, no bytes available, sleeping 1 sec...");
		::Sleep(1000); // avoid repetitive calls within short time interval
		return; // avoid blocking call of ReadFile when data are not yet available
	}
	if (m_traceLevel > 0) 
		std::cout << "[TimedExec]: HandleEventReadPipeData, numBytesAvailable=" << numBytesAvailable << std::endl;

	if (m_traceLevel > 0) 
		std::cout << "[TimedExec]: HandleEventReadPipeData, calling ReadFile......" << std::endl;
	succeeded = ReadFile(m_hStdOutRead,&buffer,MY_BUFSIZE,NULL,&m_pipeOverlapped);
	if (m_traceLevel > 0) 
		std::cout << "[TimedExec]: HandleEventReadPipeData, ... ReadFile returned" << std::endl;
	if (succeeded)
	{
		DWORD numBytesRead;
		BOOL succeeded = GetOverlappedResult(m_hStdOutRead,&m_pipeOverlapped, &numBytesRead, false /* do not wait */);
		if (succeeded)
		{
			// read data
			std::string s;
			for (unsigned long i = 0; i<numBytesRead; ++i)
			{
				if (buffer[i]!='\r') // skip output of carriage return to avoid double line breaks
				{
        			// Write captured output to stdout to watch progress of client process
					std::cout << buffer[i];
				}

				s += buffer[i];
			}

			// Write captured output to logfile for later check of memory leaks
			m_fsCapturedOut << s.c_str();
		}
		else // no data read to buffer
		{
			DWORD errCode = GetLastError();
			if (errCode == ERROR_IO_PENDING)
			{
				WriteSystemErrorMsg (m_traceLevel >= 2, errCode,
					"[TimedExec-CaptureOutput::HandleEventReadPipeData] GetOverlappedResult-Error: ERROR_IO_PENDING");
			}
			else
			{
				WriteSystemErrorMsg (m_traceLevel >= 1, errCode,
					"[TimedExec-CaptureOutput::HandleEventReadPipeData] GetOverlappedResult-Error:");
			}
		}
	}
	else // no data available
	{
		DWORD errCode = GetLastError();
		if (errCode == ERROR_IO_PENDING)
		{
			WriteSystemErrorMsg (m_traceLevel >= 2, errCode,
				"[TimedExec-CaptureOutput::HandleEventReadPipeData] ReadFile-Error: ERROR_IO_PENDING");
		}
		else if (errCode == ERROR_BROKEN_PIPE)  // regular end of pipe connnection
		{
			std::cout << "[TimedExec]: End of captured output" << std::endl;
			eventInfoMap[handle].SetFinished();
		}
		else
		{
			WriteSystemErrorMsg (m_traceLevel >= 1, errCode,
				"[TimedExec-CaptureOutput::HandleEventReadPipeData] ReadFile-Error:");
		}
	}
}

//-----------------------------------------------------------------------------

bool CaptureOutput::HasDetectedMemLeaks (bool in_performCheck)
{
    if (!m_captureIsActive)
	{
		return false;
	}

	ReleaseResourcesAfterTerminationOfExeFile();

	if (!in_performCheck)
	{
		// execution failed, it does not make sense to check for mem leaks
		return false;
	}

	if (m_traceLevel > 0) 
		std::cout << "[TimedExec]: Now checking captured output for messages about memory leaks (see file "
		<< m_fileNameCapturedOutput << "):" << std::endl;

	std::fstream outFile; // file with captured stdout and stderr of test app
	outFile.open(m_fileNameCapturedOutput.c_str());
	std::string line;
	bool memLeaksFound = false;
	while( std::getline( outFile, line ) )
	{
		// write output of test application to std out
		//std::cout << line << std::endl;

		if (line.find("Detected memory leaks") != std::string::npos)
		{
			memLeaksFound = true;
			if (m_traceLevel > 1) 
				std::cout << "[TimedExec]: ****** MEM LEAK FOUND!!! ****" << std::endl;
			break;
		}
	}
	outFile.close();

	return memLeaksFound;
}

//-----------------------------------------------------------------------------

void CaptureOutput::ReleaseResourcesAfterTerminationOfExeFile()
{
	if (m_captureIsActive)
	{
		m_fsCapturedOut.close();
	}
}

//-----------------------------------------------------------------------------
