////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
////////////////////////////////////////////////////////////////////////////////


#include "TimedExecutor.h"

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <tchar.h>
#include <io.h> // _access

#include "CaptureOutput.h"
#include "Services.h"

#include <time.h>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <fstream>

using namespace std;


///\cond do not document
//-----------------------------------------------------------------------------

TimedExecutor::TimedExecutor (void)
	: m_waitTimeSec               (0)
	, m_traceLevel                (0)
	, m_checkForMemleaks          (true)
	, m_restrictToSingleInstance  (true)
	, m_maxWaitTimeBeforeStartSec (600)
	, m_hMutexSingleInstance      (0)
	, m_mutexAquired              (false)
	, m_logResults                (true)
	, m_executionStatus           (eUNSPECIFIED_ERROR)
	, m_startExecutableTickCount  (0)
	, m_executionTimeSec          (0)
	, m_skipExecution             (false)
	, m_numRetriesAllowed         (3)
	, m_numRetriesUsed            (0)
	, m_retryIsPending            (false)
{
} // TimedExecutor::TimedExecutor

//-----------------------------------------------------------------------------

bool TimedExecutor::DumpCallInfo (int argc, _TCHAR* argv[])
{
	bool succeeded = true;
	try
	{
		if (m_traceLevel >= 2)
		{
			cout << "[TimedExec-DumpCallInfo] ----- Begin -----" << endl;
			for (int i = 0; i <argc; ++i)
			{
				cout << "[TimedExec-DumpCallInfo] arg " << i << ": >" << argv[i] << "<" << endl;
			}
			DisplayCallContext();
			cout << "[TimedExec-DumpCallInfo] ----- End -----" << endl;
		}
	}
	CATCH_ALL(TimedExec-DumpCallInfo,succeeded);
	return succeeded;

} // TimedExecutor::DumpCallInfo

//-----------------------------------------------------------------------------

void TimedExecutor::DisplayCallContext (void)
{
	cout << "[TimedExec-Context] Commandline: " << ::GetCommandLine() << endl;
	// add info about user, computername, ...

} // TimedExecutor::DisplayCallContext

//-----------------------------------------------------------------------------

void TimedExecutor::DisplaySettings (void)
{
	cout << "[TimedExec-Settings] WaitTimeSec                : " 
		<< m_waitTimeSec << endl;
	cout << "[TimedExec-Settings] ExeToStartFullPath         : >"
		<< m_executableToStartFullPath.c_str() << "<" << endl;
	cout << "[TimedExec-Settings] ExeToStartWithoutPath      : >"
		<< m_executableToStartFileName.c_str() << "<" << endl;
	cout << "[TimedExec-Settings] StartArgs                  : >"
		<< m_startArguments.c_str() << "<" << endl;
	cout << "[TimedExec-Settings] TraceLevel                 : "
		<< m_traceLevel << endl;
	cout << "[TimedExec-Settings] CheckForMemleaks           : "
		<< boolalpha <<m_checkForMemleaks << endl;
	cout << "[TimedExec-Settings] RestrictToSingleInstance   : "
		<< boolalpha << m_restrictToSingleInstance << endl;
	cout << "[TimedExec-Settings] MaxWaitTimeBeforeStartSec  : "
		<< m_maxWaitTimeBeforeStartSec << endl;
	cout << "[TimedExec-Settings] LogResults                 : "
		<< boolalpha << m_logResults << endl;
	cout << "[TimedExec-Settings] LogFilePath                : "
		<< m_logFilePath << endl;
	cout << "[TimedExec-Settings] SkipExecution              : "
		<< boolalpha << m_skipExecution << endl;

} // TimedExecutor::DisplaySettings

//-----------------------------------------------------------------------------

void TimedExecutor::DisplayUsage (void)
{
	cout << "[TimedExec-Usage]." << endl;
	cout << "[TimedExec-Usage] Syntax: TimedExec [SILENT|REGULAR|VERBOUS][MAX_NUM_RETRIES=1][SKIP_MEM_CHECK] secondsToWait exeToStart [any args to pass to the exe]" << endl;

} // TimedExecutor::DisplayUsage

//-----------------------------------------------------------------------------

bool TimedExecutor::ReadEnvironmentVariable (
	std::string const & in_envVarName,
	std::string&        out_envVarVal)
{
	const int MYBUFSIZE = 4096;
	bool varExists = false;

	DWORD dwRet;
	DWORD dwErr;

	TCHAR envVarValue [MYBUFSIZE];

	dwRet = ::GetEnvironmentVariable(TEXT(in_envVarName.c_str()), envVarValue,
		MYBUFSIZE);
	if(0 == dwRet)
	{
		dwErr = GetLastError();
		if( ERROR_ENVVAR_NOT_FOUND == dwErr )
		{
			if (m_traceLevel >= 2) cout << "[TimedExec-ReadEnvironmentVariable]: "
				<< in_envVarName.c_str() << " = <not existing>" << endl;
		}
		else
		{
			std::ostringstream os;
			os << "ERROR: Environment variable " << in_envVarName.c_str() << " could not be accessed (LastError="
				<< dwErr << "/" << GetSystemErrorDescription(dwErr).c_str() << ")" << endl;
			throw std::exception(os.str().c_str());
		}
	}
	else
	{
		if(MYBUFSIZE < dwRet)
		{
			if (m_traceLevel >=1) cout << "[TimedExec-ReadEnvironmentVariable]: WARNING: Environment variable "
				<< in_envVarName.c_str() << " too long! Value is ignored!";
		}
		else // value is accessible
		{
			if (m_traceLevel>=2)
			{
				cout << "[TimedExec-ReadEnvironmentVariable]: "
					<< in_envVarName.c_str() << " = >"
					<< envVarValue << "<" << endl;
			}
			out_envVarVal = envVarValue;
			varExists = true;
		}
	}

	return varExists;

} // TimedExecutor::ReadEnvironmentVariable

//-----------------------------------------------------------------------------

bool TimedExecutor::KillProcess (void)
{
	bool succeeded = false;
	try
	{
		if (::TerminateProcess(m_procInfo.hProcess,1))
		{
			succeeded = true;
		}
		else // termination failed
		{
			int err = GetLastError();
			cout << "[TimedExec-KillProcess]: ERROR: Process "
				<< m_executableToStartFullPath.c_str()
				<< " could not be killed (LastError="
				<< err << "/" << GetSystemErrorDescription(err).c_str()
				<< ")" << endl;
		}
	}
	CATCH_ALL(TimedExec-KillProcess,succeeded)
		return succeeded;

} // TimedExecutor::KillProcess

//-----------------------------------------------------------------------------

std::string TimedExecutor::GetExecutionStatusAsString (std::string const & in_rInfo)
{
	std::string enumStr;
	switch (m_executionStatus)
	{
	case eUNSPECIFIED_ERROR:           enumStr = "UNSPECIFIED_ERROR"; break;
	case eUSAGE_ERROR:                 enumStr = "USAGE_ERROR"; break;
	case eEXECUTION_SKIPPED:           enumStr = "EXECUTION_SKIPPED"; break;
	case eEXE_FILE_NOT_ACCESSIBLE:     enumStr = "EXE_FILE_NOT_ACCESSIBLE"; break;
	case eTIMEOUT_START_PERMISSION:    enumStr = "TIMEOUT_START_PERMISSION"; break;
	case eEXE_COULD_NOT_BE_STARTED:    enumStr = "EXE_COULD_NOT_BE_STARTED"; break;
	case eTIMEOUT_EXE_TERMINATION:     enumStr = "TIMEOUT_EXE_TERMINATION"; break;
	case eEXE_FAILED:                  enumStr = "EXE_FAILED"; break;
	case eEXE_SUCCEEDED:               enumStr = "EXE_SUCCEEDED"; break;
	case eEXE_SUCCEEDED_WITH_MEMLEAKS: enumStr = "EXE_HAS_MEMLEAKS"; break;
	default:                           enumStr = "Unexpected enum value"; break;
	}

	if (!in_rInfo.empty())
	{
		std::string resultStr ("   ");
		resultStr += in_rInfo + " (" + enumStr + ")";
		return resultStr;
	}
	else
	{
		return enumStr;
	}

} // TimedExecutor::GetExecutionStatusAsString

//-----------------------------------------------------------------------------

void TimedExecutor::StoreStartTime (void)
{
	m_startExecutableTickCount = ::GetTickCount();
	m_dateTimeExeWasStarted = GetCurrentDateTimeAsString();

} // TimedExecutor::StoreStartTime

//-----------------------------------------------------------------------------

void TimedExecutor::StoreStopTime (void)
{
	m_executionTimeSec = (::GetTickCount() - m_startExecutableTickCount)/1000.0;

} // TimedExecutor::StoreStopTime

//-----------------------------------------------------------------------------

std::string TimedExecutor::GetCurrentDateTimeAsString (void)
{
	// Store time and date
	struct tm newtime;
	__time64_t long_time;
	_time64( &long_time );
	_localtime64_s( &newtime, &long_time );

	struct tm* pTime = &newtime;

	ostringstream tmpStr;
	tmpStr << std::setw(2) << std::setfill('0') << pTime-> tm_mday << ".";
	tmpStr << std::setw(2) << std::setfill('0') << pTime-> tm_mon + 1 << ".";
	tmpStr << pTime-> tm_year + 1900 << " ";
	tmpStr << std::setw(2) << std::setfill('0') << pTime-> tm_hour << ":";
	tmpStr << std::setw(2) << std::setfill('0') << pTime-> tm_min << ":";
	tmpStr << std::setw(2) << std::setfill('0') << pTime-> tm_sec;

	return tmpStr.str();

} // TimedExecutor::GetCurrentDateTimeAsString

//-----------------------------------------------------------------------------

void TimedExecutor::WriteLineToLogfile (std::string const & in_line)
{
	// Open new or existing file
	std::ofstream logFile (m_logFilePath.c_str(), ios_base::app);
	if (!logFile)
	{
		if (m_traceLevel >=1)
			cout << "[TimedExec-WriteLineToLogfile]: ERROR: Open file "
			<< m_logFilePath << " failed!" << endl;
	}
	else
	{
		logFile << in_line << endl;
		logFile.close();
	}

} // TimedExecutor::WriteLineToLogfile

//-----------------------------------------------------------------------------

bool TimedExecutor::SetTraceLevel (int argc, _TCHAR* argv[])
{
	bool succeeded = true;
	try
	{
		std::string paramTrace;
		if (argc > 1)
		{
			paramTrace = argv[1];
		}

		// possibly the environment overrides the command line argument
		ReadEnvironmentVariable("TimedExec_Trace",paramTrace);

		if (paramTrace == "SILENT")
		{
			m_traceLevel = 0;
		}
		else if (paramTrace == "REGULAR")
		{
			m_traceLevel = 1;
		}
		else if (paramTrace == "VERBOUS")
		{
			m_traceLevel = 2;
		}
	}
	CATCH_ALL(TimedExec-SetTraceLevel,succeeded);
	if (!succeeded)
	{
		DisplayUsage();
	}
	if (!succeeded)
		m_executionStatus = eUSAGE_ERROR;
	return succeeded;

} // TimedExecutor::SetTraceLevel

//-----------------------------------------------------------------------------

bool TimedExecutor::ParseCommandLine (int argc, _TCHAR* argv[])
{
	bool succeeded = true;
	try
	{
		int posSkipMemCheck  = 1;
		int posMaxNumRetries = 1;
		int posWaitTime      = 1;
		int posExeToStart    = 2;
		int posArgs          = 3;

		// Special case: check for help request
		if (argc > 1)
		{
			std::string arg = argv[1];

			if (   (arg == "help")
				|| (arg == "-help")
				|| (arg == "?")
				|| (arg == "-?"))
			{
				DisplayUsage();
				return false;
			}
		}

		// at least expect waittime and executable to start
		if (argc < 3)
		{
			throw std::exception("required args: WaitTime and ExecutableToStart");
		}

		// Check for optional argument SILENT|VERBOUS
		if (argc > 1)
		{
			std::string arg = argv[1];

			if (   (arg == "SILENT")
				|| (arg == "REGULAR")
				|| (arg == "VERBOUS"))
			{
				++posMaxNumRetries;++posSkipMemCheck;++posWaitTime;++posExeToStart;++posArgs;
			}
		}
		// Check for optional argument MAX_NUM_RETRIES=<n>
		if (argc > posMaxNumRetries)
		{
			std::string arg = argv[posMaxNumRetries];
			std::string const MAX_NUM_RETRIES = "MAX_NUM_RETRIES=";
			std::size_t pos = arg.find(MAX_NUM_RETRIES);
			if (pos != string::npos)
			{
				std::string valStr = arg.substr(pos+MAX_NUM_RETRIES.length());
				std::istringstream is (valStr);
				is >> m_numRetriesAllowed;
				if (m_numRetriesAllowed < 0)
				{
					throw std::exception("num retries allowed must be > 0");
				}
				++posSkipMemCheck;++posWaitTime;++posExeToStart;++posArgs;
			}
		}

		// Check for optional argument SKIP_MEM_CHECK
		if (argc > posSkipMemCheck)
		{
			std::string arg = argv[posSkipMemCheck];

			if (arg == "SKIP_MEM_CHECK")
			{
				m_checkForMemleaks = false;
				++posWaitTime;++posExeToStart;++posArgs;
			}
		}

		if (argc < posArgs)
		{
			throw std::exception("required args: WaitTime and ExecutableToStart");
		}

		// read wait time
		if (argc > posWaitTime)
		{
			std::istringstream is (argv[posWaitTime]);
			is >> m_waitTimeSec;
			if (m_waitTimeSec <= 0)
			{
				throw std::exception("wait time must be > 0");
			}
		}

		// read executable path
		if (argc > posExeToStart)
		{
			m_executableToStartFullPath = argv[posExeToStart];

			// Get exe name without path
			m_executableToStartFileName = m_executableToStartFullPath;
			std::string::size_type pos = m_executableToStartFullPath.find_last_of("\\");
			if (pos!=std::string::npos)
			{
				m_executableToStartFileName = m_executableToStartFullPath.
					substr(pos+1,m_executableToStartFullPath.length() - pos);
			}
		}


		// read start arguments
		if (argc > posArgs)
		{
			for (int i = posArgs; i<argc; ++i)
			{
				std::string argStr = argv[i];
				// Remove trailing spaces and "
				Trim(argStr);
				Trim(argStr,"\"");

				// Embed argument string within " to avoid problems with
				// spaces within arguments
				m_startArguments += " \"";
				m_startArguments += argStr;
				m_startArguments += "\"";
			}
		}
	}
	CATCH_ALL(TimedExec-ParseCommandLine,succeeded);
	if (!succeeded)
	{
		DisplayUsage();
	}
	else if (m_traceLevel >= 3)
	{
		DisplaySettings();
	}
	if (!succeeded)
		m_executionStatus = eUSAGE_ERROR;
	return succeeded;

} // TimedExecutor::ParseCommandLine

//-----------------------------------------------------------------------------

bool TimedExecutor::CheckEnvironmentVariables (void)
{
	bool succeeded = true;
	try
	{
		std::string envVarVal;

		// only repeat reading, interpreting already done within SetTraceLevel
		ReadEnvironmentVariable("TimedExec_Trace",envVarVal);

		if (ReadEnvironmentVariable("TimedExec_MinWaitTime",envVarVal))
		{
			int minWaitTimeSec;
			std::istringstream is (envVarVal);
			is >> minWaitTimeSec;
			if (minWaitTimeSec > 0)
			{
				if (m_waitTimeSec < minWaitTimeSec) m_waitTimeSec = minWaitTimeSec;
			}
		}

		if (ReadEnvironmentVariable("TimedExec_MaxWaitTime",envVarVal))
		{
			int maxWaitTimeSec;
			std::istringstream is (envVarVal);
			is >> maxWaitTimeSec;
			if (maxWaitTimeSec >= 0)
			{
				if (m_waitTimeSec > maxWaitTimeSec) m_waitTimeSec = maxWaitTimeSec;
			}
		}

		if (ReadEnvironmentVariable("TimedExec_RestrictToSingleInstance",envVarVal))
		{
			m_restrictToSingleInstance = (envVarVal == "YES");
			if ((envVarVal != "YES") && (envVarVal != "NO"))
			{
				if (m_traceLevel >= 1)
					cout << "[TimedExec-CheckEnvironmentVariables]: WARNING: TimedExec_RestrictToSingleInstance has value "
					<< envVarVal.c_str() << " but allowed is only YES or NO. Assuming option NO!" << endl;
			}
		}

		if (ReadEnvironmentVariable("TimedExec_MaxWaitTimeBeforeStartSec",envVarVal))
		{
			int maxWaitTimeBeforeStartSec;
			std::istringstream is (envVarVal);
			is >> maxWaitTimeBeforeStartSec;
			if (maxWaitTimeBeforeStartSec >= 0)
			{
				m_maxWaitTimeBeforeStartSec = maxWaitTimeBeforeStartSec;
			}
			else
			{
				if (m_traceLevel >= 1)
					cout << "[TimedExec-CheckEnvironmentVariables]: WARNING: TimedExec_MaxWaitTimeBeforeStartSec < 0, value is ignored!" << endl;
			}
		}

		if (ReadEnvironmentVariable("TimedExec_SkipExecution",envVarVal))
		{
			m_skipExecution = (envVarVal == "YES");
			if ((envVarVal != "YES") && (envVarVal != "NO"))
			{
				if (m_traceLevel >= 1)
					cout << "[TimedExec-CheckEnvironmentVariables]: WARNING: TimedExec_SkipExecution has value "
					<< envVarVal.c_str() << " but allowed is only YES or NO. Assuming option NO!" << endl;
			}
		}
		if (ReadEnvironmentVariable("TimedExec_LogResults",envVarVal))
		{
			m_logResults = (envVarVal == "YES");
			if ((envVarVal != "YES") && (envVarVal != "NO"))
			{
				if (m_traceLevel >= 1)
					cout << "[TimedExec-CheckEnvironmentVariables]: WARNING: TimedExec_LogResults has value "
					<< envVarVal.c_str() << " but allowed is only YES or NO. Assuming option NO!" << endl;
			}
		}
		m_logFilePath = std::string();
		if (ReadEnvironmentVariable("TimedExec_LogFilePath",envVarVal))
		{
			m_logFilePath = envVarVal;
		}
		if (m_logFilePath.empty())
		{
			// Get complete path for exe file (e.g. "C:\MyTargetBin\MyAppname.exe")
			char appName [MAX_PATH+1];
			::GetModuleFileNameA (0, appName, MAX_PATH+1);
			std::string exeFileFullPath(appName);

			// Get directory path (e.g. "C:\MyTargetBin")
			std::string::size_type pos = exeFileFullPath.find_last_of("\\");
			if (pos!=std::string::npos)
			{
				m_logFilePath = exeFileFullPath.substr(0, pos);
			}
		}
		m_logFilePath = m_logFilePath + "\\" + "TimedExec_Log.txt";
	}
	CATCH_ALL(TimedExec-CheckEnvironmentVariables,succeeded);
	if (!succeeded)
	{
		DisplayUsage();
	}
	else if (m_traceLevel >= 2)
	{
		DisplaySettings();
	}
	if (!succeeded)
		m_executionStatus = eUSAGE_ERROR;
	return succeeded;

} // TimedExecutor::CheckEnvironmentVariables

//-----------------------------------------------------------------------------

bool TimedExecutor::ExecutionIsNotSkipped (int io_exitCode)
{
	if (m_skipExecution)
	{
		m_executionStatus = eEXECUTION_SKIPPED;
		io_exitCode = 0; // assume positive execution status because start of executable is suppressed
		if (m_traceLevel > 0) 
			cout << "[TimedExec]: HINT: Execution is skipped (to activate change environment variable TimedExec_SkipExecution)!" << endl;
	}

	return !m_skipExecution;

} // TimedExecutor::ExecutionIsNotSkipped

//-----------------------------------------------------------------------------

bool TimedExecutor::ExeFileIsAccessible (void)
{
	bool succeeded = false;
	try
	{
		if (_access (m_executableToStartFullPath.c_str(), 0) == 0) // file exists
		{
			if (_access (m_executableToStartFullPath.c_str(), 4) == 0) // file is readable
			{
				succeeded = true;
			}
			else // file is not readable
			{
				if (m_traceLevel > 0) 
					cout << "[TimedExec]: ERROR: File " << m_executableToStartFullPath.c_str()
					<< " is not readable" << endl;
			}
		}
		else // file does not exist
		{
			if (m_traceLevel > 0) 
				cout << "[TimedExec]: ERROR: File " << m_executableToStartFullPath.c_str()
				<< " does not exist" << endl;
		}
	}
	CATCH_ALL(TimedExec-ExeFileIsAccessible,succeeded)
		if (!succeeded)
			m_executionStatus = eEXE_FILE_NOT_ACCESSIBLE;
	return succeeded;

} // TimedExecutor::ExeFileIsAccessible

//-----------------------------------------------------------------------------

bool TimedExecutor::AllocateResourcesForStart (void)
{
	bool succeeded = false;
	try
	{
		if (m_restrictToSingleInstance)
		{
			long lastErr;

			// --- Open/create mutex ---

			if (m_traceLevel >=2) 
				cout << "[TimedExec]: CreateMutex..." << endl;
			m_hMutexSingleInstance = CreateMutex (0, FALSE, "TimedExec_SingleInstanceMutex");
			lastErr = GetLastError();
			if (!m_hMutexSingleInstance)
			{
				std::ostringstream os;
				os << "ERROR: accessing/creating mutex failed LastError="
					<< lastErr << "/" << GetSystemErrorDescription(lastErr).c_str() << endl;
				throw std::exception(os.str().c_str());
			}
			if (lastErr == ERROR_ALREADY_EXISTS)
			{
				if (m_traceLevel >=1) 
					cout << "[TimedExec]: Existing mutex opened (some other instance is running)" << endl;
			}
			else
			{
				if (m_traceLevel >=2) 
					cout << "[TimedExec]: New mutex created" << endl;
			}

			// --- Wait for mutex ---

			if (m_maxWaitTimeBeforeStartSec > 0)
			{ 
				if (    (m_traceLevel > 1)
					|| ((m_traceLevel > 0) && (lastErr == ERROR_ALREADY_EXISTS)))
					cout << "[TimedExec]: Waiting " << m_maxWaitTimeBeforeStartSec
					<< " sec for start permission..." << endl;
			}
			else
			{ 
				if (    (m_traceLevel > 1)
					|| ((m_traceLevel > 0) && (lastErr == ERROR_ALREADY_EXISTS)))
					cout << "[TimedExec]: Waiting INFINITE for start permission..." << endl;
			}

			DWORD waitRetVal = WaitForSingleObject(m_hMutexSingleInstance,
				m_maxWaitTimeBeforeStartSec > 0 ? m_maxWaitTimeBeforeStartSec*1000 : INFINITE);

			switch (waitRetVal)
			{
			case WAIT_OBJECT_0:
				// the ressource mutex is available before timeout elapsed
				succeeded = true;
				m_mutexAquired = true;
				if (m_traceLevel > 1)
					cout << "[TimedExec]: Start permission granted" << endl;
				break;

			case WAIT_ABANDONED:
				// the ressource mutex is available before timeout elapsed
				// other process may have terminated irregularly
				succeeded = true;
				m_mutexAquired = true;
				if (m_traceLevel > 1)
					cout << "[TimedExec]: Start permission granted (other process terminated irregularly)" << endl;
				break;

			case WAIT_TIMEOUT:
				// the ressource mutex was not available within the specified time period
				m_executionStatus = eTIMEOUT_START_PERMISSION;
				if (m_traceLevel > 0) 
					cout << "[TimedExec]: ERROR: Mutex-Timeout of " << m_maxWaitTimeBeforeStartSec 
					<< " sec elapsed. Program will not be started!" << endl;
				break;

			default: // WAIT_FAILED
				// system call failed
				{
					lastErr = GetLastError();
					std::ostringstream os;
					os << "ERROR: Waiting for start permission (mutex) failed LastError="
						<< lastErr << "/" << GetSystemErrorDescription(lastErr).c_str() << endl;
					throw std::exception(os.str().c_str());
				}
				break;
			}
		}
		else // allow multiple execution instances
		{
			// nothing to be done, immediate start without check is possible
			succeeded = true;
		}
	}
	CATCH_ALL(TimedExec-AllocateResourcesForStart,succeeded)
		return succeeded;

} // TimedExecutor::AllocateResourcesForStart

//-----------------------------------------------------------------------------

bool TimedExecutor::StartExeFile (
	int &          io_numEventsToWaitFor,
	EventInfoMap & io_eventInfos,
	HANDLE         in_hStdOut,
	HANDLE         in_hStdErr)
{
	long dwFlags = STARTF_USEPOSITION | STARTF_USESHOWWINDOW;

	PROCESS_INFORMATION piProcInfo; 
	STARTUPINFO siStartInfo;
	BOOL bSuccess = FALSE; 

	ZeroMemory( &m_procInfo, sizeof(PROCESS_INFORMATION) );
	ZeroMemory( &piProcInfo, sizeof(PROCESS_INFORMATION) );

	ZeroMemory( &siStartInfo, sizeof(STARTUPINFO) );
	siStartInfo.cb = sizeof(STARTUPINFO);
	if (m_checkForMemleaks)
	{
		siStartInfo.hStdError  = in_hStdErr;
		siStartInfo.hStdOutput = in_hStdOut;
		siStartInfo.dwFlags   |= STARTF_USESTDHANDLES;
	}

	bool succeeded = false;
	try
	{
		std::string cmdStr = m_executableToStartFullPath;
		if (m_startArguments.length()>0)
		{
			cmdStr += m_startArguments;
		}

		ostringstream oss_retry;
		if (m_numRetriesUsed > 0)
		{
			oss_retry << "retry " << m_numRetriesUsed << " ";
		}

		if (CreateProcess (
			NULL,
			LPTSTR (cmdStr. c_str()),
			NULL,
			NULL,
			TRUE, // handles are inherited
			0,
			NULL,
			NULL,
			&siStartInfo,
			&piProcInfo))
		{
			if (m_traceLevel > 0)
			{
				ostringstream oss;
				oss << "[TimedExec]: " << oss_retry.str() << "Successfully started " << cmdStr.c_str() << endl;
				printf("%s", oss.str().c_str()); // force synchronized output
			}
			m_procInfo.hProcess = piProcInfo.hProcess;
			StoreStartTime();
			++io_numEventsToWaitFor;
			io_eventInfos[m_procInfo.hProcess]=EventInfo(HandleType::PROCESS, m_procInfo.hProcess);
			succeeded = true;
			CloseHandle (piProcInfo. hThread); // we dont need the thread handle
		}
		else // exe could not be started
		{
			int err = GetLastError();
			if (m_traceLevel > 0) 
				cout << "[TimedExec]: " << oss_retry.str() << "ERROR: File " << m_executableToStartFullPath.c_str() << " could not be started (LastError="
				<< err << "/" << GetSystemErrorDescription(err).c_str() << ")" << endl;
		}
	}
	CATCH_ALL(TimedExec-StartExeFile,succeeded);
		if (!succeeded)
			m_executionStatus = eEXE_COULD_NOT_BE_STARTED;
	return succeeded;

} // TimedExecutor::StartExeFile

//-----------------------------------------------------------------------------

int TimedExecutor::HandleEventProcessTerminated(HANDLE handle, EventInfoMap&  eventInfoMap)
{
	int exitCode = 1; // assuming failed execution

	// the program terminated before timeout elapsed
	// now get the exit code of the external program
	DWORD externalExitCode;
	StoreStopTime();
	if (::GetExitCodeProcess(m_procInfo.hProcess, &externalExitCode))
	{
		if (m_traceLevel > 0) 
			cout << "[TimedExec]: Program " << m_executableToStartFullPath.c_str()
			<< " returned exit code " << externalExitCode << endl;
		// large error codes aren't detected by VS2005
		exitCode = 0 == externalExitCode ? 0 : 1;
		m_executionStatus = exitCode ? eEXE_FAILED : eEXE_SUCCEEDED;
	}
	else // could not get exit code
	{
		int err = GetLastError();
		if (m_traceLevel > 0) 
			cout << "[TimedExec-WaitForTerminationOfExeFile]: ERROR: Could not get exit code for "
			<< m_executableToStartFullPath.c_str()
			<< " (LastError="
			<< err << "/" << GetSystemErrorDescription(err).c_str()
			<< ")" << endl;
	}
	eventInfoMap[handle].SetFinished();

	return exitCode;
}

//-----------------------------------------------------------------------------

bool TimedExecutor::WaitIsNeeded (
	CaptureOutput& captureOutput,
	DWORD          startTime,
	DWORD          maxEndTime,
	int&           out_numEventsToWaitFor,
	DWORD&         out_waitTime,
	EventInfoMap&  eventInfoMap,
	MyEventHandles&     io_handles)
{
	out_numEventsToWaitFor = 0;
	out_waitTime           = 0;
	bool waitNeeded        = false;

	// Check for the handles to wait for
	for (EventInfoMap::iterator it = eventInfoMap.begin(); it != eventInfoMap.end(); ++it)
	{
		EventInfo& eventInfo = it->second;
		if (eventInfo.active)
		{
			if (eventInfo.type==HandleType::PROCESS)
			{
				DWORD curTime = GetTickCount();
				DWORD MAX_DWORD = 4294967295;
				DWORD timeElapsed = (curTime >= startTime) ? curTime - startTime : MAX_DWORD - startTime + curTime;
				if (timeElapsed >= static_cast<unsigned long>(m_waitTimeSec)*1000)
				{
					// Set minimum wait time of 1 ms to allow handling of timeout within general routine
					// at a single location in source code
					out_waitTime = 1;
				}
				else
				{
					out_waitTime = m_waitTimeSec*1000 - timeElapsed;
				}
				io_handles[out_numEventsToWaitFor]=eventInfo.handle;
				++out_numEventsToWaitFor;
				waitNeeded = true;
			}
			else if (eventInfo.type==HandleType::PIPE_STD_OUT)
			{
				if (captureOutput.IsActive())
				{
					io_handles[out_numEventsToWaitFor]=eventInfo.handle;
					++out_numEventsToWaitFor;
					waitNeeded = true;
				}
			}
		}
	}

	if (waitNeeded && !out_waitTime && captureOutput.IsActive())
    // a timeout to wait for end of client process was not set
	{
		 // wait forr reading from pipe, output may be received after process has terminated
		out_waitTime = captureOutput.GetWaitTime();
	}

	return waitNeeded;
}

//-----------------------------------------------------------------------------

int TimedExecutor::WaitForTerminationOfExeFile (CaptureOutput& captureOutput, EventInfoMap& eventInfoMap)
{
	int exitCode = 1; // assume failed execution
	bool dummy =true;
	try
	{

		bool wait=true;
		DWORD startTime = GetTickCount();
		DWORD waitTime = m_waitTimeSec > 0
			? m_waitTimeSec*1000
			: 3600*24*1000; // assume 24 h if waittime <= 0
		DWORD maxEndTime = startTime + waitTime;

		MyEventHandles eventHandles;
		while (wait) 
		{ 
			int numEventsToWaitFor = 0;
			DWORD timeout = 0;

			if (!WaitIsNeeded(captureOutput, startTime, maxEndTime, numEventsToWaitFor, timeout,
				eventInfoMap, eventHandles))
				break;

			DWORD waitRetVal = WaitForMultipleObjects( 
				numEventsToWaitFor, // number of event objects 
				eventHandles,       // array of event objects 
				FALSE,              // does not wait for all 
				timeout);           // waits indefinitely 

			if (waitRetVal == WAIT_TIMEOUT)
			{
				// Check if timeout occurred when waiting only for pipe data
				if (numEventsToWaitFor == 1)
				{
					EventInfo eventInfo = eventInfoMap[eventHandles[0]];
					if (eventInfo.type = HandleType::PIPE_STD_OUT)
					{
						if (m_traceLevel > 0) 
							cout << "[TimedExec]: ERROR: Timeout when waiting for pipe data!" << endl;
						wait = false;
						break;
					}
				}
				// the program did not terminate within the specified time period
				m_executionStatus = eTIMEOUT_EXE_TERMINATION;
				if (m_traceLevel > 0) 
					cout << "[TimedExec]: ERROR: Timeout of " << m_waitTimeSec 
					<< " sec elapsed before program has terminated!" << endl;
				// try to kill the program
				if (KillProcess())
				{
					if (m_traceLevel > 0) 
						cout << "[TimedExec]: Killed " << m_executableToStartFullPath.c_str() << endl;
				}
				wait=false;
				break;
			}
			else if ((waitRetVal >= WAIT_OBJECT_0) && ((waitRetVal <= WAIT_OBJECT_0 + 1)))
				// we wait for one of two events
			{
				DWORD idx = waitRetVal - WAIT_OBJECT_0;
				HANDLE handle = eventHandles[idx];
				if (eventInfoMap[handle].active)
				{
					HandleType::Enum handleType = eventInfoMap[handle].type;
					if (handleType == HandleType::PROCESS)
					{
						exitCode = HandleEventProcessTerminated(handle, eventInfoMap);
					}
					else if (handleType == HandleType::PIPE_STD_OUT)
					{
						captureOutput.HandleEventReadPipeData(handle, eventInfoMap);
					}
				}
				else
				{
					std::cout << "[TimedExec]: ERROR: trigger for inactive event received!" << std::endl;
				}

			}
			else // default
			{
				// system call failed
				WriteSystemErrorMsg (m_traceLevel >= 1,
					"[TimedExec]: WaitForMultipleObjects failed");
				wait=false;
			}

		}
		CloseHandle (m_procInfo. hProcess);

		if (captureOutput.HasDetectedMemLeaks(m_executionStatus == eEXE_SUCCEEDED))
		{
			exitCode = 1;
			m_executionStatus = eEXE_SUCCEEDED_WITH_MEMLEAKS;
		}

	}
	CATCH_ALL(TimedExec-WaitForTerminationOfExeFile,dummy);
	return exitCode;
}

//-----------------------------------------------------------------------------

void TimedExecutor::LogResults (void)
{
	const long WIDTH_EXE_STATUS = 56;
	const long WIDTH_EXE_FILE   = 40;
	const long WIDTH_DETAILS    = 70;

	if (m_logResults)
	{
		std::ostringstream osLeft;
		std::ostringstream osDetails;
		if (m_retryIsPending 
			&& (
			(m_executionStatus != eEXECUTION_SKIPPED)
			&& (m_executionStatus != eEXE_SUCCEEDED)
			))
		{
			osLeft << " <RETRY_PENDING>";
		}
		switch (m_executionStatus)
		{
		case eEXECUTION_SKIPPED:
			{
				osLeft << GetExecutionStatusAsString("WARNING");
			}
			break;

		case eEXE_SUCCEEDED:
			{
				osLeft    << "   SUCCESS";
				osDetails << "   StartTime: " << m_dateTimeExeWasStarted
					<< "   Execution time: " << m_executionTimeSec << " sec";
			}
			break;

		case eEXE_FAILED:
		case eEXE_SUCCEEDED_WITH_MEMLEAKS:
			{
				osLeft    << GetExecutionStatusAsString("FAILURE");
				osDetails << "   StartTime: " << m_dateTimeExeWasStarted
					<< "   Execution time: " << m_executionTimeSec << " sec";
			}
			break;

		case eTIMEOUT_EXE_TERMINATION:
			{
				osLeft    << GetExecutionStatusAsString("FAILURE");
				osDetails << "   StartTime: " << m_dateTimeExeWasStarted
					<< "   Timeout of " << m_waitTimeSec << " sec elapsed";
			}
			break;

		case eTIMEOUT_START_PERMISSION:
			{
				osLeft    << GetExecutionStatusAsString("FAILURE");
				osDetails << "   Timeout of " << m_maxWaitTimeBeforeStartSec << " sec elapsed";
			}
			break;

		case eUNSPECIFIED_ERROR:
		case eUSAGE_ERROR:
		case eEXE_FILE_NOT_ACCESSIBLE:
		case eEXE_COULD_NOT_BE_STARTED:
		default:
			{
				osLeft << GetExecutionStatusAsString("FAILURE");
			}
			break;
		}
		std::ostringstream os;
		os << GetCurrentDateTimeAsString();
		os << std::setw(WIDTH_EXE_STATUS) << std::left << osLeft.str()
			<< std::setw(WIDTH_EXE_FILE)   << std::left << m_executableToStartFileName
			<< std::setw(WIDTH_DETAILS)    << std::left << osDetails.str()
			<< "   CommandLine: " << ::GetCommandLine();
		std::string info = os.str();

		// Write to stdout
		if (m_traceLevel > 0)
		{
			ostringstream oss;
			oss << "[TimedExec]: " << info << endl;
			printf ("%s", oss.str().c_str()); // force snchronized output
		}

		// Write to logfile
		if (m_mutexAquired)
		{
			// File access is secure because multiple processes
			// are synchronized via mutex
			WriteLineToLogfile(info);
		}
	}

} // TimedExecutor::LogResults

//-----------------------------------------------------------------------------

void TimedExecutor::ReleaseResources (void)
{
	if (m_restrictToSingleInstance && m_hMutexSingleInstance)
	{
		if (m_mutexAquired)
		{
			ReleaseMutex(m_hMutexSingleInstance);
			m_mutexAquired = false;
		}
		CloseHandle (m_hMutexSingleInstance);
		m_hMutexSingleInstance = 0;
	}

} // TimedExecutor::ReleaseResources

//-----------------------------------------------------------------------------

int TimedExecutor::Run(int argc, _TCHAR* argv[])
{
	int exitCode = 1; // assume failed execution

	if (   SetTraceLevel(argc,argv) 
		&& DumpCallInfo(argc,argv)
		&& ParseCommandLine(argc,argv) 
		&& CheckEnvironmentVariables()
		&& ExecutionIsNotSkipped(exitCode))
	{
		do 
		{
			EventInfoMap eventInfoMap;
			int numEventsToWaitFor = 0;
			HANDLE hStdOut = 0;
			HANDLE hStdErr = 0;

			CaptureOutput captureOutput(m_executableToStartFullPath, m_checkForMemleaks, m_traceLevel);

			if (   ExeFileIsAccessible()
				&& AllocateResourcesForStart()
				&& captureOutput.Prepare(numEventsToWaitFor, eventInfoMap, hStdOut, hStdErr)
				&& StartExeFile(numEventsToWaitFor, eventInfoMap, hStdOut, hStdErr)
				&& captureOutput.ReleaseResourcesAfterStartOfExeFile())
			{
				exitCode = WaitForTerminationOfExeFile(captureOutput, eventInfoMap);
			}

			if ((exitCode != 0) && (m_numRetriesAllowed > 0))
			{
				m_retryIsPending = true;
				++m_numRetriesUsed;
				--m_numRetriesAllowed;
			}
			else
			{
				m_retryIsPending = false;
			}

			LogResults();
			ReleaseResources();

		} while (m_retryIsPending);
	}

	return exitCode;

} // TimedExecutor::Run

//-----------------------------------------------------------------------------

///\endcond do not document
