////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
////////////////////////////////////////////////////////////////////////////////

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <tchar.h>
#include <string>

#include "EventInfo.h"

#pragma once

class CaptureOutput;

/// Class which stores all information about the executable to start and
/// performs the starting of the executable and the waiting for its termination
class TimedExecutor
{
public:

	/// Constructor
	TimedExecutor();

	/// Start the executable specified in command line args and wait for
	/// its termination
	int Run (int argc, _TCHAR* argv[]);

private:

	// Within this application we are waiting for one or two events
    typedef HANDLE MyEventHandles[2];

	//----- Basic control -----

	// Executable to start with belonging command line params
	std::string m_executableToStartFullPath;
	std::string m_executableToStartFileName;
	std::string m_startArguments;

	// Number of seconds to wait for program termination
	long m_waitTimeSec;

	// Information about the started process
	PROCESS_INFORMATION m_procInfo;

	// Tracelevel to control the amount of information written to stdout.
	// 0 = no output at all, 2 = maximum output
	int m_traceLevel;

	//----- Check for memory leaks -----

	// In case of memory checking the output of the test application is redirected
	// to an text file. The contents of the text file will be checked when the
	// test application has terminated. Search pattern is "Detected memory leaks".
	// A found memory leak will cause an exit code of 1.

	// Flag indicating whether check for memory leaks is active
	bool m_checkForMemleaks;

	// Handle to text file with captured output of test application
	HANDLE m_hApplicationOutputFile;

	// Name of the file with captured output from test application
	// (e.g. TestApp.exe.AppOut.txt)
	std::string m_nameApplicationOutputFile;

	//----- Restrict to execution of single process -----

	// It is possible to allow only the execution of a single TimedExec instance
	// on a computer. If this feature is active and the program detects an
	// already running instance it will wait until the other instance terminates
	// (limited by Timeout of m_maxWaitTimeBeforeStartSec).
	// Possible usage in reality: Modern processors use parallel compiling of several
	// projects/makefile within a bigger project/makefile. Within the PostbuildStep
	// the unit tests may be executed. If several unit tests need the same resource
	// they can be synchronized simply by calling them via TimedExec. There is no need
	// to implement synchronization features within thre test code itself.
	bool m_restrictToSingleInstance;
	long m_maxWaitTimeBeforeStartSec;

	// Mutex used for synchronization
	HANDLE m_hMutexSingleInstance;

	// Flag indicating whether the mutex was aquired succesfully
	bool m_mutexAquired;

	//----- Info about execution results -----

	// Flag indicating whether the execution results are written to stdout
	// If we are running with property m_restrictToSingleInstance=true the
	// results are also written to a central logfile
	bool m_logResults;

	// Path for the logfile
	std::string m_logFilePath;

	// Info about execution status
	enum ExecutionStatus
	{
		eUNSPECIFIED_ERROR,
		eUSAGE_ERROR,
		eEXECUTION_SKIPPED,
		eEXE_FILE_NOT_ACCESSIBLE,
		eTIMEOUT_START_PERMISSION,
		eEXE_COULD_NOT_BE_STARTED,
		eTIMEOUT_EXE_TERMINATION,
		eEXE_FAILED,
		eEXE_SUCCEEDED,
		eEXE_SUCCEEDED_WITH_MEMLEAKS
	};
	ExecutionStatus m_executionStatus;

	// TickCount at time when exe was really started
	long m_startExecutableTickCount;
	// Formatted Date/Time string
	std::string m_dateTimeExeWasStarted;

	// Number of seconds the executable needed to execute
	double m_executionTimeSec;


	// --- allow deactivation of program ---

	// A value of true leads to immediate programm termination with exit code 0
	// No waiting for other instances and no start of given exe file is performed
	bool m_skipExecution;


	// --- allow retry ---

	// if program execution fails and there are retries left another start of the executable is performed
	long m_numRetriesAllowed;
	long m_numRetriesUsed;
	bool m_retryIsPending;

private:

	// Helper functions
	bool DumpCallInfo (int argc, _TCHAR* argv[]);
	void DisplayCallContext (void);
	void DisplaySettings (void);
	void DisplayUsage (void);
	bool ReadEnvironmentVariable (
		std::string const & in_envVarName,
		std::string&        out_envVarVal);
	bool KillProcess (void);
	std::string GetExecutionStatusAsString (std::string const & in_rInfo);
	void StoreStartTime();
	void StoreStopTime();
	std::string GetCurrentDateTimeAsString (void);
	void WriteLineToLogfile (std::string const & in_line);

	// Execution steps
	bool SetTraceLevel (int argc, _TCHAR* argv[]);
	bool ParseCommandLine (int argc, _TCHAR* argv[]);
	bool CheckEnvironmentVariables (void);
	bool ExecutionIsNotSkipped (int io_exitCode);
	bool ExeFileIsAccessible (void);
	bool AllocateResourcesForStart (void);
	bool StartExeFile (
		int &          io_numEventsToWaitFor,
		EventInfoMap & io_eventInfos,
		HANDLE         in_hStdOut,
		HANDLE         in_hStdErr);

	bool WaitIsNeeded (
		CaptureOutput&  captureOutput,
		DWORD           startTime,
		DWORD           maxEndTime,
		int&            out_numEventsToWaitFor,
		DWORD&          out_waitTime,
		EventInfoMap&   eventInfoMap,
		MyEventHandles& io_handles);

	int HandleEventProcessTerminated(HANDLE handle, EventInfoMap& eventInfoMap);
	int  WaitForTerminationOfExeFileOld (void);
	int  WaitForTerminationOfExeFile (CaptureOutput& captureOutput, EventInfoMap& eventInfoMap);
	void CheckForMemLeaksWithinOutputFile (CaptureOutput& captureOutput, int& io_exitCode);
	void LogResults (void);
	void ReleaseResources (void);

}; // class TimedExecutor

