/// \file
/// \author Gerald Fahrnholz

#pragma once

#include "CommonSources/CommonDefinitions.h"
#include "WaitableCondition.h"

namespace TestToolBox
{
    //-------------------------------------------------------------------------

    TTB_INLINE WaitableCondition::WaitableCondition (bool in_autoReset)
        : m_autoReset  (in_autoReset)
        , m_condIsTrue (false)
    {
    }

    //-------------------------------------------------------------------------

    TTB_INLINE void WaitableCondition::Set (bool in_state)
    { 
        boost::mutex::scoped_lock lock(m_mutex);
        m_condIsTrue = in_state;
        if (m_condIsTrue ) m_condition.notify_one();
    }

    //-------------------------------------------------------------------------

    TTB_INLINE bool WaitableCondition::IsTrue (void)
    {
        boost::mutex::scoped_lock lock(m_mutex);
        return m_condIsTrue;
    }

    //-------------------------------------------------------------------------

    TTB_INLINE void WaitableCondition::WaitUntilTrue (void)
    {
        boost::mutex::scoped_lock lock(m_mutex); 
        if (!m_condIsTrue)
        {
             // we really have to wait until someone calls
             // Set(true)
             m_condition.wait(lock);
        }
        if (m_autoReset) m_condIsTrue = false;
    }

    //-------------------------------------------------------------------------

    TTB_INLINE bool WaitableCondition::TimedWaitUntilTrue (long in_timeoutMs)
    {
        boost::mutex::scoped_lock lock(m_mutex);
        if (!m_condIsTrue)
        {
            boost::xtime timeInterval = 
                ConvertFromMilliSec(in_timeoutMs);
            if (!m_condition.timed_wait(lock, timeInterval))
                return false;
        }
        if (m_autoReset) m_condIsTrue = false;
        return true;
    }

    //-------------------------------------------------------------------------

    TTB_INLINE boost::xtime WaitableCondition::ConvertFromMilliSec (
        long in_timeoutMs)
    {
        double secDouble  = in_timeoutMs / 1000.0;
        int    secInt     = static_cast<int>(secDouble);
        int    nanoSecInt = static_cast<int>((secDouble - secInt) * 1000000000.0);
        boost::xtime xt;
        boost::xtime_get(&xt, boost::TIME_UTC_);
        xt.sec  += secInt;
        xt.nsec += nanoSecInt;
        while (xt.nsec >= 1000000000)
        {
            ++xt.sec;
            xt.nsec -= 1000000000;
        }
        return xt;
    }

    //-------------------------------------------------------------------------
}
