////////////////////////////////////////////////////////////////////////////////
///
/// \file
/// \author Gerald Fahrnholz
/// 
/// \brief
/// Implementation of class XmlCheck
///
////////////////////////////////////////////////////////////////////////////////

#pragma once

#include "XmlCheck.h"
#include "IEventReceiver.h"
#include "CommonSources\CommonDefinitions.h"
#include "TestEvents.h"

#ifdef TTB_USE_PUGI_XML
#include <PugiXml/pugixml.hpp>
#endif TTB_USE_PUGI_XML

#include <iostream>
#include <iomanip>

///\cond do not document

//-----------------------------------------------------------------------------

namespace TestToolBox
{
    class XmlCheckImpl;
    typedef std::shared_ptr<XmlCheckImpl> XmlCheckImplSp;

    //----------------------------------------------------------------------------
    // Header of class XmlCheckImpl
    //----------------------------------------------------------------------------

    class XmlCheckImpl : public XmlCheck
    {
    public:
        // Constructor
        XmlCheckImpl();

        // Destructor
        virtual ~XmlCheckImpl();

        void Dump(EventContext in_context) const override;

        IEventReceiver* SetEventReceiver(IEventReceiver* in_pEventReceiver) override;

        // General settings
        void SetDefaults(void) override;
        SettingsBaseSp CopyCurrentSettings() const override;
        void RestoreCurrentSettings(SettingsBaseSp in_spSettings) override;
        void SetOutputModeForWritingXmlContents(EventContext in_outputMode) override;
        void SetDetailedOutput(bool in_detailedOutput) override;
        void IgnoreAttribute(std::string const & in_attribute, bool in_ignore = true) override;
        bool IsAttributeToIgnore(std::string const & in_attribute) const override;
        void IgnoreNode(std::string const & in_nodeName, bool in_ignore = true) override;
        bool IsNodeToIgnore(std::string const & in_nodeName) const override;
        void UseAsDoubleAttribute(std::string const & in_attribute, bool in_useAsDouble = true, int in_precision = 4) override;
        bool IsDoubleAttribute(std::string const & in_attribute) const override;
        int GetPrecisionOfDoubleAttribute(std::string const & in_attribute) const override;
        void UseAsDoubleNode(std::string const & in_nodeName, bool in_useAsDouble = true, int in_precision = 4) override;
        bool IsDoubleNode(std::string const & in_nodeName) const override;
        int GetPrecisionOfDoubleNode(std::string const & in_nodeName) const override;
        void UseAsPointerAttribute(std::string const & in_attribute, bool in_useAsPointer = true) override;
        bool IsPointerAttribute(std::string const & in_attribute) const override;
        void UseAsPointerNode(std::string const & in_nodeName, bool in_useAsPointer = true) override;
        bool IsPointerNode(std::string const & in_nodeName) const override;
        void SetFuncToChangeAttributeValue(XmlServices::FuncToChangeAttributeValue in_func) override;

        // Check XML contents
        void CheckNode(pugi::xml_node const &  in_rNode, char const * in_file, long in_lineNum) override;
        void CheckSubNode(pugi::xml_node const &  in_rNode, std::string const & in_xPathExpression, char const * in_file, long in_lineNum) override;
        void CheckSubNodeWithMaxChilds(int in_maxChilds, pugi::xml_node const &  in_rNode, std::string const & in_xPathExpression, char const * in_file, long in_lineNum) override;
        bool Compare(pugi::xml_node const & in_rNodeA, pugi::xml_node const & in_rNodeB, char const * in_file, long in_lineNum) override;
        std::string GetCompareResult(void) const override;
        void CheckNode(pugi::xml_document const & in_rDoc, char const * in_file, long in_lineNum) override;
        void CheckSubNode(pugi::xml_document const &  in_rDoc, std::string const & in_xPathExpression, char const * in_file, long in_lineNum) override;
        void CheckSubNodeWithMaxChilds(int in_maxChilds, pugi::xml_document const &  in_rDoc, std::string const & in_xPathExpression, char const * in_file, long in_lineNum) override;
        bool Compare(pugi::xml_document const &  in_rDocA, pugi::xml_document const &  in_rDocB, char const * in_file, long in_lineNum) override;
        void ApplyFilterToXmlTree(pugi::xml_node& io_docFromFile) override;
        void ApplyFilterToXmlTree(pugi::xml_document& io_docFromFile) override;
        pugi::xml_node ReadFromFile(pugi::xml_document& io_docFromFile, std::string const & in_xmlFilePath) override;
        bool           SaveToFile(pugi::xml_document& io_docFromFile, std::string const & in_xmlFilePath) override;
        pugi::xml_node ReadFromString(pugi::xml_document& io_docFromString, std::string const & in_xmlTreeAsString) override;

    private:
        /// Interface to pass all kinds of test events to
        IEventReceiver* m_pEventReceiver;

        // General settings changeable by the user
        XmlCheckSettingsSp m_spSettings;

        /// Indenting to be used when writing an XML tree to test output
        static const unsigned int NUM_INDENTS_PER_SPACE = 2;

        /// Result of last call to Compare()
        std::string m_compareResult;

        struct TextWithPosInfo
        {
            TextWithPosInfo()
                : m_row(0)
                , m_col(0)
                , m_fullPath()
                , m_text()
            {}

            TextWithPosInfo(
                int                 in_row,
                int                 in_col,
                std::string const & in_fullPath,
                std::string const & in_text)
                : m_row(in_row)
                , m_col(in_col)
                , m_fullPath(in_fullPath)
                , m_text(in_text)
            {}

            int         m_row;
            int         m_col;
            std::string m_fullPath;
            std::string m_text;
        };

        // Each line contains an XML entry plus additional infos for
        // better error reporting
        typedef std::vector<TextWithPosInfo> FormattedContents;

    private:
        /// Helper function to produce more detailed output
        std::string IfDetailed(std::string const & in_str);

        /// Formating
        const char * IndentStr(unsigned int in_numIndents);
        const char * IndentEmptyStr(unsigned int in_numIndents);

        /// Write given XML node to EventReceiver or given optional container
        void WriteNode(
            pugi::xml_node const & in_rNode,    ///< XML (sub)tree to check
            unsigned int in_indent,             ///< indentation for pretty format 
            int          in_maxChildNodes = 0,  ///<maximum number of direct childs to check, 0 : all childs
            char const * in_file = 0,           ///< optional file name used in case of error
            long         in_lineNum = 0,        ///< optional line number used in case of error
            FormattedContents* in_pContents = 0 ///< optional container to write output to instead
            );       							///< of using configured EventReceiver


        /// \brief Writes attributes of given XML node to EventReceiver or given optional container
        ///
        /// Each attribute and text is reported with a separate call
        /// to EventReceiver.
        void WriteAttributes(
            pugi::xml_node const & in_rNode,     ///< XML (sub)tree to check
            unsigned int  in_indent,             ///< indentation for pretty format 
            char const *  in_file = 0,           ///< optional file name used in case of error
            long          in_lineNum = 0,        ///< optional line number used in case of error
            FormattedContents* in_pContents = 0  ///< optional container to write output to
            );

    }; // class XmlCheckImpl


    //----------------------------------------------------------------------------
    // struct XmlCheckSettings
    //----------------------------------------------------------------------------

    struct XmlCheckSettings : public SettingsBase
    {
        /// Output mode for writing XML contents
        EventContext m_outputMode;

        /// Flag indicating whether the output contains more details
        bool m_detailedOutput;

        /// Set of attribute names to ignore when checking an XML tree
        std::set<std::string> m_attributesToIgnore;

        /// Set of node names to ignore when checking an XML tree
        std::set<std::string> m_nodesToIgnore;

        /// Set of attribute names to be treated as doubles with specific precision
        std::map<std::string, int> m_doubleAttributes;

        /// Set of node names to be treated as doubles with specific precision
        std::map<std::string, int> m_doubleNodes;

        /// Set of attribute names to be treated as doubles with specific precision
        std::set<std::string> m_pointerAttributes;

        /// Set of node names to be treated as doubles with specific precision
        std::set<std::string> m_pointerNodes;

        /// Optional user supplied function to change the value of an attribute
        /// (e.g. remove unwanted text patterns)
        XmlServices::FuncToChangeAttributeValue m_funcToChangeAttributeValue;


        // default construtor
        XmlCheckSettings()
            : m_outputMode(CTX_TEST_EVENT)
            , m_detailedOutput(false)
            , m_funcToChangeAttributeValue(&XmlServices::DoNotChangeAttributeValue)
        {
        }

        // copy constructor
        XmlCheckSettings(XmlCheckSettings const & rhs)
            : m_outputMode(rhs.m_outputMode)
            , m_detailedOutput(rhs.m_detailedOutput)
            , m_attributesToIgnore(rhs.m_attributesToIgnore)
            , m_nodesToIgnore(rhs.m_nodesToIgnore)
            , m_doubleAttributes(rhs.m_doubleAttributes)
            , m_doubleNodes(rhs.m_doubleNodes)
            , m_pointerAttributes(rhs.m_pointerAttributes)
            , m_pointerNodes(rhs.m_pointerNodes)
            , m_funcToChangeAttributeValue(rhs.m_funcToChangeAttributeValue)
        {
        }

        virtual SettingsBaseSp Clone() const
        {
            XmlCheckSettingsSp spSettingsCopy = std::make_shared<XmlCheckSettings>(*this);
            return spSettingsCopy;
        };

        virtual void Dump(EventContext in_context) const;

    }; // struct XmlCheckSettings

    //----------------------------------------------------------------------------

    template<class T>
    void DumpContainer(EventContext in_context, std::string in_nameContainer, T const & in_container)
    {
        if (!in_container.empty())
        {
            TTB_WRITE_S(in_context, in_nameContainer << ":");
            for (auto const & element : in_container)
            {
                TTB_WRITE_S(in_context, "  " << element);
            }
        }
    }

    //----------------------------------------------------------------------------

    template<class T>
    void DumpMap(EventContext in_context, std::string in_nameMap, T const & in_map)
    {
        if (!in_map.empty())
        {
            TTB_WRITE_S(in_context, in_nameMap << ":");
            for (auto const & element : in_map)
            {
                TTB_WRITE_S(in_context, "  " << element.first << ": " << element.second);
            }
        }
    }

    //----------------------------------------------------------------------------

    void XmlCheckSettings::Dump(EventContext in_context) const
    {
        TTB_WRITE_S(in_context, "m_outputMode: " << m_outputMode);
        TTB_WRITE_S(in_context, "m_detailedOutput: " << m_detailedOutput);
        DumpContainer(in_context, "m_attributesToIgnore", m_attributesToIgnore);
        DumpContainer(in_context, "m_nodesToIgnore", m_nodesToIgnore);
        DumpMap(in_context, "m_doubleAttributes with precision", m_doubleAttributes);
        DumpMap(in_context, "m_doubleNodes with precision", m_doubleNodes);
        DumpContainer(in_context, "m_pointerAttributes", m_pointerAttributes);
        DumpContainer(in_context, "m_pointerNodes", m_pointerNodes);
    }

    //----------------------------------------------------------------------------
    // Implementation of class AutoRestoreXmlCheck
    //----------------------------------------------------------------------------

    TTB_INLINE AutoRestoreXmlCheck::AutoRestoreXmlCheck()
    {
        m_spSettings = TheXmlCheck()->CopyCurrentSettings();
    }

    //----------------------------------------------------------------------------

    TTB_INLINE AutoRestoreXmlCheck::~AutoRestoreXmlCheck()
    {
        TheXmlCheck()->RestoreCurrentSettings(m_spSettings);
    }

    //-------------------------------------------------------------------------
    // Implementation of XmlServices
    //----------------------------------------------------------------------------

    namespace XmlServices
    {
        TTB_INLINE void DoNotChangeAttributeValue(
            char const *  /*in_attributeName*/,
            std::string & /*io_attributeValue*/)
        {
            // do not change anything
        }
    }


    //-------------------------------------------------------------------------
    // Helper functions
    //----------------------------------------------------------------------------

    namespace XmlHelper
    {
        TTB_INLINE std::string  WritePointerValue(std::string const & in_pointer)
        {
            if (in_pointer == "")
            {
                return "not set";
            }
            else if ((in_pointer == "0") || (in_pointer == "00000000") || (in_pointer == "0x00000000"))
            {
                return "0";
            }
            else
            {
                return "exists";
            }
        }

        //----------------------------------------------------------------------------

        struct StringPosition
        {
            explicit StringPosition(
                std::string::size_type in_pos = 0)
                : m_curPos(in_pos)
                , m_lastPos(in_pos)
                , m_lineStr()
            {}

            std::string::size_type m_curPos;
            std::string::size_type m_lastPos;
            std::string            m_lineStr;
        };

        //----------------------------------------------------------------------------

        // Extract the next substring delimited by new line
        TTB_INLINE bool GetNextLineStr(
            std::string const & in_stringWithNewLines,
            StringPosition &    io_current)
        {
            if (io_current.m_lastPos == std::string::npos) return false;

            io_current.m_curPos = in_stringWithNewLines.find_first_of('\n', io_current.m_lastPos);
            io_current.m_lineStr = in_stringWithNewLines.substr(io_current.m_lastPos,
                io_current.m_curPos - io_current.m_lastPos);

            if (io_current.m_curPos != std::string::npos)
                io_current.m_lastPos = io_current.m_curPos + 1;
            else
                io_current.m_lastPos = std::string::npos;

            return true;
        }

    } // namespace XmlHelper


    //----------------------------------------------------------------------------
    // Implementation of class XmlCheck
    //----------------------------------------------------------------------------
      
    TTB_INLINE XmlCheck* TheXmlCheck(void)
    {
        return XmlCheck::Get();
    }

    //----------------------------------------------------------------------------

    TTB_INLINE XmlCheck* XmlCheck::Get(void)
    {
        if (!s_pXmlCheck) s_pXmlCheck = new XmlCheckImpl;
        return s_pXmlCheck;
    }

    //----------------------------------------------------------------------------

    TTB_INLINE void XmlCheck::Cleanup(void)
    {
        if (s_pXmlCheck)
        {
            delete s_pXmlCheck;
            s_pXmlCheck = 0;
        }
    }

    //----------------------------------------------------------------------------

    TTB_INLINE XmlCheckSp XmlCheck::CreateLocalInstance(IEventReceiver* in_pEventReceiver)
    {
        XmlCheckSp spXmlCheck = std::make_shared<XmlCheckImpl>();
        spXmlCheck->SetEventReceiver(in_pEventReceiver);
        return spXmlCheck;
    }

    //----------------------------------------------------------------------------

    TTB_INLINE pugi::xml_node XmlCheck::NodeFromXPath(
        pugi::xml_node const & in_rParentNode,
        std::string const & in_xPathExpression)
    {
        pugi::xml_node foundNode;

        pugi::xpath_node xpathNode = in_rParentNode.select_single_node(in_xPathExpression.c_str());
        if (xpathNode)
        {
            foundNode = xpathNode.node();
        }
        return foundNode;
    }


    //----------------------------------------------------------------------------
    // Implementation of class XmlCheckImpl
    //----------------------------------------------------------------------------

    TTB_INLINE XmlCheckImpl::XmlCheckImpl()
        : m_pEventReceiver(0)
        , m_spSettings(new XmlCheckSettings())
    {
        TRCF("XmlCheckImpl::XmlCheckImpl");
        TRC(TL_DEBUG, "Constructor");
    }

    //----------------------------------------------------------------------------

    TTB_INLINE XmlCheckImpl::~XmlCheckImpl()
    {
        TRCF("XmlCheckImpl::~XmlCheckImpl");
        TRC(TL_DEBUG, "Destructor");
    }

    //----------------------------------------------------------------------------

    TTB_INLINE void XmlCheckImpl::Dump(EventContext in_context) const
    {
        m_spSettings->Dump(in_context);
        if (!m_compareResult.empty())
        {
            TTB_WRITE_S(in_context, "m_compareResult: " << m_compareResult);
        }
    }

    //----------------------------------------------------------------------------

    TTB_INLINE IEventReceiver* XmlCheckImpl::SetEventReceiver(
        IEventReceiver* in_pEventReceiver)
    {
        IEventReceiver* retVal = m_pEventReceiver;
        m_pEventReceiver = in_pEventReceiver;
        return retVal;
    }

    //------------------------------------------------------------------------_

    TTB_INLINE void XmlCheckImpl::SetDefaults(void)
    {
        m_spSettings.reset(new XmlCheckSettings());
    }

    //------------------------------------------------------------------------_

    TTB_INLINE SettingsBaseSp XmlCheckImpl::CopyCurrentSettings() const
    {
        return m_spSettings->Clone();
    }

    //------------------------------------------------------------------------_

    TTB_INLINE void XmlCheckImpl::RestoreCurrentSettings(SettingsBaseSp in_spSettings)
    {
        m_spSettings = std::dynamic_pointer_cast<XmlCheckSettings>(in_spSettings);
    }

    //------------------------------------------------------------------------_

    TTB_INLINE void XmlCheckImpl::SetOutputModeForWritingXmlContents(
        EventContext in_outputMode)
    {
        if ((in_outputMode != CTX_PROT)
            && (in_outputMode != CTX_INFO)
            && (in_outputMode != CTX_TEST_EVENT))
        {
            std::ostringstream oss;
            oss << "Invalid output mode cannot be set, mode=" << in_outputMode;
            if (m_pEventReceiver)
                m_pEventReceiver->EventMsg(CTX_ERROR, oss.str().c_str());
        }
        else
        {
            m_spSettings->m_outputMode = in_outputMode;
        }
    }

    //------------------------------------------------------------------------_

    TTB_INLINE void XmlCheckImpl::SetDetailedOutput(
        bool in_detailedOutput)
    {
        m_spSettings->m_detailedOutput = in_detailedOutput;
    }

    //------------------------------------------------------------------------_

    TTB_INLINE void XmlCheckImpl::IgnoreAttribute(
        std::string const & in_attribute,
        bool                in_ignore)
    {
        if (in_ignore)
        {
            m_spSettings->m_attributesToIgnore.insert(in_attribute);
        }
        else
        {
            m_spSettings->m_attributesToIgnore.erase(in_attribute);
        }
    }

    //------------------------------------------------------------------------_

    TTB_INLINE bool XmlCheckImpl::IsAttributeToIgnore(
        std::string const & in_attribute) const
    {
        return m_spSettings->m_attributesToIgnore.find(in_attribute) != m_spSettings->m_attributesToIgnore.end();
    }

    //------------------------------------------------------------------------_

    TTB_INLINE void XmlCheckImpl::IgnoreNode(
        std::string const & in_nodeName,
        bool                in_ignore)
    {
        if (in_ignore)
        {
            m_spSettings->m_nodesToIgnore.insert(in_nodeName);
        }
        else
        {
            m_spSettings->m_nodesToIgnore.erase(in_nodeName);
        }
    }

     //------------------------------------------------------------------------_

    TTB_INLINE bool XmlCheckImpl::IsNodeToIgnore(
        std::string const & in_nodeName) const
    {
        return m_spSettings->m_nodesToIgnore.find(in_nodeName) != m_spSettings->m_nodesToIgnore.end();
    }

    //------------------------------------------------------------------------_

    TTB_INLINE void XmlCheckImpl::UseAsDoubleAttribute(
        std::string const & in_attribute,
        bool                in_useAsDouble,
        int                 in_precision)
    {
        if (in_useAsDouble)
        {
            m_spSettings->m_doubleAttributes[in_attribute] = in_precision;
        }
        else
        {
            m_spSettings->m_doubleAttributes.erase(in_attribute);
        }
    }

    //------------------------------------------------------------------------_

    TTB_INLINE bool XmlCheckImpl::IsDoubleAttribute(
        std::string const & in_attribute) const
    {
        return m_spSettings->m_doubleAttributes.find(in_attribute) != m_spSettings->m_doubleAttributes.end();
    }

    //------------------------------------------------------------------------_

    TTB_INLINE int XmlCheckImpl::GetPrecisionOfDoubleAttribute(
        std::string const & in_attribute) const
    {
        int retVal = 0;
        std::map<std::string, int>::const_iterator it = m_spSettings->m_doubleAttributes.find(in_attribute);
        if (it != m_spSettings->m_doubleAttributes.end())
        {
            retVal = it->second;
        }
        return retVal;
    }

    //------------------------------------------------------------------------_

    TTB_INLINE void XmlCheckImpl::UseAsDoubleNode(
        std::string const & in_nodeName,
        bool                in_useAsDouble,
        int                 in_precision)
    {
        if (in_useAsDouble)
        {
            m_spSettings->m_doubleNodes[in_nodeName] = in_precision;
        }
        else
        {
            m_spSettings->m_doubleNodes.erase(in_nodeName);
        }
    }

    //------------------------------------------------------------------------_

    TTB_INLINE bool XmlCheckImpl::IsDoubleNode(
        std::string const & in_nodeName) const
    {
        return m_spSettings->m_doubleNodes.find(in_nodeName) != m_spSettings->m_doubleNodes.end();
    }

    //------------------------------------------------------------------------_

    TTB_INLINE int XmlCheckImpl::GetPrecisionOfDoubleNode(
        std::string const & in_nodeName) const
    {
        int retVal = 0;
        std::map<std::string, int>::const_iterator it = m_spSettings->m_doubleNodes.find(in_nodeName);
        if (it != m_spSettings->m_doubleNodes.end())
        {
            retVal = it->second;
        }
        return retVal;
    }

    //------------------------------------------------------------------------_

    TTB_INLINE void XmlCheckImpl::UseAsPointerAttribute(
        std::string const & in_attribute,
        bool                in_useAspointer)
    {
        if (in_useAspointer)
        {
            m_spSettings->m_pointerAttributes.insert(in_attribute);
        }
        else
        {
            m_spSettings->m_pointerAttributes.erase(in_attribute);
        }
    }

    //------------------------------------------------------------------------_

    TTB_INLINE bool XmlCheckImpl::IsPointerAttribute(
        std::string const & in_attribute) const
    {
        return m_spSettings->m_pointerAttributes.find(in_attribute) != m_spSettings->m_pointerAttributes.end();
    }

    //------------------------------------------------------------------------_

    TTB_INLINE void XmlCheckImpl::UseAsPointerNode(
        std::string const & in_nodeName,
        bool                in_useAspointer)
    {
        if (in_useAspointer)
        {
            m_spSettings->m_pointerNodes.insert(in_nodeName);
        }
        else
        {
            m_spSettings->m_pointerNodes.erase(in_nodeName);
        }
    }

    //------------------------------------------------------------------------_

    TTB_INLINE bool XmlCheckImpl::IsPointerNode(
        std::string const & in_nodeName) const
    {
        return m_spSettings->m_pointerNodes.find(in_nodeName) != m_spSettings->m_pointerNodes.end();
    }

    //------------------------------------------------------------------------_

    TTB_INLINE void XmlCheckImpl::SetFuncToChangeAttributeValue(
        XmlServices::FuncToChangeAttributeValue in_func)
    {
        m_spSettings->m_funcToChangeAttributeValue = in_func;
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void XmlCheckImpl::CheckNode(
        pugi::xml_node const &  in_rNode,
        char const *            in_file,
        long                    in_lineNum)
    {
        WriteNode(in_rNode, 0, 0, in_file, in_lineNum);
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void XmlCheckImpl::CheckSubNode(
        pugi::xml_node const &  in_rNode,
        std::string const & in_xPathExpression,
        char const *        in_file,
        long                in_lineNum)
    {
        WriteNode(
            NodeFromXPath(in_rNode, in_xPathExpression),
            0, 0, in_file, in_lineNum);
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void XmlCheckImpl::CheckSubNodeWithMaxChilds(
        int                 in_maxChilds,
        pugi::xml_node const &  in_rNode,
        std::string const & in_xPathExpression,
        char const *        in_file,
        long                in_lineNum)
    {
        WriteNode(
            NodeFromXPath(in_rNode, in_xPathExpression),
            0, in_maxChilds, in_file, in_lineNum);
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE bool XmlCheckImpl::Compare(
        pugi::xml_node const &  in_rNodeA,
        pugi::xml_node const &  in_rNodeB,
        char const *            in_file,
        long                    in_lineNum)
    {
        std::ostringstream oss;

        FormattedContents contA;
        WriteNode(in_rNodeA, 0, 0, in_file, in_lineNum, &contA);

        FormattedContents contB;
        WriteNode(in_rNodeB, 0, 0, in_file, in_lineNum, &contB);

        // Compare the two text containers with read XML data
        bool differenceFound = false;
        FormattedContents::iterator itA = contA.begin();
        FormattedContents::iterator itB = contB.begin();
        while ((itA != contA.end())
            && (!differenceFound))
        {
            // Remove leading and trailing white space
            std::string strA = itA->m_text;
            TrimStr(strA);

            std::string strB;
            if (itB != contB.end())
            {
                strB = itB->m_text;
                TrimStr(strB);
            }

            // Only for debugging
            // std::cout << "Comparing " << strA << "---" << strB << std::endl;

            // Compare the strings
            if (strA != strB)
            {
                differenceFound = true;
                oss << "Difference detected\n"
                    << "lhs: " << strA
                    << "(path: " << itA->m_fullPath
                    << ", pos: " << itA->m_col << ")"
                    << std::endl;
                if (itB != contB.end())
                {
                    oss << "rhs: " << strB
                        << "(path: " << itB->m_fullPath
                        << ", pos: " << itB->m_col << ")";
                }
                else
                {
                    oss << "rhs: no data found";
                }
            }

            ++itA;
            if (itB != contB.end())
            {
                ++itB;
            }
        }

        // Check if B has additional data
        if ((!differenceFound)
            && (itA == contA.end())
            && (itB != contB.end()))
        {
            differenceFound = true;
            std::string strB = itB->m_text;
            TrimStr(strB);

            oss << "Difference detected, rhs has more data\n"
                << "rhs: " << strB
                << " (path: " << itB->m_fullPath
                << ", pos: " << itB->m_col << ")";
        }

        if (differenceFound)
        {
            m_compareResult = oss.str().c_str();
        }
        else
        {
            m_compareResult = "XML-Compare: no differences found";
        }

        if (m_pEventReceiver)
        {
            XmlHelper::StringPosition curPos;
            while (GetNextLineStr(m_compareResult, curPos))
            {
                m_pEventReceiver->Event(m_spSettings->m_outputMode, curPos.m_lineStr);
            }

        }

        return !differenceFound;
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE std::string XmlCheckImpl::GetCompareResult(void) const
    {
        return m_compareResult;
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void XmlCheckImpl::CheckNode(
        pugi::xml_document const & in_rDoc,
        char const *               in_file,
        long                       in_lineNum)
    {
        pugi::xml_node root = in_rDoc.document_element();
        CheckNode(root, in_file, in_lineNum);
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void XmlCheckImpl::CheckSubNode(
        pugi::xml_document const & in_rDoc,
        std::string const &        in_xPathExpression,
        char const *               in_file,
        long                       in_lineNum)
    {
        pugi::xml_node root = in_rDoc.document_element();
        CheckSubNode(root, in_xPathExpression, in_file, in_lineNum);
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void XmlCheckImpl::CheckSubNodeWithMaxChilds(
        int                         in_maxChilds,
        pugi::xml_document const &  in_rDoc,
        std::string const &         in_xPathExpression,
        char const *                in_file,
        long                        in_lineNum)
    {
        pugi::xml_node root = in_rDoc.document_element();
        CheckSubNodeWithMaxChilds(in_maxChilds, root, in_xPathExpression, in_file, in_lineNum);
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE bool XmlCheckImpl::Compare(
        pugi::xml_document const & in_rDocA,
        pugi::xml_document const & in_rDocB,
        char const *               in_file,
        long                       in_lineNum)
    {
        pugi::xml_node rootA = in_rDocA.document_element();
        pugi::xml_node rootB = in_rDocB.document_element();
        return Compare(rootA, rootB, in_file, in_lineNum);
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void  XmlCheckImpl::ApplyFilterToXmlTree(pugi::xml_node& io_node)
    {
        // ------  Process attributes  ----
        
        for (pugi::xml_attribute attr = io_node.first_attribute(); attr;)
        {
            pugi::xml_attribute nextAttr = attr.next_attribute();

            if (IsAttributeToIgnore(attr.name()))
            {
                io_node.remove_attribute(attr);
            }
            else
            {
                if (IsDoubleAttribute(attr.name()))
                {
                    double doubleVal = attr.as_double();
                    {
                        if (abs(doubleVal) < std::numeric_limits<double>::epsilon())
                        {
                            doubleVal = 0.0;
                        }
                        std::ostringstream oss;
                        oss << std::fixed << std::setprecision(GetPrecisionOfDoubleAttribute(attr.name())) << doubleVal;
                        attr.set_value(oss.str().c_str());
                    }
                }
                else if (IsPointerAttribute(attr.name()))
                {
                    std::ostringstream oss;
                    oss << XmlHelper::WritePointerValue(std::string(attr.as_string()));
                    attr.set_value(oss.str().c_str());
                }
            }

            attr = nextAttr;
        }

        // -----  Process node value  -----

        if (io_node.type() == pugi::node_pcdata)
        {
            pugi::xml_node parentNode = io_node.parent();
            if (IsDoubleNode(parentNode.name()))
            {
                double doubleVal = parentNode.text().as_double();
                {
                    if (abs(doubleVal) < std::numeric_limits<double>::epsilon())
                    {
                        doubleVal = 0.0;
                    }
                    std::ostringstream oss;
                    oss << std::fixed << std::setprecision(GetPrecisionOfDoubleNode(parentNode.name())) << doubleVal;
                    io_node.set_value(oss.str().c_str());
                }

            }
            else if (IsPointerNode(parentNode.name()))
            {
                std::ostringstream oss;
                oss << XmlHelper::WritePointerValue(std::string(parentNode.text().as_string()));
                io_node.set_value(oss.str().c_str());
            }
        }


        // -----  Process child nodes  -----

        for (pugi::xml_node child = io_node.first_child(); child; )
        {
            pugi::xml_node next = child.next_sibling();

            if (IsNodeToIgnore(child.name()))
            {
                child.parent().remove_child(child);
            }
            else
            {
                // recursive call for child node
                ApplyFilterToXmlTree(child);
            }
            child = next;
        }
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void XmlCheckImpl::ApplyFilterToXmlTree(pugi::xml_document& io_docFromFile)
    {
        pugi::xml_node rootNode = io_docFromFile.document_element();
        ApplyFilterToXmlTree(rootNode);
    }

    //-----------------------------------------------------------------------------

    pugi::xml_node XmlCheckImpl::ReadFromFile(pugi::xml_document& io_docFromFile, std::string const & in_xmlFilePath)
    {
        pugi::xml_node rootNode;
        pugi::xml_parse_result result = io_docFromFile.load_file(in_xmlFilePath.c_str(), pugi::parse_full | pugi::parse_trim_pcdata);
        if (!result)
        {
            std::ostringstream oss;
            oss << "Parse error: " << result.description() << ", character pos= " << result.offset;
            if (m_pEventReceiver)
                m_pEventReceiver->EventMsg(CTX_ERROR, oss.str().c_str());
        }
        else
        {
            rootNode = io_docFromFile.document_element();
        }
        return rootNode;
    }

    //-----------------------------------------------------------------------------

    bool XmlCheckImpl::SaveToFile(pugi::xml_document& io_docFromFile, std::string const & in_xmlFilePath)
    {
        bool saveSucceeded = io_docFromFile.save_file(in_xmlFilePath.c_str(), PUGIXML_TEXT("  "));
        if (!saveSucceeded)
        {
            std::ostringstream oss;
            oss << "Save error: cannot save to file " << in_xmlFilePath;
            if (m_pEventReceiver)
                m_pEventReceiver->EventMsg(CTX_ERROR, oss.str().c_str());
        }
        return saveSucceeded;
    }

    //-----------------------------------------------------------------------------

    pugi::xml_node XmlCheckImpl::ReadFromString(pugi::xml_document& io_docFromString, std::string const & in_xmlTreeAsString)
    {
        pugi::xml_node rootNode;
        pugi::xml_parse_result result = io_docFromString.load(in_xmlTreeAsString.c_str(), pugi::parse_full | pugi::parse_trim_pcdata);
        if (!result)
        {
            std::ostringstream oss;
            oss << "Parse error: " << result.description() << ", character pos= " << result.offset;
            if (m_pEventReceiver)
                m_pEventReceiver->EventMsg(CTX_ERROR, oss.str().c_str());
        }
        else
        {
            rootNode = io_docFromString.document_element();
        }
        return rootNode;
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE std::string XmlCheckImpl::IfDetailed(std::string const & in_str)
    {
        return m_spSettings->m_detailedOutput ? in_str : "";
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE const char * XmlCheckImpl::IndentStr(unsigned int in_numIndents)
    {
        static const char* INDENT = "                                        ";
        static const unsigned int LENGTH = static_cast<int>(strlen(INDENT));
        unsigned int n = in_numIndents*NUM_INDENTS_PER_SPACE;
        if (n > LENGTH) n = LENGTH;

        return &INDENT[LENGTH - n];
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE const char * XmlCheckImpl::IndentEmptyStr(unsigned int in_numIndents)
    {
        static const char* INDENT = "                                        ";
        static const unsigned int LENGTH = static_cast<int>(strlen(INDENT));
        unsigned int n = in_numIndents*NUM_INDENTS_PER_SPACE;
        if (n > LENGTH) n = LENGTH;

        return &INDENT[LENGTH - n];
    }

    //-----------------------------------------------------------------------------

    TTB_INLINE void XmlCheckImpl::WriteNode(
        pugi::xml_node const &         in_rNode,
        unsigned int       in_indent,
        int                in_maxChildNodes,
        char const *       in_file,
        long               in_lineNum,
        FormattedContents* in_pContents)
    {
        TRCF("XmlCheckImpl::WriteNode");
        TRC(TL_DEBUG, "Begin");

        //	if (!in_pNode) return;

        pugi::xml_node_type nodeType = in_rNode.type();

        std::ostringstream oss;
        oss << IndentStr(in_indent);

        bool writeAttributes = false;
        switch (nodeType)
        {
        case pugi::node_null:
            return;
            break;
        case pugi::node_document:
            oss << IfDetailed("Document");
            break;

        case pugi::node_element:
            oss << IfDetailed("Element: ") << in_rNode.name();
            writeAttributes = true;
            break;

        case pugi::node_comment:
            oss << IfDetailed("Comment: ") << in_rNode.name();
            break;

        case pugi::node_pcdata: // direct text/value within node
        {
            pugi::xml_node parentNode = in_rNode.parent();
            if (IsDoubleNode(parentNode.name()))
            {
                double doubleVal = parentNode.text().as_double();
                oss << std::fixed << std::setprecision(GetPrecisionOfDoubleNode(parentNode.name())) << doubleVal;
            }
            else if (IsPointerNode(parentNode.name()))
            {
                oss << XmlHelper::WritePointerValue(std::string(parentNode.text().as_string()));
            }
            else
            {
                oss << IfDetailed("Text: ") << parentNode.text().as_string();
            }
        }
        break;

        case pugi::node_declaration:
            oss << IfDetailed("Declaration") << " version=????";// << in_pNode->ToDeclaration()->Version();
            break;

        default:
            oss << "Unexpected nodeType=" << nodeType;
            break;
        }

        if (m_spSettings->m_detailedOutput)
        {
            // oss << " (line: " << in_pNode->Row() << ", col: " << in_pNode->Column() << ")";
        }

        if (oss.str().length() > 0)
        {
            if (in_pContents)
            {
                in_pContents->push_back(TextWithPosInfo(
                    0/*in_pNode->Row()*/, in_rNode.offset_debug(), in_rNode.path()/*GetFullPath(in_pNode)*/,
                    oss.str().c_str()));
            }
            else if (m_pEventReceiver)
            {
                m_pEventReceiver->EventMsg(m_spSettings->m_outputMode, oss.str().c_str());
            }
        }

        if (writeAttributes)
        {
            WriteAttributes(in_rNode, in_indent + 1, in_file, in_lineNum, in_pContents);
        }

        // Iterate over children
        int numReportedChilds = 0;

        for (pugi::xml_node child : in_rNode.children())
        {
            if (IsNodeToIgnore(child.name())) continue;
            WriteNode(child, nodeType != pugi::node_document ? in_indent + 1 : in_indent, 0, in_file, in_lineNum, in_pContents);
            ++numReportedChilds;
            if (in_maxChildNodes && (numReportedChilds >= in_maxChildNodes))
                break;
        }

        TRC(TL_DEBUG, "End");
    }

    //----------------------------------------------------------------------------

    TTB_INLINE void XmlCheckImpl::WriteAttributes(
        pugi::xml_node const &         in_rNode,
        unsigned int       in_indent,
        char const *       /*in_file*/,
        long               /*in_lineNum*/,
        FormattedContents* in_pContents)
    {
        // if ( !in_pElement ) return;

        //		TiXmlAttribute* pAttrib=in_pElement->FirstAttribute();
        int    numAttributes = 0;
        int    intVal;
        double doubleVal;

        for (pugi::xml_attribute attr : in_rNode.attributes())
        {
            if (!IsAttributeToIgnore(attr.name()))
            {
                std::ostringstream oss;
                oss << IndentStr(in_indent)
                    << attr.name() << ": ";
                if (IsDoubleAttribute(attr.name()))
                {
                    doubleVal = attr.as_double();
                    {
                        if (abs(doubleVal) < std::numeric_limits<double>::epsilon())
                        {
                            doubleVal = 0.0;
                        }
                        oss << std::fixed << std::setprecision(GetPrecisionOfDoubleAttribute(attr.name())) << doubleVal;
                    }
                }
                else if (IsPointerAttribute(attr.name()))
                {
                    oss << XmlHelper::WritePointerValue(std::string(attr.as_string()));
                }
                else
                {
                    // Give the user a chance to change / filter attribute text
                    std::string attributeValueAsText = attr.as_string();
                    m_spSettings->m_funcToChangeAttributeValue(attr.name(), attributeValueAsText);

                    oss << attributeValueAsText;

                    if (m_spSettings->m_detailedOutput)
                    {
                        // Try to read as integer
                        intVal = attr.as_int();
                        oss << ", as int: " << intVal;

                        // Try to read as double
                        doubleVal = attr.as_double();
                        oss << ", as double: " << doubleVal;
                    }
                }

                if (m_spSettings->m_detailedOutput)
                {
                    oss << " (offset: " << in_rNode.offset_debug() << ")"; // TODO: calculate line/row?
                }

                if (in_pContents)
                {
                    in_pContents->push_back(TextWithPosInfo(
                        0/*pAttrib->Row()*/, in_rNode.offset_debug(), in_rNode.path() + "/" + attr.name(),
                        oss.str().c_str()));
                }
                else if (m_pEventReceiver)
                {
                    m_pEventReceiver->EventMsg(m_spSettings->m_outputMode, oss.str().c_str());
                }

                numAttributes++;
            }
        }

        if (m_spSettings->m_detailedOutput && (numAttributes > 0))
        {
            std::ostringstream oss;
            oss << IndentEmptyStr(in_indent) << numAttributes << " attributes";
            if (in_pContents)
            {
                in_pContents->push_back(TextWithPosInfo(
                    0/*pAttrib->Row()*/, in_rNode.offset_debug(), in_rNode.path(),
                    oss.str().c_str()));
            }
            else if (m_pEventReceiver)
            {
                m_pEventReceiver->EventMsg(m_spSettings->m_outputMode, oss.str().c_str());
            }
        }

    } // XmlCheckImpl::WriteAttributes

    //----------------------------------------------------------------------------

    __declspec(selectany) XmlCheck* XmlCheck::s_pXmlCheck = 0;
    __declspec(selectany) const unsigned int XmlCheckImpl::NUM_INDENTS_PER_SPACE;

    //----------------------------------------------------------------------------

} // namespace TestToolBox

  ///\endcond
